#!python
# Copyright (c) 2000-2001 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject
import cvslib, uriparse, os, process, re, sys
import timeline
import koprocessutils

class KoCVS:
    _com_interfaces_ = [components.interfaces.koICVS, components.interfaces.nsIObserver]
    _reg_desc_ = "Komodo CVS Support"
    _reg_contractid_ = "@activestate.com/koCVS;1"
    _reg_clsid_ = "{CE39F3D2-7103-493f-8A77-EBD15071CD57}"
    
    def __init__(self):
        # generic cvs instance to do work with
        timeline.enter('KoCVS.__init__')
        self.cvs = cvslib.CVS()
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)
        self.fileSvc = components.classes["@activestate.com/koFileService;1"].\
                getService(components.interfaces.koIFileService);
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
        self.reasonNotFunctional = None
        self.isFunctional = None
        self._reVersion = re.compile(r'System\s\((.*?)\)\s([\d|\.]*?)\s',re.M|re.VERBOSE)
        self._globalPrefs.addObserver(self)

        self._userPath = koprocessutils.getUserEnv()["PATH"].split(os.pathsep)
        self._executable = None
        self._setExecutableFromPrefs()
        self.redetermineIfFunctional()
        timeline.leave('KoCVS.__init__')

    def Observe(self, subject, topic, value):
        if topic == self._globalPrefs.id and \
           subject == 'cvsExecutable':
            self.set_executable(value)
    
    def _setExecutableFromPrefs(self):
        executable = self._globalPrefs.getStringPref('cvsExecutable')
        if self._executable != executable:
            self.set_executable(executable)
            try:
                self._observerSvc.notifyObservers(self, 'subsystem_changed', 'cvs')
            except COMException:
                pass # Ignore error if there is no observer.
        
    def set_executable(self, executable):
        self._executable = executable
        # we want to allow self._executable to be None, but
        # we cannot allow self.cvs._cvs to be None
        if self._executable:
            self.cvs._cvs = self._executable
        else:
            self.cvs._cvs = 'cvs'
        
    def get_executable(self):
        return self._executable
    
    def redetermineIfFunctional(self):
        # XXX if there are multiple cvs executables, we should determine
        # which ones are usable, which is the latest version, etc.,
        # and stuff that information someplace
        timeline.enter('redetermineIfFunctional')
        import which
        assert which._version_ >= (1,0,0)
        cvsInfoEx = components.classes["@activestate.com/koCVSInfoEx;1"].\
                    createInstance(components.interfaces.koIAppInfoEx)
        if not self._executable:
            try:
                if sys.platform.startswith('win'):
                    exts = ['.exe']
                else:
                    exts = None
                cvs = which.which('cvs', exts=exts, path=self._userPath)
            except which.WhichError:
                pass
            else:
                # this call sets cvsInfoEx.executablePath
                cvsInfoEx.getInstallationPathFromBinary(cvs)
        else:
            cvsInfoEx.executablePath = self._executable;

        if not cvsInfoEx.executablePath:
            self.isFunctional = 0
            self.reasonNotFunctional = "no 'cvs' executable could be found"
        else:
            version = cvsInfoEx.version
            self.isFunctional = 0
            if not version:
                self.reasonNotFunctional = \
                    "the executable is not a CVS executable or "\
                    "is otherwise incompatible with Komodo"
            else:
                self.isFunctional = 1
                self.reasonNotFunctional = None
        # XXX it seems to me that determining whether an executable
        # is functional does not mean we should change the subsystem
        try:
            self._observerSvc.notifyObservers(self, 'subsystem_changed', 'cvs')
        except COMException:
            pass # Ignore error if there is no observer.
        timeline.leave('redetermineIfFunctional')

    def _splitFile(self, file,splitdir):
        # figure out what repositories files are in
        # and split them up into seperate lists
        # we also convert to local paths here
        try:
            path = uriparse.URIToLocalPath(file);
            if path:
                if not splitdir and os.path.isdir(path):
                    return path,''
                path = os.path.normpath(path)
                basedir = os.path.dirname(path)
                filename = os.path.basename(path)
                #print "_splitfile %s:%s:%s"% (path,basedir,filename)
                # make sure the path is in cvs
                root = self.cvs.getFileRoot(basedir)
                if root:
                    return basedir, filename
        except cvslib.CVSLibError, e:
            # we'll ignore files that are not in cvs
            pass
        return None,None
    
    def _splitFiles(self, files, splitdir=0):
        # figure out what repositories files are in
        # and split them up into seperate lists
        # we also convert to local paths here
        _files = {}
        for file in files:
            basedir,filename = self._splitFile(file,splitdir)
            if basedir:
                if not basedir in _files:
                    _files[basedir] = []
                _files[basedir].append(filename)
        return _files
    
#    cvs diff [-lNR] [rcsdiff-options]
#    [[-r rev1 | -D date1] [-r rev2 | -D date2]] [files...] 
#	-l	Local directory only, not recursive
#	-R	Process directories recursively.
#	-D d1	Diff revision for date against working file.
#	-D d2	Diff rev1/date1 against date2.
#	-N	include diffs for added and removed files.
#	-r rev1	Diff revision for rev1 against working file.
#	-r rev2	Diff rev1/date1 against rev2.
#	--ifdef=arg	Output diffs in ifdef format.
    def diff(self, files, options, external):
        #print "cvs diff ",repr(files)
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        diff = ''
        for basedir in _files.keys():
            #print "diff files [%s] :%s" % (basedir,repr(_files[basedir]))
            try:
                result = self.cvs.diff(files=_files[basedir],
                                   cwd=basedir,
                                   diffOpts=options.split())
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            
            if 'stdout' in result:
                diff += result['stdout']
            elif 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))

        # Convert line-endings to the currently expected
        import eollib
        eolpref = self._globalPrefs.getStringPref('endOfLine')
        try:
            desiredEOL = eollib.eolPref2eol[eolpref]
        except KeyError:
            desiredEOL = eollib.EOL_PLATFORM
        diff = eollib.convertToEOLFormat(diff, desiredEOL)

        return diff

#    cvs add [-k rcs-kflag] [-m message] files...
#	-k	Use "rcs-kflag" to add the file with the specified kflag.
#	-m	Use "message" for the creation log.
    def add(self, files, mode, message):
        #add(self, files, mode=None, msg=None, cwd=None)
        # first we figure out what repositories files are in
        # and split them up into seperate lists
        # we init the cvs object with the root we retreived
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files,1)
        output = ''
        for basedir in _files.keys():
            try:
                result = self.cvs.add(files=_files[basedir], mode=mode, msg=message, cwd=basedir)
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            
            if 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))
                output += result['stderr']
            if 'stdout' in result:
                output += result['stdout']
            
        # do notification that status may have changed
        # we have to use the original urls passed in since
        # they are modified to local paths in _splitFiles
        for file in files:
            if file: self._observerSvc.notifyObservers(self, 'file_status_now', file)

        return output
    
#    cvs commit [-nRlf] [-m msg | -F logfile] [-r rev] files...
#	-n	Do not run the module program (if any).
#	-R	Process directories recursively.
#	-l	Local directory only (not recursive).
#	-f	Force the file to be committed; disables recursion.
#	-F file	Read the log message from file.
#	-m msg	Log message.
#	-r rev	Commit to this branch or trunk revision.
    def commit(self, files, message, options):
        #print "cvs commit ",repr(files)
        # save the message to a temp file
        msgfile = self.fileSvc.makeTempFile('.cvs','w')
        msgfile.puts(message)
        msgfile.flush()
        msgfile.close()
        msgpath = msgfile.path

        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            #print "commit files [%s] :%s" % (basedir,repr(_files[basedir]))
            try:
                result = self.cvs.commit(_files[basedir],msgFile=msgpath,cwd=basedir)
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            #print repr(result)
            if 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))
                output += result['stderr']
            if 'stdout' in result:
                output += result['stdout']
        
        del msgfile
        
        for file in files:
            self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
    
#    cvs remove [-flR] [files...]
#	-f	Delete the file before removing it.
#	-l	Process this directory only (not recursive).
#	-R	Process directories recursively.
    def remove(self, files, force, recursive):
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            try:
                result = self.cvs.remove(_files[basedir], force, recursive,cwd=basedir)
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))
                output += result['stderr']
            if 'stdout' in result:
                output += result['stdout']
            
        for file in files:
            self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
    
#    cvs update [-APdflRp] [-k kopt] [-r rev|-D date] [-j rev]
#    [-I ign] [-W spec] [files...]
#	-A	Reset any sticky tags/date/kopts.
#	-P	Prune empty directories.
#	-d	Build directories, like checkout does.
#	-f	Force a head revision match if tag/date not found.
#	-l	Local directory only, no recursion.
#	-R	Process directories recursively.
#	-p	Send updates to standard output (avoids stickiness).
#	-k kopt	Use RCS kopt -k option on checkout.
#	-r rev	Update using specified revision/tag (is sticky).
#	-D date	Set date to update from (is sticky).
#	-j rev	Merge in changes made between current revision and rev.
#	-I ign	More files to ignore (! to reset).
#	-W spec	Wrappers specification line.
    def update(self, files, options):
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            try:
                result = self.cvs.update(_files[basedir],cwd=basedir)
            except cvslib.CVSLibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in result:
                self.lastErrorSvc.setLastError(0, str(result['stderr']))
                
            # build a better status output for komodo
            if 'files' in result:
                for file in result['files']:
                    localFileName = os.path.normpath(os.path.join(basedir,file['file']))
                    url = uriparse.localPathToURI(localFileName)
                    #print "%s %s %s" %(file['status'],file['file'],url)
                    output += "%s %s %s\n" % (file['status'],file['file'],url)
                    
        for file in files:
            self._observerSvc.notifyObservers(self, 'file_status_now', file)
        
        return output

    def revert(self, files, options):
        # XXX unfortunately this function requires knowledge of
        # item properties
        self.lastErrorSvc.setLastError(0, '')
        output = ''
        for file in files:
            kofile = self.fileSvc.getFileFromURI(file)
            
            # splitfile does more than initial appearances
            basedir,filename = self._splitFile(file,0)
            #print "attempting to revert file ",file
            if kofile.sccAction:
                # we are edit, delete or sync
                try:
                    if kofile.sccAction == 'delete':
                        #print "adding deleted file back into cvs"
                        result = self.cvs.add(files=filename,cwd=basedir)
                    else:
                        #print "doing CLEAN update"
                        result = self.cvs.update(files=filename,clean=1,cwd=basedir)
                except cvslib.CVSLibError, e:
                    self.lastErrorSvc.setLastError(0, str(e))
                    return ''
                if 'stderr' in result:
                    output += result['stderr']
                    self.lastErrorSvc.setLastError(0, str(result['stderr']))
                if 'stdout' in result:
                    output += result['stdout']
                self._observerSvc.notifyObservers(self, 'file_status_now', file)
        return output
   
