#!python
# Copyright (c) 2003 ActiveState Corp.
# See the file LICENSE.txt for licensing information.

"""Komodo DBGP XPCOM component and service implementations.

Dev Notes:
- Mostly the Komodo DBGP XPCOM components are light wrappers around the
  equivalently named classes in dbgpServerAPI.py
"""

import os
import sys
import types

from xpcom import components, COMException, ServerException, nsError
from xpcom.server import WrapObject, UnwrapObject
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC

import dbgpServerAPI, dbgpErrors
from dbgpErrors import DBGPError
from koTreeView import TreeView
import socket, string, base64, types
import threading
from xml.dom import minidom

import uriparse
import logging

log = logging.getLogger("dbgp")
#log.setLevel(logging.DEBUG)

# due to the multithreaded nature of the classes in this file,
# it may be best to have these initialized in the main thread
# when this file is initialy loaded, and use the global proxy
# vars from the classes
# we must also avoid doing this during xpcom registration by
# not doing this in main, so this is actually filled in during the
# managers init.
_proxyMgr = None
_observerSvc = None
_observerProxy = None
_globalPrefsSvc = None
_globalPrefsSvcProxy = None

def _initProxies():
    global _proxyMgr, _observerSvc, _observerProxy, _globalPrefsSvc, _globalPrefsSvcProxy
    _proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
        getService(components.interfaces.nsIProxyObjectManager)
    _observerSvc = components.classes["@mozilla.org/observer-service;1"].\
        getService(components.interfaces.nsIObserverService)
    _observerProxy = _proxyMgr.getProxyForObject(None,
        components.interfaces.nsIObserverService, _observerSvc,
        PROXY_ALWAYS | PROXY_ASYNC)
    _globalPrefsSvc = components.classes["@activestate.com/koPrefService;1"].\
        getService(components.interfaces.koIPrefService)
    _globalPrefsSvcProxy = _proxyMgr.getProxyForObject(None,
        components.interfaces.koIPrefService, _globalPrefsSvc,
        PROXY_ALWAYS | PROXY_SYNC)

    
dbgpServerAPI._dbgpDataType = dbgpServerAPI.dbgpDataType
class koDBGPDataType(dbgpServerAPI._dbgpDataType):
    _com_interfaces_ = [components.interfaces.koIDBGPDataType]
    _reg_clsid_ = "{E5D94B47-48B8-4795-AC39-7D994B8CACEE}"
    _reg_contractid_ = "@activestate.com/koDBGPDataType;1"
dbgpServerAPI.dbgpDataType = koDBGPDataType

dbgpServerAPI._dbgpBreakpoint = dbgpServerAPI.dbgpBreakpoint
class koDBGPBreakpoint(dbgpServerAPI._dbgpBreakpoint):
    _com_interfaces_ = [components.interfaces.koIDBGPBreakpoint]
    _reg_clsid_ = "{690C35D6-3CD4-46C2-802B-EA8BE30A045D}"
    _reg_contractid_ = "@activestate.com/koDBGPBreakpoint;1"

    __safe_for_unpickling__ = 1 # required so unpickling will use __getinitargs__'s retval
    def __getstate__(self):
        # For persisting breakpoints via the pickle module.
        state = {}
        for attrName in self._attrs:
            state[attrName] = getattr(self, attrName)
        return state
    def __getinitargs__(self):
        # Define this to ensure dbgpBreakpoint.__init__() gets called
        # on unpickling.
        return tuple()

dbgpServerAPI.dbgpBreakpoint = koDBGPBreakpoint


dbgpServerAPI._dbgpSpawnpoint = dbgpServerAPI.dbgpSpawnpoint
class koDBGPSpawnpoint(dbgpServerAPI._dbgpSpawnpoint):
    _com_interfaces_ = [components.interfaces.koIDBGPSpawnpoint]
    _reg_clsid_ = "{1B829885-35C7-4102-BA35-3B7522AFDB3B}"
    _reg_contractid_ = "@activestate.com/koDBGPSpawnpoint;1"

    __safe_for_unpickling__ = 1 # required so unpickling will use __getinitargs__'s retval
    def __getstate__(self):
        # For persisting breakpoints via the pickle module.
        state = {}
        for attrName in self._attrs:
            state[attrName] = getattr(self, attrName)
        return state
    def __getinitargs__(self):
        # Define this to ensure dbgpBreakpoint.__init__() gets called
        # on unpickling.
        return tuple()

dbgpServerAPI.dbgpSpawnpoint = koDBGPSpawnpoint


dbgpServerAPI._dbgpContextType = dbgpServerAPI.dbgpContextType
class koDBGPContextType(dbgpServerAPI._dbgpContextType):
    _com_interfaces_ = [components.interfaces.koIDBGPContextType]
    _reg_clsid_ = "{045EAAA9-28F6-463F-8FA2-044EF6163FDC}"
    _reg_contractid_ = "@activestate.com/koDBGPContextType;1"
dbgpServerAPI.dbgpContextType = koDBGPContextType

dbgpServerAPI._dbgpStackFrame = dbgpServerAPI.dbgpStackFrame
class koDBGPStackFrame(dbgpServerAPI._dbgpStackFrame):
    _com_interfaces_ = [components.interfaces.koIDBGPStackFrame]
    _reg_clsid_ = "{1054AE33-C433-4954-ADCF-7259096415C7}"
    _reg_contractid_ = "@activestate.com/koDBGPStackFrame;1"
dbgpServerAPI.dbgpStackFrame = koDBGPStackFrame

dbgpServerAPI._dbgpProperty = dbgpServerAPI.dbgpProperty
class koDBGPProperty(dbgpServerAPI._dbgpProperty):
    _com_interfaces_ = [components.interfaces.koIDBGPProperty]
    _reg_clsid_ = "{9483F743-D051-4D2F-8C88-B9EFA77D32B1}"
    _reg_contractid_ = "@activestate.com/koDBGPProperty;1"

    # XPCOM getter to avoid returning unescaped binary data (29345)
    def get_value(self):
        if type(self.value) == types.UnicodeType:
            # unicode type is ok to return un-repr'd
            return self.value
        try:
            return unicode(self.value)
        except:
            return repr(self.value)[1:-1] # remove quotes

    def setValue(self, value, type):
        try:
            dbgpServerAPI._dbgpProperty.setValue(self, value, type)
        except DBGPError, ex:
            errno, errmsg = dbgpServerAPI.getDBGPErrorInfo(ex)
            lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"].\
                getService(components.interfaces.koILastErrorService)
            lastErrorSvc.setLastError(errno, errmsg)
            raise ServerException(nsError.NS_ERROR_FAILURE, errmsg)

dbgpServerAPI.dbgpProperty = koDBGPProperty

dbgpServerAPI._dbgpSession = dbgpServerAPI.dbgpSession
class koDBGPSession(dbgpServerAPI._dbgpSession):
    _com_interfaces_ = [components.interfaces.koIDBGPSession]
    _reg_clsid_ = "{367AA61A-9DEF-409D-B85C-E5209E7FF984}"
    _reg_contractid_ = "@activestate.com/koDBGPSession;1"

    def __init__(self, sessionHost):
        dbgpServerAPI._dbgpSession.__init__(self, sessionHost)
        self._lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"].\
            getService(components.interfaces.koILastErrorService)
        self.lastNotification = None
       
    def get_status(self):
        try:
            return dbgpErrors.status_names.index(self.statusName)
        except ValueError, e:
            return dbgpErrors.STATUS_STARTING
    
    def get_reason(self):
        return dbgpErrors.reason_names.index(self.reason)

    # notify is not an xpcom interface
    def notify(self, name, data):
        # THREAD WARNING
        # this function is called from the command loop thread.  Do
        # not do anything here that will result in another command
        # being sent to the client, that will result in a lockup
        # the observerProxy is async, so it will not block the thread
        log.debug('notification %r: %r' % (name, data))
        self.lastNotification = data
        try:
            _observerProxy.notifyObservers(self,
                                'debugger_session_notify',
                                name)
        except COMException, e:
            log.debug('No debugger notify observers %s', e)

    # stateChange is not an xpcom interface
    def stateChange(self, node):
        # THREAD WARNING
        # this function is called from the command loop thread.  Do
        # not do anything here that will result in another command
        # being sent to the client, that will result in a lockup
        # the observerProxy is async, so it will not block the thread
        if node:
            dbgpServerAPI._dbgpSession.stateChange(self, node)
        log.debug('Notify state change %r for %r' % (self.statusName, self))
        try:
            _observerProxy.notifyObservers(self,
                                'debugger_session_state_change',
                                self.statusName)
        except COMException, e:
            log.debug('No debugger state change observers %s', e)
    
    def resume(self, action):
        result = dbgpServerAPI._dbgpSession.resume(self, action)
        self.stateChange(None)
        return result
        
    def get_maxChildren(self):
        return self.maxChildren

    def get_maxData(self):
        return self.maxData

    def get_maxDepth(self):
        return self.maxDepth
    
    def get_application(self):
        return self._application
    
    def get_address(self):
        return self._clientAddr[0]
    
    def get_port(self):
        return self._clientAddr[1]
    
    def get_supportsDetach(self):
        return self._supportsOptionalCommand('detach')
    
    def get_supportsInteract(self):
        return self._supportsOptionalCommand('interact')

    def spawnpointSet(self, sp):
        # Override subclass for XPCOM Error handling.
        # The Tcl client can, for example, return this:
        #   DBGPError: (u'Spawnpoints not allowed on this line', u'203')
        try:
            return dbgpServerAPI._dbgpSession.spawnpointSet(self, sp)
        except DBGPError, ex:
            errno, errmsg = dbgpServerAPI.getDBGPErrorInfo(ex)
            self._lastErrorSvc.setLastError(errno, errmsg)
            raise ServerException(nsError.NS_ERROR_FAILURE, errmsg)

    def breakpointSet(self, bp):
        # Override subclass for XPCOM Error handling.
        try:
            return dbgpServerAPI._dbgpSession.breakpointSet(self, bp)
        except DBGPError, ex:
            errno, errmsg = dbgpServerAPI.getDBGPErrorInfo(ex)
            self._lastErrorSvc.setLastError(errno, errmsg)
            raise ServerException(nsError.NS_ERROR_FAILURE, errmsg)

dbgpServerAPI.dbgpSession = koDBGPSession

class koDBGPTerminalInterrupt:
    _com_interfaces_ = [components.interfaces.koITerminalInterrupt]
    _reg_clsid_ = "{760fc589-04b9-4183-8517-89dc3f24b01b}"
    _reg_contractid_ = "@activestate.com/koDBGPTerminalInterrupt;1"
    
    def __init__(self, application):
        self.application = application
    
    def puts(self, data):
        return self.write(data)

    def write(self, data):
        if self.application.currentSession.interactivePrompt:
            return self.application.currentSession.interact(data) != 0
        return 0

dbgpServerAPI._dbgpApplication = dbgpServerAPI.dbgpApplication
class koDBGPApplication(dbgpServerAPI._dbgpApplication):
    _com_interfaces_ = [components.interfaces.koIDBGPApplication]
    _reg_clsid_ = "{3FE1F468-BACA-4252-A7F8-A15E826276BD}"
    _reg_contractid_ = "@activestate.com/koDBGPApplication;1"

    def __init__(self, appMgr):
        dbgpServerAPI._dbgpApplication.__init__(self, appMgr)
        self.interactiveHandler = koDBGPTerminalInterrupt(self)
        self.terminal = None
    
    def get_currentSession(self):
        return self.currentSession
    
    def set_currentSession(self, session):
        sess = UnwrapObject(session)
        if not self.haveSession(sess):
            raise Exception('session is not part of this application')
        self.currentSession = sess

dbgpServerAPI.dbgpApplication = koDBGPApplication


dbgpServerAPI._dbgpBreakpointManager = dbgpServerAPI.dbgpBreakpointManager
class koDBGPBreakpointManager(dbgpServerAPI._dbgpBreakpointManager,
                              TreeView):
    _com_interfaces_ = [components.interfaces.koIDBGPBreakpointManager,
                        components.interfaces.nsITreeView]
    _reg_clsid_ = "{3E4A2213-1483-4F6C-A8A7-CFCA4698FE7D}"
    _reg_contractid_ = "@activestate.com/koDBGPBreakpointManager;1"
    
    def __init__(self):
        dbgpServerAPI._dbgpBreakpointManager.__init__(self)

        # nsITreeView setup stuff
        TreeView.__init__(self) # for debugging: , debug="bp")
        self._rowData = {
            # <breakpoint guid>: <row data dict>
        }
        self._order = [] # list of breakpoint guid's in table display order
        self._tree = None
        # Key (usually a <treecol/> id) by which the list (self._order)
        # is sorted.
        self._sortedBy = None
        self._sortReversed = 0 # boolean indicating if currently reverse sorted
        self.tabMgr = None
        
        #XXX Have to keep these constants in sync with value in markers.js.
        self.MARKNUM_SPAWNPOINT_ENABLED = 4
        self.MARKNUM_SPAWNPOINT_DISABLED = 3
        self.MARKNUM_BREAKPOINT_ENABLED = 2
        self.MARKNUM_BREAKPOINT_DISABLED = 1
        self._bpMarkerNums = [
            self.MARKNUM_SPAWNPOINT_ENABLED,
            self.MARKNUM_SPAWNPOINT_DISABLED,
            self.MARKNUM_BREAKPOINT_ENABLED,
            self.MARKNUM_BREAKPOINT_DISABLED,
        ]
        self._bpType2markerNum = {
            # Map a breakpoint type and state to a marker number.
            ('line', 'enabled'): self.MARKNUM_BREAKPOINT_ENABLED,
            ('line', 'disabled'): self.MARKNUM_BREAKPOINT_DISABLED,
            ('call', 'enabled'): self.MARKNUM_BREAKPOINT_ENABLED,
            ('call', 'disabled'): self.MARKNUM_BREAKPOINT_DISABLED,
            ('return', 'enabled'): self.MARKNUM_BREAKPOINT_ENABLED,
            ('return', 'disabled'): self.MARKNUM_BREAKPOINT_DISABLED,
            ('exception', 'enabled'): self.MARKNUM_BREAKPOINT_ENABLED,
            ('exception', 'disabled'): self.MARKNUM_BREAKPOINT_DISABLED,
            ('conditional', 'enabled'): self.MARKNUM_BREAKPOINT_ENABLED,
            ('conditional', 'disabled'): self.MARKNUM_BREAKPOINT_DISABLED,
            ('watch', 'enabled'): self.MARKNUM_BREAKPOINT_ENABLED,
            ('watch', 'disabled'): self.MARKNUM_BREAKPOINT_DISABLED,
            ('spawn', 'enabled'): self.MARKNUM_SPAWNPOINT_ENABLED,
            ('spawn', 'disabled'): self.MARKNUM_SPAWNPOINT_DISABLED,
        }
        self._markerHandles = {
            # <breakpoint guid>: <marker handle from ISciMoz.addMarker()>
        }

    #---- Methods for keeping breakpoint scintilla markers and the
    #     breakpoints in sync.
    def _getMarkerNum(self, bp):
        try:
            return self._bpType2markerNum[(bp.type, bp.state)]
        except KeyError:
            log.warn("Could not determine scintilla marker number of "
                     "breakpoint of type '%s' and state '%s' on "
                     "line %s of file '%s': no restoring marker",
                     bp.type, bp.state, bp.lineno, bp.filename)
            return None

    def _addMarker(self, bp):
        log.debug("koDBGPBreakpointManager._addMarker(%s#%s)",
                  os.path.basename(bp.filename), bp.lineno)
        if bp.filename and bp.lineno > 0:
            views = self.topView.getViewsByTypeAndURIX(1, "editor", bp.filename)
            if views:
                view = views[0] # marker mods on any view are carried to all others
                view = view.QueryInterface(components.interfaces.koIScintillaView)
                markerNum = self._getMarkerNum(bp)
                if markerNum is not None:
                    log.info("add %s/%s breakpoint marker to %s, line %d",
                             bp.type, bp.state, view.document.baseName,
                             bp.lineno)
                    markerHandle = view.scimoz.markerAdd(bp.lineno-1, markerNum)
                    self._markerHandles[bp.getGuid()] = markerHandle

    def _removeMarker(self, bp):
        log.debug("koDBGPBreakpointManager._removeMarker(%s#%s)",
                  os.path.basename(bp.filename), bp.lineno)
        if bp.filename and bp.lineno > 0:
            views = self.topView.getViewsByTypeAndURIX(1, "editor", bp.filename)
            if views:
                view = views[0] # marker mods on any view are carried to all others
                view = view.QueryInterface(components.interfaces.koIScintillaView)
                markerNum = self._getMarkerNum(bp)
                if markerNum is not None:
                    log.info("remove %s/%s breakpoint marker to %s, line %d",
                             bp.type, bp.state, view.document.baseName,
                             bp.lineno)
                    view.scimoz.markerDelete(bp.lineno-1, markerNum)
                    del self._markerHandles[bp.getGuid()]

    def _removeAllMarkers(self):
        log.debug("koDBGPBreakpointManager._removeAllMarkers()")
        views = self.topView.getViewsByTypeX(1, "editor")
        for view in views:
            view = view.QueryInterface(components.interfaces.koIScintillaView)
            for markerNum in self._bpMarkerNums:
                view.scimoz.markerDeleteAll(markerNum)
        self._markerHandles = {}

    def addMarkersForDocument(self, document, scintilla):
        log.debug("koDBGPBreakpointManager.addMarkersForDocument(%r, scintilla)",
                  document.baseName)
        if not document.file: # then this is a new document w/o an associated file
            return
        uri = document.file.URI
        scimoz = scintilla.scimoz
        #XXX This is really inefficient. Should make self._breakpoints more
        #    efficient for this lookup (or at least construct a cached
        #    data structure for this). Should do stress testing to see if
        #    there is a need.
        for bp in self._breakpoints.values():
            #XXX What about case-insensitivity on Windows?
            if bp.filename == uri and bp.lineno > 0:
                markerNum = self._getMarkerNum(bp)
                if markerNum is not None:
                    markerHandle = scimoz.markerAdd(bp.lineno-1, markerNum)
                    self._markerHandles[bp.getGuid()] = markerHandle

    def updateBreakpointsFromMarkers(self, document, scintilla):
        log.debug("koDBGPBreakpointManager.updateBreakpointsFromMarkers(%r, scintilla)",
                  document.baseName)
        if not document.file: # then this is a new document w/o an associated file
            return
        uri = document.file.URI
        scimoz = scintilla.scimoz
        #XXX As above, it might be necessary/helpful to cache the breakpoints
        #    (or their guids) by filename so don't have to iterate through
        #    all breakpoints.
        try:
            for bp in self._breakpoints.values():
                #XXX What about case-insensitivity on Windows?
                if bp.filename == uri and bp.lineno > 0:
                    guid = bp.getGuid()
                    markerHandle = self._markerHandles[guid]
                    lineno = scimoz.markerLineFromHandle(markerHandle)
                    # Make sure we do NOT forward this change to current
                    # debug sessions because they still have the unmodified
                    # file content.
                    #XXX Actually shouldn't the change only not be forwarded
                    #    if that file has already been entered in this debug
                    #    session. Presumably, though, the file has be entered
                    #    already because it is current and the user is
                    #    editing it.
                    if lineno == -1:
                        log.info("breakpoint %s marker removed while editing: "
                                 "removing breakpoint", guid)
                        self._silentlyRemoveBreakpoint(bp)
                        continue
                    lineno += 1 # Breakpoint linenos are 1-based, Scimoz 0-based.
                    if lineno != bp.lineno:
                        #XXX Subtle bug: if a breakpoint moves onto a previous
                        #    breakpoint the markers get out of sync. We _could_
                        #    just delete the breakpoint at that point. In any
                        #    case, this should be rare.
                        log.info("breakpoint %s marker move from line %s to %s: "
                                 "updating breakpoint", guid, bp.lineno, lineno)
                        self._silentlyUpdateBreakpointLineno(bp, lineno)
        except Exception, e:
            log.exception(e, "Could not update breakpoints for '%s'.", uri)
            raise

    def _silentlyRemoveBreakpoint(self, bp):
        self._lock.acquire()
        try:
            guid = bp.getGuid()
            del self._breakpoints[guid]

            del self._rowData[guid] # 1.
            row = self._order.index(guid)
            del self._order[row] # 2.
            # 3. marker already updated, that is part of the point
            if self._tree: # 4.
                self._tree.beginUpdateBatch()
                self._tree.rowCountChanged(row, -1)
                self._tree.invalidateRange(row, len(self._order))
                self._tree.endUpdateBatch()
            if self.tabMgr: # 5.
                self.tabMgr.numBreakpointsChanged()
        finally:
            self._lock.release()

    def _silentlyUpdateBreakpointLineno(self, bp, lineno):
        self._lock.acquire()
        try:
            bp.lineno = lineno
            guid = bp.getGuid()
            self._generateRowData(guid) # 1.
            # 2. don't need to re-sort: editing can't change order
            # 3. marker already updated, that is part of the point
            if self._tree: # 4.
                row = self._order.index(guid)
                self._tree.invalidateRow(row)
            # 5. unnecessary, no change in number of breakpoints
        finally:
            self._lock.release()


    #---- Breakpoint persistence
    def _getBreakpointsFile(self):
        """Return the name of the file in which breakpoints are persisted."""
        koDirs = components.classes["@activestate.com/koDirs;1"].getService()
        return os.path.join(koDirs.hostUserDataDir, "breakpoints.pickle")

    def restoreBreakpoints(self):
        import pickle
        bpFile = self._getBreakpointsFile()
        breakpointList = []
        if os.path.exists(bpFile):
            try:
                fin = open(bpFile, 'r')
                try:
                    breakpointList = pickle.load(fin)
                finally:
                    fin.close()
            except (pickle.PickleError, EnvironmentError, EOFError), ex:
                log.exception("Could not restore breakpoints from '%s'.",
                              bpFile)
            else:
                log.info("restored breakpoints from '%s'", bpFile)

        # We need to assign runtime guids to the breakpoints and create
        # the self._breakpoints mapping properly.
        self._breakpoints = {}
        self._order = []
        for bp in breakpointList:
            guid = self._makeBreakpointGuid()
            bp._guid = guid
            self._breakpoints[guid] = bp
            self._order.append(guid)

        for guid in self._order:
            self._generateRowData(guid)
        self.sort()

    def saveBreakpoints(self):
        import pickle
        bpFile = self._getBreakpointsFile()
        try:
            fout = open(bpFile, 'w')
            try:
                # We just save the breakpoints themselves (and NOT their
                # runtime GUIDs, or the GUID->breakpoint mapping).
                pickle.dump(self._breakpoints.values(), fout, 0) #XXX bin=1
            finally:
                fout.close()
        except (pickle.PickleError, EnvironmentError), ex:
            log.error("Could not save breakpoints to '%s'.", bpFile)
        log.info("saved breakpoints from '%s'", bpFile)


    #---- Changing breakpoints and reacting to breakpoint changes.
    def addBreakpoint(self, bp):
        # Have to unwrap the koIDBGPBreakpoint here, otherwise it does not
        # compare as expected when trying to remove.
        bp = UnwrapObject(bp)
        dbgpServerAPI._dbgpBreakpointManager.addBreakpoint(self, bp)

    def removeAllBreakpointsInURI(self, uri):
        log.debug("koDBGPBreakpointManager.removeAllBreakpointsInURI(uri='%s')",
                  uri)
        for bp in self._breakpoints.values():
            #XXX What about case-insensitive Windows?
            if bp.filename == uri:
                self.removeBreakpoint(bp.getGuid())

    def toggleBreakpointState(self, guid):
        """Toggle the state of the given breakpoint btwn "enabled" and
        "disabled".
        
            "guid" is a dbgpBreakpoint guid number
        """
        bp = self._breakpoints[guid]
        log.debug("koDBGPBreakpointManager.toggleBreakpointState(guid=%s): %s#%s",
                  guid, os.path.basename(bp.filename), bp.lineno)
        newState = None
        if bp.state == "enabled":
            newState = "disabled"
        elif bp.state == "disabled":
            newState = "enabled"
        if newState:
            newBp = bp.clone()
            newBp.state = newState
            self.updateBreakpoint(bp.getGuid(), newBp)

    def toggleAllBreakpointStates(self):
        for bp in self._breakpoints.values():
            if bp.state == "enabled":
                newState = "disabled" # if any bp is enabled, disable all
                break
        else:
            newState = "enabled" # if no bp's are enabled, enable all
        for bp in self._breakpoints.values():
            if bp.state not in ("enabled", "disabled"):
                continue # Don't screw up "deleted" breakpoints.
            if bp.state != newState:
                newBp = bp.clone()
                newBp.state = newState
                #XXX Might be nice to add an optional attrs=None argument
                #    to self.updateBreakpoint().
                self.updateBreakpoint(bp.getGuid(), newBp)

    #---- post* and pre* methods for reacting to changes in the breakpoints
    # Each of these methods must do the following:
    # 1. update self._rowData
    # 2. update self._order
    # 3. update any editor-gutter markers
    # 4. if there is a Breakpoints tree, then update the tree:
    #    .invalidate*(), .rowCountChanged()
    # 5. if the number of breakpoints changed, let the "Breakpoints" tab
    #    manager know
    def postAddBreakpoint(self, bp):
        guid = bp.getGuid()
        self._generateRowData(guid)  # 1.
        self._order.append(guid) # 2.
        self._sort(self._sortedBy)
        row = self._order.index(guid)
        self._addMarker(bp) # 3.
        if self._tree: # 4.
            self._tree.beginUpdateBatch()
            self._tree.rowCountChanged(row, 1)
            self._tree.invalidateRange(row, len(self._order))
            self._tree.endUpdateBatch()
        if self.tabMgr: # 5.
            self.tabMgr.numBreakpointsChanged()

    def postRemoveBreakpoint(self, bp):
        guid = bp.getGuid()
        del self._rowData[guid] # 1.
        row = self._order.index(guid)
        del self._order[row] # 2.
        self._removeMarker(bp) # 3.
        if self._tree: # 4.
            self._tree.beginUpdateBatch()
            self._tree.rowCountChanged(row, -1)
            self._tree.invalidateRange(row, len(self._order))
            self._tree.endUpdateBatch()
        if self.tabMgr: # 5.
            self.tabMgr.numBreakpointsChanged()

    def postRemoveAllBreakpoints(self):
        self._rowData = {} # 1.
        lengthChange = len(self._order)
        self._order = [] # 2.
        self._removeAllMarkers() # 3.
        if self._tree: # 4.
            self._tree.beginUpdateBatch()
            self._tree.rowCountChanged(0, -lengthChange)
            self._tree.invalidate()
            self._tree.endUpdateBatch()
        if self.tabMgr: # 5.
            self.tabMgr.numBreakpointsChanged()

    def preUpdateBreakpoint(self, bp):
        self._removeMarker(bp) # 3. remove marker before update...
    def postUpdateBreakpoint(self, bp, attrs):
        guid = bp.getGuid()
        self._generateRowData(guid) # 1.
        self._sort(self._sortedBy) # 2. re-sort
        row = self._order.index(guid)
        self._addMarker(bp) # 3. ...and re-add marker after
        if self._tree: # 4.
            self._tree.beginUpdateBatch()
            self._tree.invalidateRange(row, len(self._order))
            self._tree.endUpdateBatch()
        # 5. unnecessary, no change in number of breakpoints

    #---- Getting breakpoint info.
    def get_currentBreakpoint(self):
        row = self._tree.selection.currentIndex
        if row == -1:
            raise ServerException(nsError.NS_ERROR_FAILURE,
                                  "no currently selected breakpoint")
        guid = self._order[row]
        return self._breakpoints[guid]

    def get_currentBreakpointGuid(self):
        row = self._tree.selection.currentIndex
        if row == -1:
            raise ServerException(nsError.NS_ERROR_FAILURE,
                                  "no currently selected breakpoint")
        return self._order[row]

    def get_currentBreakpointName(self):
        row = self._tree.selection.currentIndex
        if row == -1:
            raise ServerException(nsError.NS_ERROR_FAILURE,
                                  "no currently selected breakpoint")
        guid = self._order[row]
        return self._rowData[guid]["breakpoints-name"]

    def getBreakpointAtFileAndLine(self, filename, line):
        for bp in self._breakpoints.values():
            if bp.filename == filename and bp.lineno == line and bp.type != "spawn":
                return bp
        else:
            return None
    def getSpawnpointAtFileAndLine(self, filename, line):
        for bp in self._breakpoints.values():
            if bp.type == "spawn" and bp.filename == filename and bp.lineno == line:
                return bp
        else:
            return None

    def copyBreakpointsToNewURI(self, uri_old, uri_new):
        log.debug("koDBGPBreakpointManager.copyBreakpointsToNewURI(old='%s', new='%s')",
                  uri_old, uri_new)
        for bp in self._breakpoints.values():
            #XXX What about case-insensitive Windows?
            if bp.filename == uri_old:
                nbp = bp.clone()
                nbp.filename = uri_new
                self.addBreakpoint(nbp)

    #---- nsITreeView implementation for the "Breakpoints" tab
    #TODO:
    # - add support for multi-column sorting control
    # - add methods to style somethings, e.g. "(no condition)" could be
    #   made grey, likewise "break always"
    def _generateRowData(self, guid):
        """Translate the data for the breakpoint with the given guid for
        display in the "Breakpoints" tree.
        
        Generally we are taking VS.NET's lead here.
        """
        #XXX Currently not showing these attributes: temporary, exceptionName
        bp = self._breakpoints[guid]
        conditionDesc = "(no condition)"
        typeAndState = "breakpoint-%s" % bp.state
        if bp.type == "conditional":
            #XXX is it an error if bp.expression is null or the empty string?
            conditionDesc = "when '%s' is true" % bp.expression
        elif bp.type == "spawn":
            conditionDesc = "N/A"
            typeAndState = "spawnpoint-%s" % bp.state
        if bp.type == "spawn":
            hitCountDesc = "N/A"
        elif bp.hitValue:
            if bp.hitCondition in (None, ">="):
                hitCountDesc = "break when hit count is greater than or equal to %s" % bp.hitValue
            elif bp.hitCondition == "==":
                hitCountDesc = "break when hit count is equal to %s" % bp.hitValue
            elif bp.hitCondition == "%":
                hitCountDesc = "break when hit count is a multiple of %s" % bp.hitValue
            else:
                log.error("unknown hit condition: '%s'" % bp.hitCondition)
                hitCountDesc = "???"
        else:
            hitCountDesc = "break always"
        rowData = {
            "breakpoints-name": bp.getName(),
            "breakpoints-language": bp.language,
            "breakpoints-condition": conditionDesc,
            "breakpoints-hit-count": hitCountDesc,
            "breakpoints-file": uriparse.displayPath(bp.filename),
            "breakpoints-type-and-state": "", # don't want a label, just an image
            "state": bp.state,
            "type": bp.type,
            # Used for sorting image column on both break vs spawn and
            # enabled vs disabled.
            "type-and-state": typeAndState,
            "guid": guid, #XXX May no longer be necessary.
        }
        self._rowData[guid] = rowData
    
    def _sort(self, sortBy):
        """Do the actual sorting of the rows by the given key."""
        def cond(guid1, guid2, sortBy=sortBy, rowData=self._rowData):
            return cmp(rowData[guid1][sortBy], rowData[guid2][sortBy])

        try:
            self._order.sort(cond)
            if self._sortReversed:
                self._order.reverse()
        except KeyError:
            log.error("Cannot sort breakpoints by: '%s'" % sortBy)
            raise

    def sort(self, sortBy="breakpoints-name"):
        """Sort the current rows by the given key.
        
            "sortBy" is the key by which the sort. Usually this is a
                Breakpoints <treecol/> id. If not specified it defaults
                to the "breakpoints-name" key/column.
        
        If already sorted by this key then reverse the sorting order.
        Note that the breakpoints tree supports sorting on multiple keys
        (as does bugzilla for example).
        """
        log.debug("sort breakpoints by '%s'", sortBy)
        if sortBy == "breakpoints-type-and-state":
            sortBy = "type-and-state"
        if self._sortedBy == sortBy:
            self._order.reverse()
            self._sortReversed = not self._sortReversed
        else:
            self._sortReversed = 0  # reset to non-reversed sort for a new key
            self._sort(sortBy)
        self._sortedBy = sortBy
        if self._tree:
            self._tree.beginUpdateBatch()
            self._tree.invalidate()
            self._tree.endUpdateBatch()

    def setTree(self, tree):
        TreeView.setTree(self, tree)
        if self._tree:
            self._tree.beginUpdateBatch()
            self._tree.rowCountChanged(0, len(self._order))
            self._tree.invalidate()
            self._tree.endUpdateBatch()

    def get_rowCount(self):
        return len(self._order)

    def getCellText(self, row, col):
        try:
            guid = self._order[row]
            datum = self._rowData[guid][col]
        except IndexError:
            return ""
        except KeyError:
            log.error("unknown breakpoint column id: '%s'" % col)
            return ""
        if type(datum) not in types.StringTypes:
            datum = str(datum)
        return datum

    def cycleCell(self, row, col):
        if col == "breakpoints-type-and-state":
            # Toggle the breakpoint enabled/disabled state.
            try:
                guid = self._order[row]
                self.toggleBreakpointState(guid)
            except IndexError:
                pass

    def getImageSrc(self, row, col):
        if col != "breakpoints-type-and-state":
            return ""
        try:
            guid = self._order[row]
            type_ = self._rowData[guid]["type"]
            state = self._rowData[guid]["state"]
            if type_ == "spawn" and state == "enabled":
                return "chrome://jaguar/skin/images/spawnpoint_enabled.gif"
            elif type_ == "spawn" and state == "disabled":
                return "chrome://jaguar/skin/images/spawnpoint_disabled.gif"
            if state == "enabled":
                return "chrome://jaguar/skin/images/breakpoint_enabled.gif"
            elif state == "disabled":
                return "chrome://jaguar/skin/images/breakpoint_disabled.gif"
            elif state == "deleted":
                log.warn("did not exist a breakpoint with state='deleted': "
                         "row %d" % row)
                return "chrome://jaguar/skin/global/images/delete.png"
            else:
                log.error("unexpected breakpoint state (row %d): '%s' "
                          "(type '%s')" % (row, state, type_))
                return ""
        except IndexError:
            return ""
    
dbgpServerAPI.dbgpBreakpointManager = koDBGPBreakpointManager

    
dbgpServerAPI._dbgpManager = dbgpServerAPI.dbgpManager
class koDBGPManager(dbgpServerAPI._dbgpManager):
    _com_interfaces_ = [components.interfaces.koIDBGPManager,
                        components.interfaces.nsIObserver]
    _reg_clsid_ = "{EC99C1B5-9E83-4200-BF69-F06D3B72A243}"
    _reg_contractid_ = "@activestate.com/koDBGPManager;1"
    
    def __init__(self):
        dbgpServerAPI._dbgpManager.__init__(self)
        _initProxies()
        global _globalPrefsSvcProxy, _observerProxy
        self._prefs = _globalPrefsSvc.prefs
        self._observerProxy = _observerProxy
        self._lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"].\
            getService(components.interfaces.koILastErrorService)

        # XXX for now this will be done in debugger.js
        # Observers will be QI'd for a weak-reference, so we must keep the
        # observer alive ourself, and must keep the COM object alive,
        # _not_ just the Python instance!!!
        # self._wrapped = WrapObject(self, components.interfaces.nsIObserver)
        # self._prefs.prefObserverService.addObserver(self._wrapped, 'remoteDebuggerListenPort', 1)
        # self._prefs.prefObserverService.addObserver(self._wrapped, 'dbgpProxyEnabled', 1)
        # self._prefs.prefObserverService.addObserver(self._wrapped, 'dbgpProxyKey', 1)
        # self._prefs.prefObserverService.addObserver(self._wrapped, 'dbgpProxyListenAddress', 1)

    def getDBGPBreakpointManager(self):
        bpMgr = components.classes["@activestate.com/koDBGPBreakpointManager;1"]\
                .getService(components.interfaces.koIDBGPBreakpointManager)
        return bpMgr

    def _getKey(self):
        proxyKey = None
        if self._prefs.hasPref("dbgpProxyKey"):
            proxyKey = self._prefs.getStringPref("dbgpProxyKey")
        if not proxyKey:
            proxyKey = os.getenv('USER',os.getenv('USERNAME','NOKEY'))
        return proxyKey

    def getKey(self):
        if self._server_key is None:
            return self._getKey()
        else:
            return self._server_key

    def start(self):
        self._lastErrorSvc.setLastError(0, '')
        if self.isListening():
            return
        # get address/port from prefs, then start the listener
        port = 9000
        proxyEnabled = 0
        if self._prefs.hasPref("remoteDebuggerListenPort"):
            port = self._prefs.getLongPref("remoteDebuggerListenPort")
        if self._prefs.hasPref("dbgpProxyEnabled"):
            proxyEnabled = self._prefs.getBooleanPref("dbgpProxyEnabled")
        if proxyEnabled:
            proxyKey = ''
            proxyAddr = '127.0.0.1'
            proxyPort = 9001
            port = 0 # system selected port
            proxyKey = self._getKey()
            if self._prefs.hasPref("dbgpProxyListenAddress"):
                addr = self._prefs.getStringPref("dbgpProxyListenAddress")
                (proxyAddr, _port,) = string.split(addr,':')
                if _port:
                    proxyPort = int(_port)

            log.debug('setting proxy key to %s', proxyKey)
            self.setKey(proxyKey)
            log.debug('setting proxy address to %s:%d', proxyAddr, proxyPort)
            self.setProxy(proxyAddr, proxyPort)
        else:
            self.setProxy('', 0)
        
        self.listen('', port)

    def listen(self, address, port):
        try:
            dbgpServerAPI._dbgpManager.listen(self, address, port)
        except DBGPError, ex:
            errno, errmsg = dbgpServerAPI.getDBGPErrorInfo(ex)
            self._lastErrorSvc.setLastError(errno, errmsg)
            raise ServerException(nsError.NS_ERROR_FAILURE, errmsg)
        try:
            self._observerProxy.notifyObservers(None, 'debugger_listener', 'on')
        except COMException, e:
            pass

    def stop(self):
        try:
            dbgpServerAPI._dbgpManager.stop(self)
        except DBGPError, ex:
            errno, errmsg = dbgpServerAPI.getDBGPErrorInfo(ex)
            self._lastErrorSvc.setLastError(errno, errmsg)
            # likely the proxy died, but we still stopped listening, notify we're off
        try:
            self._observerProxy.notifyObservers(None, 'debugger_listener', 'off')
        except COMException, e:
            pass

    def isListening(self):
        if self._listener and self._listener.checkListening():
            return 1
        return 0
    
    def onConnect(self, session, client, addr):
        # before any communication, we can decide if we want
        # to allow the connection here.  return 0 to deny
        # XXX TODO add ip filtering capability
        return dbgpServerAPI._dbgpManager.onConnect(self, session, client, addr)
    
    def notifyInit(self, session, init):
        try:
            log.debug('Notify debugger_init for %r' % session)
            self._observerProxy.notifyObservers(session, 'debugger_init', '')
        except COMException, e:
            log.debug('No debugger init observers %s', e)

    def observe(self, prefSet, prefName, prefSetID):
        #print "got notify for %r:%r:%r" %(prefSet, prefName, prefSetID)
        if prefName in ['remoteDebuggerListenPort',
                        'dbgpProxyEnabled',
                        'dbgpProxyKey',
                        'dbgpProxyListenAddress']:
            self.stop()
            self.start()

    def get_address(self):
        if self._listener:
            return self._listener._address
        return ''

    def get_port(self):
        if self._listener:
            return self._listener._port
        return 0

    def get_proxyAddress(self):
        return self._proxyAddr

    def get_proxyPort(self):
        return self._proxyPort

    def get_currentSessions(self):
        return len(self.getApplicationList())

    def get_currentConnections(self):
        c = 0
        apps = self.getApplicationList()
        for app in apps:
            c = c + app.sessionCount()
        return c

    def get_totalConnections(self):
        if self._listener:
            return self._listener._totalConnections
        return 0

dbgpServerAPI.dbgpManager = koDBGPManager

