# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

from xpcom import components
sci_constants = components.interfaces.ISciMoz
from xpcom.server import WrapObject, UnwrapObject
import scimozindent
import logging
import re
import eollib
log = logging.getLogger('koLanguageCommandHandler')
indentlog = logging.getLogger('koLanguageCommandHandler.indenting')
jumplog = logging.getLogger('koLanguageCommandHandler.jump')

import string
import timeline

"""
The generic command handler is appropriate for all languages.

  It handles things like:
    block selection
    brace matching and jumps to matching brace
    transpose
    ctrl-space
    indentation control
    
"""

class GenericCommandHandler:
    _com_interfaces_ = [components.interfaces.koIViewController,
                        components.interfaces.nsIController,
                        components.interfaces.nsIObserver]
    _reg_desc_ = "Command Handler for All files"
    _reg_contractid_ = "@activestate.com/koGenericCommandHandler;1"
    _reg_clsid_ = "{B383592C-5343-4AA7-9419-04D1B34EC906}"
    
    def __init__(self):
        log.info("in __init__ for GenericCommandHandler")
        self._completeWordState = None
        self._view = None
        self.sysUtils = components.classes["@activestate.com/koSysUtils;1"].\
            getService(components.interfaces.koISysUtils)

    def __del__(self):
        log.info("in __del__ for GenericCommandHandler")
        
    def set_view(self, view):
        if view:
            self._view = view.QueryInterface(components.interfaces.koIScintillaView)
            self._prefs = view.prefs
            self._scimoz = self._view.scimoz
        else:
            self._view = None
            self._prefs = None
            self._scimoz = None

    def get_view(self):
        return self._view

    def _is_cmd_fontFixed_enabled(self):
        return 1

    def _do_cmd_fontFixed(self):
        self._view.alternateFaceType = not self._view.alternateFaceType
        self._view.prefs.setBooleanPref('editUseAlternateFaceType',
                                        not self._view.prefs.getBooleanPref('editUseAlternateFaceType'))

    def _is_cmd_viewWhitespace_enabled(self):
        return self._scimoz.viewWS

    def _do_cmd_viewWhitespace(self):
        self._view.scimoz.viewWS = not self._view.scimoz.viewWS

    def _is_cmd_viewLineNumbers_enabled(self):
        return self._scimoz.getMarginWidthN(0) > 0

    def _updateLineNumberMargin(self):
        numLinesToAccountFor = max(1000,self._scimoz.lineCount*2)
        padding = 5
        textWidth = self._scimoz.textWidth(0, str(numLinesToAccountFor))
        self._scimoz.setMarginWidthN(0, textWidth + padding)

    def _do_cmd_viewLineNumbers(self):
        shown = self._scimoz.getMarginWidthN(0) > 0
        shown = not shown
        if shown:
            self._updateLineNumberMargin()
        else:
            self._scimoz.setMarginWidthN(0, 0)

    def _is_cmd_viewIndentationGuides_enabled(self):
        return self._scimoz.indentationGuides

    def _do_cmd_viewIndentationGuides(self):
        self._scimoz.indentationGuides = not self._scimoz.indentationGuides

    def _is_cmd_viewEOL_enabled(self):
        return self._scimoz.viewEOL

    def _do_cmd_viewEOL(self):
        self._scimoz.viewEOL = not self._scimoz.viewEOL

    def _is_cmd_wordWrap_enabled(self):
        return self._scimoz.wrapMode

    def _do_cmd_wordWrap(self):
        if self._scimoz.wrapMode == self._scimoz.SC_WRAP_NONE:
            self._scimoz.wrapMode = self._scimoz.SC_WRAP_WORD
            self._scimoz.layoutCache = self._scimoz.SC_CACHE_PAGE
        else:
            self._scimoz.wrapMode = self._scimoz.SC_WRAP_NONE
            self._scimoz.layoutCache = self._scimoz.SC_CACHE_NONE

    def _is_cmd_jumpToCorrespondingLine_enabled(self):
        return self._view.languageObj.name == 'Diff'

    def _findMatchingLineAndCol(self, position):
        """Return the tuple (filename,line,col) corresponding to where in the
        <filename> file that the specified character <position> in the current
        view (self._view) maps to.
        
        This deals with understanding diff "chunk" headers
        and offsets relative to those.
        
        Ideally this code should work for all diff formats without
        knowledge of what the relevant pref is, since it can be used in
        files 
        """
        scin = self._view.scimoz

        # Look up the buffer from the 'position' until we find the
        # line identifying the file -- for p4, it starts with ==== //
        p4duRE = "@@ -\d+,\d+ \+(\d+)," # -du
        p4defaultRE = "\d+\(,.\d+\)?[acd](\d+)\(,.\d+\)?" # default
        p4dcRE = "*** \d+,\d+ ****"
        jumpRE = re.compile(p4duRE)
        # the fileRE is a regex which tells us what file
        # our chunk is in.
        fileRE = re.compile("==== .*#\d+ - (.*) ====") # XXX p4 specific???

        startLineNo = scin.lineFromPosition(position)
        lineStart = scin.positionFromLine(startLineNo)
        lineEnd = scin.getLineEndPosition(startLineNo)
        line = scin.getTextRange(lineStart, lineEnd)
        startLine = line
        offset = -2 # lines between current line and chunk header

        # If we're in the middle of a deletion block, we'll start
        # 'thinking' from the _end_ of the current deletion block.
        while line.startswith('-'):
            jumplog.info('Skipping line: %r', line)
            startLineNo += 1
            lineStart = scin.positionFromLine(startLineNo)
            lineEnd = scin.getLineEndPosition(startLineNo)
            line = scin.getTextRange(lineStart, lineEnd)
        
        chunkLineStart = None
        # walking backwards from the current line up
        for lineNo in range(startLineNo, -1, -1):
            lineStart = scin.positionFromLine(lineNo)
            lineEnd = scin.getLineEndPosition(lineNo)
            line = scin.getTextRange(lineStart, lineEnd)
            jumplog.info('Looking at line: %r', line)
            if chunkLineStart == None:
                # if we haven't found the chunk header yet
                chunkMatch = jumpRE.search(line)
                if chunkMatch:
                    chunkLineStart = int(chunkMatch.group(1))
                    jumplog.info('found chunk header: chunkLineStart = %d',
                                 chunkLineStart)
                    # found it!.  This tells us where in the
                    # file the current chunk starts. (we don't
                    # yet know which file we're dealing with however.
            else: # we have a chunk tag, look for the filename
                match = fileRE.search(line)
                # do we have a line indicating what file we're working
                # against?
                if match:
                    filename = match.group(1)
                    jumplog.info('found filename: %r', filename)
                    lineNo = chunkLineStart + offset
                     # -1 because of space, max to avoid negative cols.
                    column = max(0, scin.getColumn(position)-1)
                    return filename, lineNo, column
            if line.startswith('- ') or line.startswith('? ') or \
               line.startswith('??'):
                # we need to skip 'deletion', ndiff '? ' lines
                # and chunk headers since they are
                # not in the current version of the file and so
                # do not contribute to the offset computation.
                jumplog.info("Discounting: %r", line)
                continue
            if not chunkLineStart:
                jumplog.info("Adding 1 to offset for %r -> offset = %d",
                             line, offset)
                offset += 1
        return None, None, None
    
    def _do_cmd_jumpToCorrespondingLine(self):
        scin = self._view.scimoz
        position = scin.currentPos
        observerSvc = components.classes["@mozilla.org/observer-service;1"].\
                    getService(components.interfaces.nsIObserverService);
        fname1, anchorLine, anchorCol = self._findMatchingLineAndCol(scin.anchor)
        fname2, currentPosLine, currentPosCol = self._findMatchingLineAndCol(scin.currentPos)
        if fname1 != fname2:
            log.warn("Can't jump to a selection that spans files!.  Picking the first.")
            fname2, currentPosLine, currentPosCol = fname1, anchorLine, anchorCol
        if fname1 is None:
            log.warn("Unable to find jump target")
            return
        data = "%s\t%d,%d-%d,%d" % (fname1,
                                     anchorLine + 1,
                                     anchorCol + 1,
                                     currentPosLine + 1,
                                     currentPosCol + 1)
        observerSvc.notifyObservers(None, 'open_file', data)


    def _do_cmd_transientMarkSet(self):
        self._view.transientMark = self._view.scimoz.currentPos

    def _do_cmd_transientMarkMoveBack(self):
        sm = self._view.scimoz
        mark = self._view.transientMark
        if mark == -1:
            return
        sm.anchor = sm.currentPos = mark
        self._view.rotateTransientMarkRing()
        sm.scrollCaret()

    def _do_cmd_cutRegion(self):
        sm = self._view.scimoz
        mark = self._view.transientMark
        if mark == -1:
            return
        sm.anchor = mark
        sm.cut()
        sm.sendUpdateCommands("clipboard")
        
    def _do_cmd_transientMarkExchangeWithPoint(self):
        mark = self._view.transientMark
        if mark == -1:
            return
        sm = self._view.scimoz
        self._view.transientMark = self._view.scimoz.currentPos
        sm.currentPos = sm.anchor = mark
        sm.scrollCaret()

    def _do_cmd_editReflow(self):
        """ Reflow -- currently only works for paragraphs in text documents
        needs to be tweaked for comments and strings in code.
        """
        
        scin = self._view.scimoz
        import reflow
        # First figure out what paragraphs the current selection spans.
        # Walk back to the first blank line or to the top of the
        # document.
        # first we move the starting 'position of interest' to the beginning
        # of the current line
        startLineNo = scin.lineFromPosition(scin.selectionStart)

        if 0: # I think Trent has convinced me it's a bad idea.
            # we will swallow lines back from our current position if
            # they're not empty
            while startLineNo and getLine(scin, startLineNo-1).strip():
                startLineNo -= 1

        scin.targetStart = scin.positionFromLine(startLineNo)

        if (scin.getColumn(scin.selectionEnd) == 0 and
            scin.selectionEnd != 0 and
            scin.selectionEnd != scin.selectionStart and
            not getLine(scin, scin.lineFromPosition(scin.selectionEnd)).strip()):
            # if our selection ends after the newline chars and the next
            # line is blank, we don't want to skip over that line and go
            # on to the next paragraph.
            #
            # in other words:
            #
            #   adsa<|asdasjdlakjdas
            #   asdasdsa
            #   |>
            #
            # is a case where we don't want to deal w/ line 3 above, but
            #
            #   <|>asdlkhas dkljas kldjalkdsj saldj
            #
            # we obviously want to deal with line 1.
            #
            scin.selectionEnd -= 1
            _steppedBack = True
        else:
            _steppedBack = False

        endLineNo = scin.lineFromPosition(scin.selectionEnd)

        if (scin.selectionEnd != scin.selectionStart and 
            scin.getColumn(scin.selectionEnd) == 0):
            # we don't really want to include that line by default
            endLineNo -= 1

        # let's find out what indents we're going to count as being part of
        # the last relevant paragraph.
        para = reflow.Para(reflow.Line(getLine(scin, endLineNo)))
        
        if scin.selectionEnd == scin.selectionStart:
            # We're going to reflow including lines that are part of the
            # paragraph being considered.  To figure that out, we look for
            # lines which, among other criteria, have the same styling
            # as the current line. The styling is derived from the last
            # _non-whitespace_ character in the line we're ending with.
            curStyle = scin.getStyleAt(scin.getLineEndPosition(endLineNo)-1)
    
            while 1:
                # if we reach end of document, stop going on
                if endLineNo+1 >= scin.lineCount:
                    break
                # if the next line contains styles (not in whitespace
                # characters) which aren't like the current style, then
                # don't include said line in reflowable region
                try:
                    for pos in range(scin.positionFromLine(endLineNo+1),
                                     scin.getLineEndPosition(endLineNo+1)):
                        c = scin.getWCharAt(pos)
                        s = scin.getStyleAt(pos)
                        if c in ' \t': continue
                        if s != curStyle:
                            #print "breaking", c,s
                            raise ValueError
                        #print "ok with", c
                except ValueError:
                    break
                # if the next line doesn't fit in to the current paragraph,
                # then stop.
                if not para.accept(reflow.Line(getLine(scin, endLineNo+1))):
                    break
                para.append(reflow.Line(getLine(scin, endLineNo+1)))
                endLineNo += 1
        scin.targetEnd = scin.getLineEndPosition(endLineNo)
        start = scin.positionFromLine(startLineNo)
        end = min(scin.textLength, scin.getLineEndPosition(endLineNo))
        
        # if we have whitespace to our left we'll want to make sure that we end
        # up with whitespace to our left
        hadWhitespaceToLeft = scin.anchor <= scin.currentPos and scin.getWCharAt(scin.currentPos-1) in ' \t'

        if end <= start:  # no point -- nothing to reflow
            return
        text = scin.getTextRange(start, end)
        
        # we want to find out where (relative to non-whitespace characters) the cursor
        # is _before_ the reflow, so we can try to maintain that after reflow.
        textBeforeCurrentPos = scin.getTextRange(scin.targetStart, scin.currentPos)
        textBeforeCurrentPosNoWS = ''.join([word.strip() for word in textBeforeCurrentPos.split()])
        numChars = len(textBeforeCurrentPosNoWS)

        # we now know what text to reflow.
        # we need the eol characters to use for reflowing
        eol = eollib.eol2eolStr[eollib.scimozEOL2eol[scin.eOLMode]]
        # boom
        # reflow doesn't know how to deal with tabs of non-8-space width
        # so we'll do a conversion here, then convert back after
        text = text.expandtabs(scin.tabWidth)
        reflowed = reflow.reflow(text, scin.edgeColumn, eol)
        if scin.useTabs:
            lines = reflowed.splitlines(1)
            for pos in range(len(lines)):
                line = lines[pos]
                if line:
                    raw, effective = classifyws(line, scin.tabWidth)
                    ntabs, nspaces = divmod(effective, scin.tabWidth)
                    lines[pos] = '\t' * ntabs + ' ' * nspaces + line[raw:]
            reflowed = ''.join(lines)
        orig = scin.getTextRange(scin.targetStart, scin.targetEnd)
        if orig == reflowed:
            return  # do nothing
        # replacing what we've come up with as a selection
        scin.replaceTarget(len(reflowed), reflowed)
        # now we look for where we want to put the cursor
        # we'll move N characters from the start of the range, where
        # N is 'numChars' as determined above.
        numCharsSkipped = 0
        i = 0
        while numCharsSkipped < numChars and i < len(reflowed):
            if reflowed[i] not in string.whitespace:
                numCharsSkipped += 1
            i += 1
        if hadWhitespaceToLeft:
            i += 1
        scin.currentPos = scin.anchor = min(scin.textLength, scin.targetStart + i)
        if _steppedBack:
            # step forward to the next line
            scin.currentPos = scin.anchor = min(scin.textLength,
                                                scin.positionFromLine(scin.lineFromPosition(scin.currentPos)+1))
        scin.xOffset = 0

    def _do_cmd_editCenterVertically(self):
        # center the buffer on the cursor (vertically)
        scimoz = self._scimoz
        fvl = scimoz.firstVisibleLine
        curLineNo = scimoz.lineFromPosition(scimoz.currentPos)
        linesOnScreen = scimoz.linesOnScreen
        scimoz.lineScroll(0, (curLineNo-fvl)-(linesOnScreen//2))

    def _do_cmd_editMoveCurrentLineToTop(self):
        # scroll the buffer so the current line is at the top
        scimoz = self._scimoz
        fvl = scimoz.firstVisibleLine
        curLineNo = scimoz.lineFromPosition(scimoz.currentPos)
        scimoz.lineScroll(0, curLineNo-fvl)

    def isCommandEnabled( self, command_name ):
        # Result: boolean
        # In: param0: wstring
        meth = getattr(self, "_is_%s_enabled" % (str(command_name),), None)
        if meth is None:
            rc = 1
        else:
            rc = meth()
        log.debug("koLanguageCommandHandler - isCommandEnabled '%s' -> %d", command_name, rc)
        return rc

    def supportsCommand( self, command_name ):
        # Result: boolean
        # In: param0: wstring
        rc = hasattr(self, "_do_" + command_name)
        log.debug("koLanguageCommandHandler - SupportsCommand '%s' -> %d", command_name, rc)
        return rc

    def doCommand( self, command_name ):
        # Result: void - None
        # In: param0: wstring
        meth = getattr(self, "_do_" + command_name)
        try:
            # Some CommandHandler's do not have a wrapper, e.g. TerminalCommandHandler
            #if self.wrapper is not None:
            #    self.wrapper.beginBatchOperation()
            #try:
                meth()
            #finally:
                #if self.wrapper is not None:
                #    self.wrapper.endBatchOperation()
        except KeyboardInterrupt: # User cancelled a dialog.
            pass

    def _do_cmd_newlineExtra(self):
        self._do_cmd_newline(None, 1)

    def _do_cmd_newlineBare(self):
        self._do_cmd_newline('none', 0)

    def _do_cmd_newlineSame(self):
        self._do_cmd_newline('plain', 0)

    def _do_cmd_newline(self, indentStyle=None, continueComments=0):
        if indentStyle is None:
            indentStyle = self._prefs.getStringPref('editAutoIndentStyle')
        sm = self._scimoz
        sm.beginUndoAction()
        try:
            try:
                if indentStyle == 'none':
                    indent = ''
                else:
                    indent = self._view.languageObj.computeIndent(self._scimoz,
                                                                  indentStyle,
                                                                  continueComments)
                # Optionally clean up whitespace characters to the left of cursor if there
                # is something else there (leave whitespace-only lines alone, as otherwise
                # typing { <cr><cr> } up-arrow is frustrating.
                lineNo = sm.lineFromPosition(sm.currentPos)
                lineStart = sm.positionFromLine(lineNo)
                stuffToLeft = sm.getTextRange(lineStart, sm.currentPos)
                if stuffToLeft.strip():
                    while (sm.currentPos > lineStart and
                           sm.getWCharAt(sm.positionBefore(sm.currentPos)) in ' \t'):
                        sm.deleteBack()
                    # OPTIONALLY delete whitespace to the right of the 
                    # current cursor position
                    if 1: # XXX make a pref, give a UI.
                        lineEnd = sm.getLineEndPosition(lineNo)
                        textToRight = sm.getTextRange(sm.currentPos, lineEnd)
                        while sm.getWCharAt(sm.currentPos) in ' \t':
                            sm.clear()
                
                sm.newLine()
                if indent:
                    sm.addText(self.sysUtils.byteLength(indent), indent) # consider using replaceTarget??
                sm.chooseCaretX()
            except Exception, e:
                log.exception("Error in cmd_newline handling")
        finally:
            sm.endUndoAction()
        

    def _getIndentWidthForLine(self, lineNo, upto=None):
        indent = self._getIndentForLine(lineNo, upto)
        return len(indent.expandtabs(self._scimoz.tabWidth))

    def _getIndentForLine(self, lineNo, upto=None):
        sm = self._scimoz
        lineStart = sm.positionFromLine(lineNo)
        if upto:
            lineEnd = min(upto, sm.getLineEndPosition(lineNo))
        else:
            lineEnd = sm.getLineEndPosition(lineNo)
        line = sm.getTextRange(lineStart, lineEnd)
        indentLength = len(line)-len(line.lstrip())
        return line[:indentLength]

    def _getIndentWidthFromIndent(self, indentstring):
        """ Given an indentation string composed of tabs and spaces,
        figure out how long it is (get the tabWidth from scimoz) """
        expandedstring = indentstring.expandtabs(self._scimoz.tabWidth)
        return len(expandedstring)

    def _getFoldLevel(self, line):
        """Return a fold level for the current line that guarantees that:
            - if at the top level, the fold level is zero
            - the deeper you are the higher the fold level
        It does *not* guarantee that:
            - a depth of three means then level is three
        """
        
        # _getFoldLevel depends on the document being styled.  It currently
        # is only called from _do_cmd_blockSelect below, which calls this within
        # loops, so the colourise call is done first rather than here.  topFoldLevel
        # is also set before using this function
        
        foldLevel = self._scimoz.getFoldLevel(line)
        foldLevel &= self._scimoz.SC_FOLDLEVELNUMBERMASK
        foldLevel -= self.__topFoldLevel
        return foldLevel

    def _do_cmd_blockSelect(self):
        """Use the lexer's fold markers to increase the selection.
        If the current selection *is* a complete "fold block" then increase
        to the next fold block. If the current selection is *not* a complete
        fold block then select the fold block of the shallowest end of the
        selection.

        The logic is complicated somewhat because, the line preceeding a
        block at a certain level should be included in the block selection.
        For example, in the following code, if "Ctrl-B" is hit with the
        cursor on the "print 1" line, the "for element in mylist:" line
        should be included in the selection.

        LEVEL   CODE
        0       def foo(a):
        1           mylist = ["hello", 42, "there", 3.14159]
        1           while 1:
        2               for element in mylist:
        3                   print 1
        3                   if type(element) == type(""):
        4                       print "element %s is a string" % element
        1           print
        """
        #log.debug("----------------- BLOCK SELECT ---------------------")
        # Ensure cursor is at the start of the current selection (so we
        # don't have to deal with the reverse condition in the calculations
        # below.
        if self._scimoz.anchor < self._scimoz.currentPos:
            self._scimoz.anchor, self._scimoz.currentPos =\
                self._scimoz.currentPos, self._scimoz.anchor

        startLine = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        endLine = self._scimoz.lineFromPosition(self._scimoz.anchor)
        lastLine = self._scimoz.lineFromPosition(self._scimoz.textLength-1)
        
        # if the lexer doesn't support folding, select all and get out of here fast.
        lexer = self._view.languageObj.getLanguageService(
            components.interfaces.koILexerLanguageService)
        if not lexer.supportsFolding:
            self._scimoz.selectAll()
            return

        # since we need all the fold data, ensure the entire document is styled
        if self._scimoz.endStyled < self._scimoz.length:
            self._scimoz.colourise(self._scimoz.endStyled, -1)

        self.__topFoldLevel = self._scimoz.getFoldLevel(0)
        self.__topFoldLevel &= self._scimoz.SC_FOLDLEVELNUMBERMASK

        # If the next line after the startLine is "deeper" then move the
        # startLine to there. See the doc string for why this is done.
        if (startLine != lastLine and
            self._getFoldLevel(startLine+1) > self._getFoldLevel(startLine)):
            startLine += 1

        # If there is a selection and selectionEnd is at the start of a line
        # then move up the end of the previous line because this simplifies
        # the subsequent logic.
        if (self._scimoz.selText != ""
            and self._scimoz.getColumn(self._scimoz.anchor) == 0):
            endLine -= 1
            self._scimoz.anchor = self._scimoz.getLineEndPosition(endLine)
        #log.debug("BLOCKSELECT: startLine=%d, endLine=%d, lastLine=%d",
        #          startLine, endLine, lastLine)

        # Determine the minimum fold level in the current selection.
        minFoldLevel = self._getFoldLevel(startLine)
        for line in range(startLine, endLine):
            #log.debug("BLOCKSELECT: foldLevel for line %d is: %x",
            #          line, self._getFoldLevel(line))
            minFoldLevel = min(minFoldLevel, self._getFoldLevel(line))
        #log.debug("BLOCKSELECT: minFoldLevel is %x", minFoldLevel)

        # Determine if the current selection is a complete fold block.
        # - selection start must be at the start of a line
        # - selection end must be at the end of a line
        # - start and end fold level must both be at the bound of the
        #   minFoldLevel
        if (self._scimoz.getColumn(self._scimoz.currentPos) == 0
            and self._scimoz.anchor == self._scimoz.getLineEndPosition(endLine)
            and (startLine == 0
                 or (self._getFoldLevel(startLine-1) < minFoldLevel
                     and self._getFoldLevel(startLine)
                         > self._getFoldLevel(startLine-1))
                )
            and (endLine == lastLine
                 or (self._getFoldLevel(endLine+1) < minFoldLevel
                     and self._getFoldLevel(endLine)
                         > self._getFoldLevel(endLine+1))
                )
           ):
            #log.debug("BLOCKSELECT: *is* a complete fold block")
            if startLine == 0:
                targetFoldLevel = 0
            else:
                targetFoldLevel = self._getFoldLevel(startLine-1)
        else:
            #log.debug("BLOCKSELECT: condition: start of selection on column 0? %s",
            #          self._scimoz.getColumn(self._scimoz.currentPos) == 0)
            #log.debug("BLOCKSELECT: condition: end of selection at eol? %s",
            #          self._scimoz.anchor == self._scimoz.getLineEndPosition(endLine))
            #log.debug("BLOCKSELECT: condition: startLine at minFoldLevel boundary? %s",
            #          (startLine == 0 or
            #           (self._getFoldLevel(startLine-1) < minFoldLevel
            #            and self._getFoldLevel(startLine)
            #                > self._getFoldLevel(startLine-1))))
            #log.debug("BLOCKSELECT: condition: endLine at minFoldLevel boundary? %s",
            #          (endLine == lastLine or
            #           (self._getFoldLevel(endLine+1) < minFoldLevel
            #            and self._getFoldLevel(endLine)
            #                > self._getFoldLevel(endLine+1))))
            #log.debug("BLOCKSELECT: is *not* a complete fold block")
            targetFoldLevel = minFoldLevel
        #log.debug("BLOCKSELECT: targetFoldLevel is %x", targetFoldLevel)

        # increase the selection to the target fold level
        if targetFoldLevel == 0:
            #log.debug("BLOCKSELECT: select all")
            self._scimoz.anchor = self._scimoz.textLength
            self._scimoz.currentPos = 0
            self._scimoz.ensureVisibleEnforcePolicy(0)
        else:
            # expand upwards
            while startLine > 0:  #XXX are line's in SciMoz 0- or 1-based???
                #log.debug("BLOCKSELECT: level(%x) up one line(%d) > target level(%x)?",
                #          self._getFoldLevel(startLine-1), startLine-1,
                #          targetFoldLevel)
                if self._getFoldLevel(startLine-1) >= targetFoldLevel:
                    startLine -= 1
                else:
                    break 
            # select the leading line in the block as well
            if startLine != 0:
                startLine -= 1

            # expand downwards
            while endLine < lastLine:
                #log.debug("BLOCKSELECT: level(%x) down one line(%d) > target level(%x)?",
                #          self._getFoldLevel(endLine+1), endLine+1,
                #          targetFoldLevel)
                if self._getFoldLevel(endLine+1) >= targetFoldLevel:
                    endLine += 1
                else:
                    break 

            # make the selection
            #log.debug("BLOCKSELECT: startLine=%d, endLine=%d",
            #          startLine, endLine)
            if endLine == lastLine:
                self._scimoz.anchor = self._scimoz.textLength
            else:
                self._scimoz.anchor = self._scimoz.positionFromLine(endLine+1)
            self._scimoz.currentPos = self._scimoz.positionFromLine(startLine)
            self._scimoz.ensureVisibleEnforcePolicy(startLine)

    def _is_cmd_comment_enabled(self):
        commenter = self._view.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        return commenter is not None

    def _do_cmd_comment(self):
        commenter = self._view.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        if not commenter:
            # not all languages have commentor services, and the commands
            # are not always updated to improve performance.  Log it as
            # an error
            log.error("%s does not have a commenter service!",
                      self._view.languageObj.name)
            return
        commenter.comment(self._scimoz)

    def _is_cmd_uncomment_enabled(self):
        commenter = self._view.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        return commenter is not None

    def _do_cmd_uncomment(self):
        commenter = self._view.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        if not commenter:
            # not all languages have commentor services, and the commands
            # are not always updated to improve performance.  Log it as
            # an error
            log.error("%s does not have a commenter service!",
                      self._view.languageObj.name)
            return
        commenter.uncomment(self._scimoz)

    def _do_cmd_convertUpperCase(self):
        selStart = self._scimoz.selectionStart
        selEnd = self._scimoz.selectionEnd
        self._scimoz.replaceSel(self._scimoz.selText.upper())
        self._scimoz.selectionStart = selStart
        self._scimoz.selectionEnd = selEnd

    def _do_cmd_convertLowerCase(self):
        selStart = self._scimoz.selectionStart
        selEnd = self._scimoz.selectionEnd
        self._scimoz.replaceSel(self._scimoz.selText.lower())
        self._scimoz.selectionStart = selStart
        self._scimoz.selectionEnd = selEnd

    def _is_cmd_selectToMatchingBrace_enabled(self):
        matchingBrace = self._scimoz.braceMatch(self._scimoz.positionBefore(self._scimoz.currentPos))
        log.info("matchingBrace @ %d : %d", self._scimoz.currentPos, matchingBrace)
        if matchingBrace != -1: return 1 # we're there
        # also check to the right
        if self._scimoz.currentPos < self._scimoz.textLength:
            matchingBrace = self._scimoz.braceMatch(self._scimoz.currentPos)
            log.info("matchingBrace @ %d : %d", self._scimoz.currentPos, matchingBrace)
            return matchingBrace != -1
        else:
            log.info("matchingBrace: returning 0") 
            return 0

    _is_cmd_jumpToMatchingBrace_enabled = _is_cmd_selectToMatchingBrace_enabled
    _is_cmd_braceMatch_enabled = _is_cmd_selectToMatchingBrace_enabled

    def _do_cmd_braceMatch(self):
        if not self._view.languageObj.supportsBraceHighlighting:
            return
        braceAtCaret, braceOpposite, isInside = self._findMatchingBracePosition(self._scimoz.currentPos)
        
        if braceAtCaret != -1 and braceOpposite == -1:
            self._scimoz.braceBadLight(braceAtCaret)
            self._scimoz.highlightGuide = 0
        else:
            chBrace = self._scimoz.getWCharAt(braceAtCaret)
            self._scimoz.braceHighlight(braceAtCaret, braceOpposite)
            columnAtCaret = self._scimoz.getColumn(braceAtCaret)
            if chBrace == ':':
                lineStart = self._scimoz.lineFromPosition(braceAtCaret)
                indentPos = self._scimoz.getLineIndentPosition(lineStart)
                indentPosNext = self._scimoz.getLineIndentPosition(lineStart+1)
                columnAtCaret = self._scimoz.getColumn(indentPos)
                columnAtCaretNext = self._scimoz.getColumn(indentPosNext)
                indentSize = self._scimoz.indent
                if (columnAtCaretNext - indentSize > 1):
                    columnAtCaret = columnAtCaretNext - indentSize 
            columnOpposite = self._scimoz.getColumn(braceOpposite)
            self._scimoz.highlightGuide = min(columnAtCaret, columnOpposite)

    def _do_cmd_selectToMatchingBrace(self):
        self._goMatchingBrace(1)

    def _do_cmd_jumpToMatchingBrace(self):
        self._goMatchingBrace(0)

    def _goMatchingBrace(self, select):
        braceAtCaret, braceOpposite, isInside = self._findMatchingBracePosition(self._scimoz.currentPos)
        log.info("braceAtCaret: %d, braceOpposite: %d, isInside: %s", braceAtCaret, braceOpposite, isInside)
        # Convert the character positions into caret positions based on whether
        # the caret position was inside or outside the braces.
        if isInside:
            if braceOpposite > braceAtCaret:
                braceAtCaret += 1
            else:
                braceOpposite += 1
        else: # Outside
            if braceOpposite > braceAtCaret:
                braceOpposite += 1
            else:
                braceAtCaret += 1
        if braceOpposite >= 0:
            self._ensureRangeVisible(braceOpposite, braceOpposite)
            if select:
                self._scimoz.anchor = braceAtCaret
                self._scimoz.currentPos = braceOpposite
            else:
                self._scimoz.anchor = braceOpposite
                self._scimoz.currentPos = braceOpposite
        self._scimoz.chooseCaretX()

    def _ensureRangeVisible(self, posStart, posEnd, enforcePolicy=1):
        lineStart = self._scimoz.lineFromPosition(min(posStart, posEnd))
        lineEnd = self._scimoz.lineFromPosition(max(posStart, posEnd))
        for line in range(lineStart, lineEnd+1):
            if enforcePolicy:
                self._scimoz.ensureVisibleEnforcePolicy(line)
            else:
                self._scimoz.ensureVisible(line)

    # Find the brace that matches the brace before or after the given caret
    # position. The character before has precedence if it is a brace.
    # Returns three values: whether the caret is inside the braces,
    # and the indexes of the current and the matching brace.
    def _findMatchingBracePosition(self, caretPos, sloppy=1):
        mask = self._view.languageObj.stylingBitsMask
        isInside = 0
        braceAtCaret = -1
        braceOpposite = -1
        charBefore = '\0'
        styleBefore = 0
        textLength = self._scimoz.textLength
        if caretPos > textLength:
            caretPos = textLength
        if (caretPos > 0) :
            charBefore = self._scimoz.getWCharAt(self._scimoz.positionBefore(caretPos))
            styleBefore = self._scimoz.getStyleAt(self._scimoz.positionBefore(caretPos)) & mask
        # Priority goes to character before caret
        if charBefore and self._view.languageObj.getBraceIndentStyle(charBefore, styleBefore):
            braceAtCaret = caretPos - 1
        
        colonMode = 0
        if self._view.languageObj.name == 'Python' and ':' == charBefore:
            braceAtCaret = caretPos - 1
            colonMode = 1

        isAfter = 1
        if sloppy and (braceAtCaret < 0) and \
           0 <= caretPos < textLength: # XXX check last edge condition
            # No brace found so check other side
            charAfter = self._scimoz.getWCharAt(caretPos)
            styleAfter = self._scimoz.getStyleAt(caretPos) & mask
            
            if charAfter and (charAfter in "[](){}") and \
               self._view.languageObj.getBraceIndentStyle(charAfter, styleAfter):
                braceAtCaret = caretPos
                isAfter = 0

            if self._view.languageObj.name == 'Python' and ':' == charAfter:
                braceAtCaret = caretPos
                colonMode = 1
                
        if braceAtCaret >= 0:
            if colonMode:
                lineStart = self._scimoz.lineFromPosition(braceAtCaret)
                lineMaxSubord = self._scimoz.getLastChild(lineStart, -1)
                braceOpposite = self._scimoz.getLineEndPosition(lineMaxSubord)
            else:
                braceOpposite = self._scimoz.braceMatch(braceAtCaret)
                
        if braceOpposite > braceAtCaret:
            isInside = isAfter
        else:
            isInside = not isAfter
        return braceAtCaret, braceOpposite, isInside

    def _is_cmd_folding_enabled(self):
        return self._view.languageObj.foldable

    _is_cmd_foldExpand_enabled = _is_cmd_folding_enabled
    _is_cmd_foldExpandAll_enabled = _is_cmd_folding_enabled
    _is_cmd_foldCollapse_enabled = _is_cmd_folding_enabled
    _is_cmd_foldCollapseAll_enabled = _is_cmd_folding_enabled
    
    def _do_cmd_foldExpand(self):
        lineno = self._scimoz.lineFromPosition(self._scimoz.currentPos);
        if (self._scimoz.getFoldLevel(lineno) & self._scimoz.SC_FOLDLEVELHEADERFLAG) and \
            not self._scimoz.getFoldExpanded(lineno):
            self._scimoz.toggleFold(lineno)

    def _do_cmd_foldExpandAll(self):
        for lineno in range(0, self._scimoz.lineCount):
            if (self._scimoz.getFoldLevel(lineno) & self._scimoz.SC_FOLDLEVELHEADERFLAG) and \
                not self._scimoz.getFoldExpanded(lineno):
                    self._scimoz.toggleFold(lineno)

    def _do_cmd_foldCollapse(self):
        lineno = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        while not (self._scimoz.getFoldLevel(lineno) & self._scimoz.SC_FOLDLEVELHEADERFLAG):
            if lineno == 1 or not self._scimoz.getFoldExpanded(lineno):
                return
            lineno -= 1
        self._scimoz.toggleFold(lineno)
        self._scimoz.gotoLine(lineno)

    def _do_cmd_foldCollapseAll(self):
        if self._scimoz.endStyled < self._scimoz.length:
            self._scimoz.colourise(self._scimoz.endStyled,-1)
        for lineno in range(0, self._scimoz.lineCount):
            if (self._scimoz.getFoldLevel(lineno) & self._scimoz.SC_FOLDLEVELHEADERFLAG) and \
                self._scimoz.getFoldExpanded(lineno):
                self._scimoz.toggleFold(lineno)


    def _asktabwidth(self):
        dialogproxy = components.classes['@activestate.com/asDialogProxy;1'].\
            getService(components.interfaces.asIDialogProxy)
        tabwidth = self._view.prefs.getLongPref('tabWidth')
        value = dialogproxy.prompt("Tab Width (between 0 and 16)", str(tabwidth), "OK", None)
        if value is not None:
            return int(value)
        else:
            return value
                                  

    def _do_cmd_tabify(self):
        selection = self._scimoz.selText
        tabwidth = self._asktabwidth()
        if tabwidth is None: return
        lines = selection.splitlines(1)
        for pos in range(len(lines)):
            line = lines[pos]
            if line:
                raw, effective = classifyws(line, tabwidth)
                ntabs, nspaces = divmod(effective, tabwidth)
                lines[pos] = '\t' * ntabs + ' ' * nspaces + line[raw:]
        self._scimoz.replaceSel(''.join(lines))

    def _do_cmd_untabify(self):
        selection = self._scimoz.selText
        lines = selection.splitlines(1)
        tabwidth = self._asktabwidth()
        if tabwidth is None: return
        for pos in range(len(lines)):
            lines[pos] = lines[pos].expandtabs(tabwidth)
        self._scimoz.replaceSel(''.join(lines))

    def _do_cmd_backSmart(self):
        # if the pref is off, just do a regular delete
        if (self._scimoz.currentPos != self._scimoz.anchor or
            not self._view.prefs.getBooleanPref('editBackspaceDedents')):
            self._scimoz.deleteBack()
            return
        # If there is only whitespace to the left of us (on the current line), we delete enough characters
        #    to do a proper alignment ( this is currently done by IDLE, and works well w.r.t. tabs, spaces, etc.
        # Else we just do a regular backspace.
        lineNo = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        lineStart = self._scimoz.positionFromLine(lineNo)
        before = self._scimoz.getTextRange(lineStart, self._scimoz.currentPos)
        if not before or before.strip() != '':
            # we're either at the beginning of the line or
            # somewhere in the line with text to our left -- just do a backspace.
            self._scimoz.deleteBack()
            return
        # the following is taken from idle (with variable renamings)
        # here we've got only whitespace to our left.
        # Delete whitespace left, until hitting a real char or closest
        # preceding virtual tab stop.
        # Ick.  It may require *inserting* spaces if we back up over a
        # tab character!  This is written to be clear, not fast.
        tabwidth = self._scimoz.tabWidth
        indentwidth = self._scimoz.indent
        have = len(before.expandtabs(tabwidth))
        assert have > 0
        want = int((have - 1) / indentwidth) * indentwidth
        ncharsdeleted = 0
        while 1:
            before = before[:-1]
            ncharsdeleted = ncharsdeleted + 1
            have = len(before.expandtabs(tabwidth))
            if have <= want or before[-1] not in " \t":
                break
        self._scimoz.beginUndoAction()
        try:
            for i in range(ncharsdeleted):
                self._scimoz.deleteBack()
            if have < want:
                inserted = ' ' * (want - have)
                self._scimoz.replaceSel(inserted)
        finally:
            self._scimoz.endUndoAction()

    def _insertIndent(self):
        sm = self._scimoz
        currentLineNo = sm.lineFromPosition(sm.currentPos)
        lineStart = sm.positionFromLine(currentLineNo)
        toLeft = sm.getTextRange(lineStart, sm.currentPos)
        if not toLeft.strip(): # we've got nothing but whitespace to our left:
            currentIndentWidth = self._getIndentWidthForLine(currentLineNo,
                                                            sm.currentPos)
            numIndents, extras = divmod(currentIndentWidth, self._scimoz.indent)
            numIndents += 1
            newIndentWidth = numIndents * self._scimoz.indent
            newIndent = scimozindent.makeIndentFromWidth(self._scimoz, newIndentWidth)
            sm.anchor = sm.positionFromLine(currentLineNo)
            sm.replaceSel(newIndent)
        else:
            # this isn't quite what vs.net does -- vs.net converts
            # all of the whitespace around the cursor to tabs
            # if appropriate
            startCol = sm.getColumn(sm.currentPos)
            numIndents, extras = divmod(startCol, self._scimoz.indent)
            numIndents += 1
            targetCol = numIndents * self._scimoz.indent
            if sm.useTabs:
                sm.replaceSel('\t')
            else:
                sm.replaceSel(' ' * (targetCol - startCol))

    def _insertDedent(self):
        sm = self._scimoz
        startCol = sm.getColumn(sm.currentPos)
        if self._scimoz.indent == 0:
            log.error("scimoz indent was 0, should never happen")
            return
        numIndents, extras = divmod(startCol, self._scimoz.indent)
        if numIndents and not extras:
            numIndents -= 1
        targetCol = numIndents * self._scimoz.indent
        sm.beginUndoAction()
        try:
            while (sm.getColumn(sm.currentPos) > targetCol and
                   sm.getWCharAt(sm.positionBefore(sm.currentPos)) in ' \t'):
                sm.deleteBack()
            curCol = sm.getColumn(sm.currentPos)
            if curCol < targetCol:
                d = targetCol-curCol
                sm.addText(d, d*' ')
        finally:
            sm.endUndoAction()

    def _do_cmd_dedent(self):
        sm = self._scimoz

        # Do tab autocompletion, assuming it's in progress.
        if sm.autoCActive():
            sm.autoCComplete()
            sm.scrollCaret()
            return

        if sm.currentPos == sm.anchor and \
           self._prefs.getBooleanPref('editTabCompletes') and \
           sm.getWCharAt(sm.positionBefore(sm.currentPos)) in self.wordchars:
            self._doCompleteWord(1)
            return

        selectionStartLine = sm.lineFromPosition(sm.selectionStart)
        endOfSelectionStartLine = sm.getLineEndPosition(selectionStartLine)
        selectionEndLine = sm.lineFromPosition(sm.selectionEnd)
        startColumn = sm.getColumn(sm.selectionStart)

        # Do we have a selection?  If no, then it's 'insert a backwards tab'
        if sm.currentPos == sm.anchor:
            if startColumn == 0:
                sm.currentPos = endOfSelectionStartLine
                self._regionShift(-1)
                sm.currentPos = sm.anchor
            else:
                self._insertDedent()
            return

        # If we have a selection, we first figure out if it's within-line or
        # either whole-line or multi-line.
        
        if (selectionStartLine != selectionEndLine or
            startColumn == 0 and sm.selectionEnd == endOfSelectionStartLine):
            return self._regionShift(-1)
        else:
            return self._insertDedent()

    wordchars = string.letters + string.digits + "_"
    
    def _do_cmd_completeWord(self):
        self._doCompleteWord(0)

    def _do_cmd_completeWordBack(self):
        self._doCompleteWord(1)

    def _doCompleteWord(self, backwards):
        # adapted from IDLE's AutoExpand code, with the additional
        # tweak of 'going backwards'
        curinsert = self._scimoz.currentPos
        lineNo = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        startofLinePos = self._scimoz.positionFromLine(lineNo)
        endofLinePos = self._scimoz.getLineEndPosition(lineNo)
        curline = self._scimoz.getTextRange(startofLinePos, endofLinePos)
        if not self._completeWordState:
            words = self._getwords()
            index = 0
        else:
            words, index, insert, line = self._completeWordState
            if insert != curinsert or line != curline:
                words = self._getwords()
                index = 0
        if not words:
            return
        word = self._getprevword()
        self._scimoz.anchor = self._scimoz.currentPos - len(word)
        self._scimoz.replaceSel('')
        if backwards:
            newword = words[(index-2)  % len(words)]
            index = (index - 1) % len(words)
        else:
            newword = words[index]
            index = (index + 1) % len(words)
        self._scimoz.replaceSel(newword)
        curinsert = self._scimoz.currentPos
        lineNo = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        startofLinePos = self._scimoz.positionFromLine(lineNo)
        endofLinePos = self._scimoz.getLineEndPosition(lineNo)
        curline = self._scimoz.getTextRange(startofLinePos, endofLinePos)
        
        self._completeWordState = words, index, curinsert, curline
        self._scimoz.scrollCaret()

    def _getwords(self):
        # this is where we would do magic to look through other buffers
        word = self._getprevword()
        if not word:
            return []
        before = self._scimoz.getTextRange(0, self._scimoz.currentPos)
        wbefore = re.findall(r"\b" + word + r"\w+\b", before)
        del before
        after = self._scimoz.getTextRange(self._scimoz.currentPos, self._scimoz.textLength)
        wafter = re.findall(r"\b" + word + r"\w+\b", after)
        del after
        if not wbefore and not wafter:
            return []
        words = []
        dict = {}
        # search backwards through words before
        wbefore.reverse()
        for w in wbefore:
            if dict.get(w):
                continue
            words.append(w)
            dict[w] = w
        # search onwards through words after
        for w in wafter:
            if dict.get(w):
                continue
            words.append(w)
            dict[w] = w
        words.append(word)
        return words

    def _getprevword(self):
        curinsert = self._scimoz.currentPos
        lineno = self._scimoz.lineFromPosition(self._scimoz.currentPos)
        startofLinePos = self._scimoz.positionFromLine(lineno)
        line = self._scimoz.getTextRange(startofLinePos, curinsert)
        i = len(line)
        while i > 0 and line[i-1] in self.wordchars:
            i = i-1
        return line[i:]

    def _do_cmd_indent(self):
        sm = self._scimoz

        # Do tab autocompletion, assuming it's in progress.
        if sm.autoCActive():
            shouldAdjust = 0
            # if we're closing a </ tag, then adjust it.
            # Closing a </ tag is defined as "in XML, going backwards,
            # we hit a < before a whitespace character, and
            # there is a / to the right of that <.
            if self._view.languageObj.supportsSmartIndent == 'XML':
                lineno = self._scimoz.lineFromPosition(sm.currentPos)
                startofLinePos = sm.positionFromLine(lineno)
                line = sm.getTextRange(startofLinePos, sm.currentPos)
                lastWS = max(line.rfind(' '), line.rfind('\t'))
                lastBracket = line.rfind('<')
                if lastWS < lastBracket and \
                   lastBracket+1 < len(line) and \
                    line[lastBracket+1] == '/':
                    shouldAdjust = 1
            sm.autoCComplete()
            if shouldAdjust:
                scimozindent.adjustClosingXMLTag(sm)
            sm.scrollCaret()
            return

        if sm.currentPos == sm.anchor:
            if self._prefs.getBooleanPref('editTabCompletes') and \
               sm.getWCharAt(sm.currentPos-1) in self.wordchars:
                self._doCompleteWord(0)
                return
    
            # Do we have a selection?  If no, then it's 'insert a tab'
            self._insertIndent()
            return
        
        # If we have a selection, we first figure out if it's within-line or
        # either whole-line or multi-line.
        
        selectionStartLine = sm.lineFromPosition(sm.selectionStart)
        selectionEndLine = sm.lineFromPosition(sm.selectionEnd)
        startColumn = sm.getColumn(sm.selectionStart)
        endOfSelectionStartLine = sm.getLineEndPosition(selectionStartLine)
        if (selectionStartLine != selectionEndLine or
            startColumn == 0 and sm.selectionEnd == endOfSelectionStartLine):
            return self._regionShift(1)
        else:
            return self._insertIndent()

    def _regionShift(self, shift):
        """ This code shifts regions of text left or right depending on the
        sign of 'shift' (which should be +1 for a rightward shift or
        -1 for a leftward shift
        
        First we adjust the selection to consist only of whole lines.
        
        Then we look for the delta in number of spaces that should be applied
        to the first line which has a non-zero indent.
        
        Then we apply that delta to each line in the selection, ensuring that
        the indentations that result respect the use of tabs as specified
        in the widget
        """
        sm = self._scimoz
        indentlog.info("doing _regionShift by shift: %s", shift)
        # first adjust the selection to span full lines.
        anchorLine = sm.lineFromPosition(sm.anchor)
        currentPosLine = sm.lineFromPosition(sm.currentPos)
        if sm.anchor < sm.currentPos:
            anchorFirst = 1
            startLine = anchorLine
            endLine = currentPosLine
        else:
            anchorFirst = 0
            endLine = anchorLine
            startLine = currentPosLine
            # ensure that anchor starts for the sake of our computation
            # things get fixed at the end.
            sm.anchor, sm.currentPos = sm.currentPos, sm.anchor

        anchorColumn = sm.getColumn(sm.anchor)
        currentPosColumn = sm.getColumn(sm.currentPos)
        if anchorColumn != 0:
            l = sm.lineFromPosition(sm.anchor)
            sm.anchor = sm.positionFromLine(l)
        if currentPosColumn != 0:
            # adjust to end of line
            l = sm.lineFromPosition(sm.currentPos)
            sm.currentPos = sm.getLineEndPosition(l)
        # figure out what delta we need to apply to the first line
        for line in range(startLine, max(endLine, startLine+1)):
            currentIndentWidth = self._getIndentWidthForLine(line)
            if shift == 1 or currentIndentWidth:
                break
        if currentIndentWidth == 0 and shift == -1:
            return # nothing to do, nothing's indented!
        numIndents, extras = divmod(currentIndentWidth, self._scimoz.indent)
        if shift == 1:
            numIndents += 1
        elif shift == -1 and numIndents and not extras:
            numIndents -= 1
        newIndentWidth = numIndents * self._scimoz.indent
        delta = newIndentWidth - currentIndentWidth
        indentlog.info("delta = %d", delta)
        # apply that delta to each line in the region
        region = sm.getTextRange(sm.anchor, sm.currentPos)
        lines = region.splitlines(1) # keep line ends
        data = []
        for line in lines:
            count = 0
            for char in line:
                if char in ' \t':
                    count += 1
                else:
                    break
            indent = line[:count].expandtabs(self._scimoz.tabWidth)
            rest = line[count:]
            numspaces = max(len(indent) + delta, 0)
            indent = scimozindent.makeIndentFromWidth(self._scimoz, numspaces)
            newline = indent + rest
            data.append(newline)
        region = ''.join(data)
        before = min(sm.currentPos, sm.anchor)
        sm.targetStart = sm.anchor
        sm.targetEnd = sm.currentPos
        sm.replaceTarget(len(region), region)
        if anchorFirst:
            sm.anchor = before
            sm.currentPos = before + len(region)
        else:
            sm.currentPos = before
            sm.anchor = before + len(region)

    def unexpandtabs(self, line):
        numspaces = len(line)-len(line.lstrip())
        rest = line[numspaces:]
        numtabs, numspaces = divmod(numspaces, self._scimoz.tabWidth)
        indent = '\t'*numtabs + ' '*numspaces
        return indent + rest


def classifyws(s, tabwidth):
    raw = effective = 0
    for ch in s:
        if ch == ' ':
            raw = raw + 1
            effective = effective + 1
        elif ch == '\t':
            raw = raw + 1
            effective = (effective / tabwidth + 1) * tabwidth
        else:
            break
    return raw, effective

def getLine(scin, lineNo):
    lineStart = scin.positionFromLine(lineNo)
    lineEnd = scin.getLineEndPosition(lineNo)
    line = scin.getTextRange(lineStart, lineEnd)
    return line