#!python
# Copyright (c) 2004 ActiveState Corp, a division of Sophos Plc.
# See the file LICENSE.txt for licensing information.

"""Code Intelligence: database handling

Usage:
    XXX how to use the CIDB
"""

import os, sys
import threading
import time
import xml.sax
from xml.sax.handler import ContentHandler, ErrorHandler
import md5
import pprint

import logging
import sqlite
import process

from codeintel.common import *
from codeintel import ScanRequest



#---- globals

log = logging.getLogger("codeintel.database")
#log.setLevel(logging.INFO)



#---- internal support

def _version2versionInfo(version):
    verInfo = []
    for part in version.split('.'):
        try:
            part = int(part)
        except ValueError:
            pass
        verInfo.append(part)
    return tuple(verInfo)


def _detectXMLEncoding(content):
    "Return the encoding of the given XML content or None if it cannot tell."
    if not content.startswith("<?xml"):
        return None
    import re
    firstline = re.split(r"\r\n|\r|\n", content, 1)[0]
    match = re.search('encoding="([^"]*)"', firstline)
    if not match:
        return None
    encoding = match.group(1)
    return encoding


class _Cix2CidbHandler(ContentHandler):
    """SAX Event Content Handler to update a CIDB database with CIX."""
    MIN_CIX_VERSION = (0,1)

    def __init__(self, cu, manager):
        ContentHandler.__init__(self)
        self.cu = cu                # an open CIDB cursor
        self.manager = manager      # a codeintel.Manager instance
        # State variables
        self.tagstack = []          # stack of CIX tags
        self.scope_id_stack = None  # stack of symbol scope ids
        self.scope_stack = None     # stack of symbol scope names
        self.file_id = None         # CIDB row id for the current file
        self.module_id = None       # CIDB row id for the current module
        # The CIDB row id of the last added symbol.
        # This is used to know to which symbol to attach <doc> content.
        self.last_symbol_id = None
        # used to accumulate character content in a <doc>
        self.doc = None
        # The (<table>, <id>) of the item (a 'symbol' entry or a 'base'
        # entry) with which <type/> tag information can be associated.
        self.last_typeable_row = (None, None)
        
        # Remember data about the "last successful <file/>" for error
        # recovery.
        self.lsf_path = None
        self.lsf_generator = None
        self.lsf_language = None

    def startElement(self, tag, attrs):
        #if log.isEnabledFor(logging.DEBUG):
        #    attrDict = {}
        #    for key, value in attrs.items():
        #        attrDict[key] = value
        #    log.debug("cix: startElement %s %s", tag, attrDict)
        self.tagstack.append(tag)
        funcName = "start_"+tag
        func = getattr(self, funcName, None)
        if func is not None:
            func(attrs)
    def endElement(self, tag):
        #log.debug("cix: endElement %s", tag)
        funcName = "end_"+tag
        func = getattr(self, funcName, None)
        if func is not None:
            func()
        self.tagstack.pop()

    def start_codeintel(self, attrs):
        # validate the CIX version
        if not attrs.has_key("version"):
            raise CIXError("no 'version' tag on 'codeintel' element")
        version = attrs.get("version")
        versionInfo = _version2versionInfo(version)
        if versionInfo < self.MIN_CIX_VERSION:
            raise CIXError("CIX content version, %s, is unsupported" % version)
        #log.debug("cix: version %s", version)

    def start_file(self, attrs):
        """<file .../>
        
        Attributes:
            "path" (required) path of the file scanned. This path should
                not be made absolute if the given path was relative. The
                path must use Unix-style dir separators, '/', so that
                CIX can be cross-platform.
            "generator" (required) is the name of the language engine. This
                must match the name of the language engine driver (as
                defined in engines.py) exactly.
            "langauge" (required) is the language of the scanned content.
            "mtime" (required for successful scan) is a last modified time of
                the file in seconds since the "epoch".
            "md5" (required for successful scan) is the MD5 hexdigest of the
                scanned content.
            "error" (required for failed scan) is a string describing why the
                scan failed. This is intended to be an internal-only form of
                CIX.
        """
        # Must be child of <codeintel/> tag.
        if self.tagstack[-2] != "codeintel":
            raise CIXError("<file> tag must be an immediate child of a "
                           "<codeintel> tag")
        try:
            path = attrs["path"]
        except KeyError, ex:
            raise CIXError("<file> tag has no 'path' attribute")
        cpath = canonicalizePath(path)
        try:
            generator = attrs["generator"]
        except KeyError, ex:
            raise CIXError("<file> tag has no 'generator' attribute")
        try:
            language = attrs["language"]
        except KeyError, ex:
            raise CIXError("<file> tag has no 'language' attribute")
        if attrs.has_key("mtime"):
            mtime = attrs["mtime"]
        else:
            mtime = None
        if attrs.has_key("md5"):
            md5sum = attrs["md5"]
        else:
            md5sum = None
        if attrs.has_key("error"):
            error = attrs["error"]
        else:
            error = None
        if not ((mtime and md5sum and error is None) or
                (mtime is None and md5sum is None and error is not None)):
            raise CIXError("illegal <file> tag: must define 'mtime' and "
                           "'md5' for a successful scan or 'error' for a "
                           "failed scan")

        # Save these for error recovery.
        self.lsf_path = path
        self.lsf_generator = generator
        self.lsf_language = language
        
        #log.debug("cix: update files and scans: (%r, %r, %r, %r, %r, %r)",
        #          cpath, generator, language, mtime, md5sum, error)
        cu = self.cu
        cu.execute("SELECT id FROM language WHERE name=%s", language)
        if cu.rowcount:
            language_id = int(cu.fetchone()[0])
        else: # there is no row for this language, add one
            cu.execute("INSERT INTO language (name) VALUES (%s)", language)
            language_id = int(cu.lastrowid)
        self.language_id = language_id
        
        cu.execute("SELECT id FROM file WHERE compare_path=%s "
                   "AND language_id=%s", cpath, language_id)
        if cu.rowcount:
            file_id = int(cu.fetchone()[0])
        else:
            # there is no row for this file, add a new one
            cu.execute("INSERT INTO file (path, compare_path, language_id) "
                       "VALUES (%s, %s, %s)", canonicalizePath(path, False),
                       cpath, language_id)
            file_id = int(cu.lastrowid)
        self.file_id = file_id
            
        cu.execute("SELECT id FROM scan WHERE file_id=%s AND generator=%s",
                   file_id, generator)
        if cu.rowcount:
            # update this scan row
            scan_id = int(cu.fetchone()[0])
            if md5sum:
                cu.execute("UPDATE scan SET mtime=%s, md5=%s, error=%s "
                           "WHERE id=%s", mtime, md5sum, error, scan_id)
            else:
                # Intentionally do NOT null out 'md5' value.
                cu.execute("UPDATE scan SET error=%s WHERE id=%s",
                           error, scan_id)
        else:
            # there is no scan entry for the file/generator combo, add one
            cu.execute("INSERT INTO scan (file_id, generator, mtime, md5, error) "
                       "VALUES (%s, %s, %s, %s, %s)",
                       file_id, generator, mtime, md5sum, error)
            scan_id = int(cu.lastrowid)
        self.scan_id = scan_id

        self._old_imports = self._getFileImportData(cu, scan_id)
        # If the scan succeeded then, clear all previous scan data for this
        # file and generator in preparation for adding the data from
        # this CIX document.
        if md5sum:
            #starttime = time.clock()
            cu.execute("DELETE FROM module WHERE scan_id=%s", scan_id)
            cu.execute("DELETE FROM symbol WHERE scan_id=%s", scan_id)
            cu.execute("DELETE FROM base WHERE scan_id=%s", scan_id)
            cu.execute("DELETE FROM import WHERE scan_id=%s", scan_id)
            cu.execute("DELETE FROM type WHERE scan_id=%s", scan_id)
            #endtime = time.clock()
            #print "XXX DELETEing old scan took %.3f sec" % (endtime-starttime)
    def end_file(self):
        # Schedule changed import statements for scanning.
        #XXX Eventually should make this depend on Komodo's
        #    codeintel_completion_languages pref.
        if self.lsf_language in ("Python",): #XXX limit to langs using CI-based completion
            new = self._getFileImportData(self.cu, self.scan_id)
            old = self._old_imports
            changed = [i for i in new if i not in old]
            if changed:
                handler = self.manager.getImportHandler(self.lsf_language)
                cwd = os.path.dirname(os.path.abspath(self.lsf_path))
                for (module, symbol) in changed:
                    try:
                        mpath, kind, scannable = handler.findModuleOnDisk(
                            module, symbol, cwd=cwd)
                    except ImportError:
                        pass
                    else:
                        if scannable:
                            log.info("request scan for import '%s'", mpath)
                            r = ScanRequest(mpath, self.lsf_language,
                                            PRIORITY_BACKGROUND)
                            self.manager.addRequest(r)
        self.file_id = None
        self.scan_id = None

    def _getFileImportData(self, cu, scan_id):
        """Return a unique list of import data for the given scan.
        Here "unique" means only the import row data relevant to identifying
        the file imported.
        """
        cu.execute("SELECT module, symbol FROM import WHERE scan_id=%d",
                   scan_id)
        unique = {}
        for row in cu.fetchall():
            unique[tuple(row)] = 1
        #return unique.keys()
        return unique

    # Handle possible first child <doc/> tag of any module or symbol tag
    # (e.g. <module>, <class>).
    def characters(self, content):
        if self.tagstack[-1] == "doc":
            #log.debug("cix: doc characters %r", content)
            # Note: I am specifically NOT unencoding the lower-ASCII char
            #       entities, for example: '&#8;' is not being changed
            #       to '\b'.
            if self.doc:
                self.doc += content
            else:
                self.doc = content
    def end_doc(self):
        if self.doc:
            if self.tagstack[-2] == "module":
                table = "module"
                id_ = self.module_id
            else:
                table = "symbol"
                id_ = self.last_symbol_id
            doc = self.doc
            self.cu.execute("UPDATE %s SET doc=%s WHERE id=%d",
                            table, doc, id_)
        self.doc = None

    def start_module(self, attrs):
        """<module .../>
        
        Attributes:
            "name" (required)
            "line" (optional, default 1) is a 1-based line number on which
                the module definition begins
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this module.
        """
        if self.tagstack[-2] != "file":
            raise CIXError("<module> tag must be an immediate child of a "
                           "<file> tag")
        file_id = self.file_id
        scan_id = self.scan_id
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<module> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <module> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = 1
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        doc = None
        self.cu.execute("""INSERT INTO module
                            (file_id, scan_id, name, line, attributes, doc)
                           VALUES (%(file_id)d, %(scan_id)d, %(name)s,
                                   %(line)d, %(attributes)s, %(doc)s)""",
                        locals())
        self.module_id = self.cu.lastrowid
        self.scope_id_stack = [None]
        self.scope_stack = []
    def end_module(self):
        self.scope_id_stack = None
        self.scope_stack = None
        self.module_id = None

    def start_function(self, attrs):
        """<function .../>
        
        Attributes:
            "name" (required)
            "line" (optional) is a 1-based line number on which the
                definition begins. This is only really optional to support
                CIX data for binary files where a line number doesn't
                make sense. For scanned files a line should really be
                specified.
            "lineend" (optional, though strongly recommended) is a 1-based
                line number for the last line on which the definition occurs.
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this function -- for example,
                "virtual public".
            "signature" (optional, string)
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        scope_id = self.scope_id_stack[-1] or 0
        scope = '.'.join(self.scope_stack)
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<function> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <function> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = None
        if attrs.has_key("lineend"):
            try:
                lineend = attrs["lineend"]
            except ValueError, ex:
                raise CIXError("illegal <function> 'lineend' attribute: must "
                               "be integer: %r" % attrs["lineend"])
        else:
            lineend = None
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        if attrs.has_key("signature"):
            signature = attrs["signature"]
        else:
            signature = None
        type_ = ST_FUNCTION
        doc = None
        self.cu.execute("""INSERT INTO symbol
                            (file_id, scan_id, module_id, scope_id, scope,
                             name, signature, line, lineend, type, attributes,
                             doc)
                           VALUES (%(file_id)d, %(scan_id)d, %(module_id)s,
                                   %(scope_id)s, %(scope)s, %(name)s,
                                   %(signature)s, %(line)s, %(lineend)s,
                                   %(type_)d, %(attributes)s, %(doc)s)""",
                        locals())
        self.last_symbol_id = self.cu.lastrowid
        self.scope_id_stack.append(self.last_symbol_id)
        self.scope_stack.append(name)
    def end_function(self):
        self.scope_id_stack.pop()
        self.scope_stack.pop()

    def start_class(self, attrs):
        """<class .../>
        
        Attributes:
            "name" (required)
            "line" (optional) is a 1-based line number on which the
                definition begins. This is only really optional to support
                CIX data for binary files where a line number doesn't
                make sense. For scanned files a line should really be
                specified.
            "lineend" (optional, though strongly recommended) is a 1-based
                line number for the last line on which the definition occurs.
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this class.
            "signature" (optional, string)
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        scope_id = self.scope_id_stack[-1] or 0
        scope = '.'.join(self.scope_stack)
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<class> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <class> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = None
        if attrs.has_key("lineend"):
            try:
                lineend = attrs["lineend"]
            except ValueError, ex:
                raise CIXError("illegal <class> 'lineend' attribute: must "
                               "be integer: %r" % attrs["lineend"])
        else:
            lineend = None
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        if attrs.has_key("signature"):
            signature = attrs["signature"]
        else:
            signature = None
        type_ = ST_CLASS
        doc = None
        self.cu.execute("""INSERT INTO symbol
                            (file_id, scan_id, module_id, scope_id, scope,
                             name, signature, line, lineend, type, attributes,
                             doc)
                           VALUES (%(file_id)d, %(scan_id)d, %(module_id)s,
                                   %(scope_id)s, %(scope)s, %(name)s,
                                   %(signature)s, %(line)s, %(lineend)s,
                                   %(type_)d, %(attributes)s, %(doc)s)""",
                        locals())
        self.last_symbol_id = self.cu.lastrowid
        self.scope_id_stack.append(self.last_symbol_id)
        self.scope_stack.append(name)
    def end_class(self):
        self.scope_id_stack.pop()
        self.scope_stack.pop()

    def start_interface(self, attrs):
        """<interface .../>
        
        Attributes:
            "name" (required)
            "line" (optional) is a 1-based line number on which the
                definition begins. This is only really optional to support
                CIX data for binary files where a line number doesn't
                make sense. For scanned files a line should really be
                specified.
            "lineend" (optional, though strongly recommended) is a 1-based
                line number for the last line on which the definition occurs.
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this class.
            "signature" (optional, string)
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        scope_id = self.scope_id_stack[-1] or 0
        scope = '.'.join(self.scope_stack)
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<interface> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <interface> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = None
        if attrs.has_key("lineend"):
            try:
                lineend = attrs["lineend"]
            except ValueError, ex:
                raise CIXError("illegal <interface> 'lineend' attribute: must "
                               "be integer: %r" % attrs["lineend"])
        else:
            lineend = None
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        if attrs.has_key("signature"):
            signature = attrs["signature"]
        else:
            signature = None
        type_ = ST_INTERFACE
        doc = None
        self.cu.execute("""INSERT INTO symbol
                            (file_id, scan_id, module_id, scope_id, scope,
                             name, signature, line, lineend, type, attributes,
                             doc)
                           VALUES (%(file_id)d, %(scan_id)d, %(module_id)s,
                                   %(scope_id)s, %(scope)s, %(name)s,
                                   %(signature)s, %(line)s, %(lineend)s,
                                   %(type_)d, %(attributes)s, %(doc)s)""",
                        locals())
        self.last_symbol_id = self.cu.lastrowid
        self.scope_id_stack.append(self.last_symbol_id)
        self.scope_stack.append(name)
    def end_interface(self):
        self.scope_id_stack.pop()
        self.scope_stack.pop()

    def start_variable(self, attrs):
        """<variable .../>
        
        Attributes:
            "name" (required)
            "line" (optional) is a 1-based line number on which the
                definition begins. This is only really optional to support
                CIX data for binary files where a line number doesn't
                make sense. For scanned files a line should really be
                specified.
            "lineend" (optional, though strongly recommended) is a 1-based
                line number for the last line on which the definition occurs.
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this variable -- for example,
                "virtual public".
            "signature" (optional, string)
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        scope_id = self.scope_id_stack[-1] or 0
        scope = '.'.join(self.scope_stack)
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<variable> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <variable> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = None
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        if attrs.has_key("signature"):
            signature = attrs["signature"]
        else:
            signature = None
        type_ = ST_VARIABLE
        doc = None
        self.cu.execute("""INSERT INTO symbol
                            (file_id, scan_id, module_id, scope_id, scope,
                             name, signature, line, type, attributes, doc)
                           VALUES (%(file_id)d, %(scan_id)d, %(module_id)s,
                                   %(scope_id)s, %(scope)s, %(name)s,
                                   %(signature)s, %(line)s, %(type_)d,
                                   %(attributes)s, %(doc)s)""",
                        locals())
        self.last_symbol_id = self.cu.lastrowid
        self.scope_id_stack.append(self.last_symbol_id)
        self.scope_stack.append(name)
        self.last_typeable_row = ("symbol", self.last_symbol_id)
    def end_variable(self):
        self.scope_id_stack.pop()
        self.scope_stack.pop()
        self.last_typeable_row = (None, None)

    def start_argument(self, attrs):
        """<argument .../> - a function argument
        
        Attributes:
            "name" (required)
            "line" (optional) is a 1-based line number on which the argument
                definition begins
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this argument -- for example,
                "virtual public".
            "signature" (optional, string)
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        scope_id = self.scope_id_stack[-1] or 0
        scope = '.'.join(self.scope_stack)
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<argument> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <argument> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = None
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        if attrs.has_key("signature"):
            signature = attrs["signature"]
        else:
            signature = None
        type_ = ST_ARGUMENT
        doc = None
        self.cu.execute("""INSERT INTO symbol
                            (file_id, scan_id, module_id, scope_id, scope,
                             name, signature, line, type, attributes, doc)
                           VALUES (%(file_id)d, %(scan_id)d, %(module_id)s,
                                   %(scope_id)s, %(scope)s, %(name)s,
                                   %(signature)s, %(line)s, %(type_)d,
                                   %(attributes)s, %(doc)s)""",
                        locals())
        self.last_symbol_id = self.cu.lastrowid
        self.scope_id_stack.append(self.last_symbol_id)
        self.scope_stack.append(name)
        self.last_typeable_row = ("symbol", self.last_symbol_id)
    def end_argument(self):
        self.scope_id_stack.pop()
        self.scope_stack.pop()
        self.last_typeable_row = (None, None)

    def start_classref(self, attrs):
        """<classref .../>, i.e. a <class> base
        
        Attributes:
            "name" (required)
        """
        symbol_id = self.scope_id_stack[-1]
        scan_id = self.scan_id
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<classref> tag has no 'name' attribute")
        type_ = BT_CLASSREF
        self.cu.execute("INSERT INTO base (symbol_id, scan_id, name, type) "
                        "VALUES (%(symbol_id)d, %(scan_id)d, %(name)s, %(type_)s)",
                        locals())
        self.last_typeable_row = ("base", self.cu.lastrowid)
    def end_classref(self):
        self.last_typeable_row = (None, None)

    def start_interfaceref(self, attrs):
        """<interfaceref .../>, i.e. a <class> interface or <interface> base
        
        Attributes:
            "name" (required)
        """
        symbol_id = self.scope_id_stack[-1]
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<interfaceref> tag has no 'name' attribute")
        type_ = BT_INTERFACEREF
        self.cu.execute("INSERT INTO base (symbol_id, name, type) "
                        "VALUES (%(symbol_id)d, %(name)s, %(type_)s)",
                        locals())
        self.last_typeable_row = ("base", self.cu.lastrowid)
    def end_interfaceref(self):
        self.last_typeable_row = (None, None)

    def start_type(self, attrs):
        """<type .../>, i.e. a possible type for a <variable/>, <argument/>
        <classref/> or <interfaceref/>.
        
        Attributes:
            "type" (required) is a CITDL expression
            "score" (optional) is a numeric value that _may_ be used to
                indicate higher probabilities of some type guesses over
                others.
        XXX Could add an optional line number that could be useful for
            showing how the type determination was made.
        """
        if self.lsf_language not in ("Python",):
            # OPTIMIZATION: Don't bother collecting type information for
            # languages that don't use CI-based completion (the only use
            # for the type info).
            return
        
        scan_id = self.scan_id
        table, target_id = self.last_typeable_row
        if table is None or target_id is None:
            raise CIXError("<type> tag has no appropriate target symbol or base")
        if table == "symbol":
            symbol_id = target_id
            base_id = None
        elif table == "base":
            symbol_id = None
            base_id = target_id
        else:
            raise CIXError("unexpected target table for <type> tag: '%s'" % table)
        try:
            citdl = attrs["type"]
        except KeyError, ex:
            raise CIXError("<type> tag has no 'type' attribute")
        if attrs.has_key("score"):
            try:
                score = int(attrs["score"])
            except ValueError, ex:
                raise CIXError("optional 'score' attribute of <type> tag "
                               "must be an integer: %r" % attrs["score"])
        else:
            score = None
        self.cu.execute("""INSERT INTO type
                            (scan_id, symbol_id, base_id, citdl, score)
                           VALUES (%(scan_id)d, %(symbol_id)s, %(base_id)s,
                                   %(citdl)s, %(score)s)""",
                        locals())

    def start_import(self, attrs):
        """<import .../>
        
        Attributes:
            "module" (required) is the name of the module (or file) being
                imported.
            "symbol" (optional) is the name of a symbol being imported from
                'module', if one is specified. E.g.:
                    Python: from time import asctime
                    Perl:   use File::Path qw(mkpath);
            "alias" (optional) is an alias for 'name' if any, or for 'module',
                if not. E.g.
                    Python:     import xml as XML
                    Python:     from time import time as clock
            "line" (optional) is a 1-based line number on which the import
                definition begins
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        # We can't use None/NULL as the scope_id to indicate "module-level
        # scope" because you can't SELECT on that, at least not with SQLite.
        scope_id = self.scope_id_stack[-1] or 0
        try:
            module = attrs["module"]
        except KeyError, ex:
            raise CIXError("<import> tag has no 'module' attribute")
        try:
            line = int(attrs["line"])
        except KeyError, ex:
            line = None
        except ValueError, ex:
            raise CIXError("illegal <import> 'line' attribute: must "
                           "be an integer: %r" % attrs["line"])
        if attrs.has_key("symbol"):
            symbol = attrs["symbol"]
        else:
            symbol = None
        if attrs.has_key("alias"):
            alias = attrs["alias"]
        else:
            alias = None
        self.cu.execute("""INSERT INTO import
                            (file_id, scan_id, module_id, scope_id, line,
                             module, symbol, alias)
                           VALUES (%(file_id)d, %(scan_id)d, %(module_id)d,
                                   %(scope_id)d, %(line)s, %(module)s,
                                   %(symbol)s, %(alias)s)""",
                        locals())



#---- public classes

class Database:
    """A friendlier interface to a Code Intelligence database (CIDB).
    
    Usage:
        db = Database(manager, cx)
        print db.getVersion()
    
    Note that you may only use a Database from one thread because a CI
    database connection cannot be shared by multiple threads.
    """
    CIDB_VERSION = (0, 14)

    def __init__(self, manager, cx):
        """Create a Database instance.
        
            "manager" is the controlling codeintel.Manager instance.
            "cx" is an open connection to the CIDB.
        """
        self.manager = manager
        self.cx = cx

    def create(self):
        """Setup the database for first use."""
        # Create the tables.
        fname = os.path.join(os.path.dirname(__file__), "cidb_schema.sql")
        createTablesSql = open(fname, "r").read()
        cu = self.cx.cursor()
        cu.execute(createTablesSql)
        
        # Setup some "meta" table information.
        ver = '.'.join(map(str, self.CIDB_VERSION))
        cu.execute("INSERT INTO meta(name, value) "
                   "VALUES ('version', %s)", ver)
        import socket
        cu.execute("INSERT INTO meta(name, value) "
                   "VALUES ('hostname', %s)", socket.gethostname())
        cu.execute("INSERT INTO meta(name, value) "
                   "VALUES ('platform', %s)", getPlatform())

        self.cx.commit()

    def update(self, cix, request=None, driver=None, recover=1):
        """Update the database with the given CIX-format results.
        
            "cix" is the XML content of the scan results for the given
                request. The XML must be in the std Code Intelligence XML
                format.
            "request" (optional) is the ScanRequest instance that resulted in
                the given CIX.
            "driver" (optional) is the LanguageEngineDriver instance that
                was used to generate the given CIX.
            "recover" is an internal boolean for recursive calls indicating
                if error recovery should be attempted. It is only attempted
                on the first time through

        Can raise a CodeIntelError if there is a problem.
        """
        # xml.sax expects an encoded string. Ensure that that's what
        # 'cix' is.
        if isinstance(cix, unicode):
            encoding = _detectXMLEncoding(cix)
            if not encoding:
                encoding = "utf-8" # fallback to trying UTF-8
            try:
                cix = cix.encode(encoding)
            except UnicodeError, ex:
                raise CIXError("error updating database with CIX: could not "
                               "encoding the unicode CIX content: "
                               "encoding='%s'" % encoding)

        cu = self.cx.cursor()
        try:
            ch = _Cix2CidbHandler(cu, self.manager)
            try:
                xml.sax.parseString(cix, ch)
            except (CIXError, xml.sax.SAXException, sqlite.Error,
                    UnicodeError), ex:
                self.cx.rollback() # Drop any changes that *were* made.
                if not recover:
                    raise
                # Try to recover.
                if ch.lsf_path: # means parsing got as far as a <file/>
                    fattrs = {"path": ch.lsf_path,
                              "generator": ch.lsf_generator,
                              "language": ch.lsf_language,
                              "error": str(ex)}
                    cix2 = '<?xml version="1.0" encoding="UTF-8"?>'\
                           '<codeintel version="0.1"> <file%s/> </codeintel>'\
                           % xmlattrstr(fattrs)
                    return self.update(cix2, request, driver, recover=0)
                elif request and driver:
                    fattrs = {"path": request.path,
                              "generator": driver.name,
                              "language": request.language,
                              "error": str(ex)}
                    cix2 = '<?xml version="1.0" encoding="UTF-8"?>'\
                           '<codeintel version="0.1"> <file%s/> </codeintel>'\
                           % xmlattrstr(fattrs)
                    return self.update(cix2, request, driver, recover=0)
                else:
                    raise
            self.cx.commit()
        finally:
            cu.close()

    def removeFile(self, path):
        cu = self.cx.cursor()
        try:
            cpath = canonicalizePath(path)
            cu.execute("SELECT id FROM file WHERE compare_path=%s", cpath)
            if not cu.rowcount:
                log.warn("could not remove file '%s': not in CIDB", path)
                return
            file_id = int(cu.fetchone()[0])
            cu.execute("SELECT id FROM scan WHERE file_id=%s", file_id)
            for (scan_id,) in cu.fetchall():
                cu.execute("DELETE FROM module WHERE scan_id=%s", scan_id)
                cu.execute("DELETE FROM symbol WHERE scan_id=%s", scan_id)
                cu.execute("DELETE FROM base WHERE scan_id=%s", scan_id)
                cu.execute("DELETE FROM import WHERE scan_id=%s", scan_id)
                cu.execute("DELETE FROM type WHERE scan_id=%s", scan_id)
                cu.execute("DELETE FROM scan WHERE id=%s", scan_id)
            cu.execute("DELETE FROM file WHERE id=%s", file_id)
            self.cx.commit()
        finally:
            cu.close()

    def getVersion(self):
        cu = self.cx.cursor()
        cu.execute("SELECT name, value FROM meta WHERE name='version'")
        return cu.fetchone()[1]

    def getVersionInfo(self):
        version = self.getVersion()
        return _version2versionInfo(version)

    def getHostname(self):
        cu = self.cx.cursor()
        cu.execute("SELECT name, value FROM meta WHERE name='hostname'")
        return cu.fetchone()[1]

    def getPlatform(self):
        cu = self.cx.cursor()
        cu.execute("SELECT name, value FROM meta WHERE name='platform'")
        return cu.fetchone()[1]


