#!python
# Copyright (c) 2004 Sophos Inc.
# See the file LICENSE.txt for licensing information.

"""Code Intelligence: CILE drivers and registry"""

import os
import sys

import logging
import process

from codeintel.common import *


#---- globals

log = logging.getLogger("codeintel.engines")



#---- public classes

class LanguageEngineDriver:
    """Virtual base class for Language Engine Drivers
    
    A Language Engine Driver knows how to use a facility (be it a Python
    module, a separate app, a service) for scanning content of a specific
    language for code intelligence data; then update that database
    with that data.  The "facility" is generally known as a "Language Engine".
    
    Subclasses should override the .scan() method.
    """
    name = "Virtual Base" # subclasses should override this

    def scan(self, request):
        """Scan the given file and return data as a CIX document.

            "request" is a ScanRequest instance.

        This method MUST be re-entrant. The scheduler typically runs a pool
        of scans simultaneously so individual Language Engine Drivers can
        be called into from multiple threads.
        
        Returns a CIX document (XML) if the scan was successful.  Raises a
        LanguageEngineError if there was a problem scanning.  I.e. a
        driver should be resistant to Language Engine hangs and crashes.
        """
        raise NotImplementedError("virtual method: you must override this in a subclass")


class _SuperDaveEngineDriver(LanguageEngineDriver):
    """A crashy Language Engine Driver for Python files.
    
    As the name implies the SuperDave language engine crashes (and
    sometimes hangs) a lot. This is useful for showing an example of how
    to write a Language Engine Driver that is resistant to a language
    engine that can crash or hang.
    """
    name = "SuperDave"

    def __init__(self):
        # Find the 'superdavele' executable to use as the Language Engine.
        dname = os.path.normpath(os.path.join(os.path.dirname(__file__), os.pardir))
        if sys.platform.startswith("win"):
            fname = "superdavele.exe"
        else:
            fname = "superdavele"
        self.superdaveleExe = os.path.join(dname, fname)
        if not os.path.exists(self.superdaveleExe):
            raise CodeIntelError(
                "Could not find SuperDave Language Engine: '%s'\n"
                "You may need to run the Makefile.<platform> in the "
                "src/codeintel directory to build this test executable."
                % self.superdaveleExe)

    def scan(self, request):
        # Randomly either work, crash, or hang.
        import random, bisect
        argvs = [
            (10, [self.superdaveleExe, "crash"]),   # crash 10% of the time
            (20, [self.superdaveleExe, "hang"]),    # hang 10% of the time
            (30, [self.superdaveleExe, "HANG"]),    # hang w/100%CPU 10% of the time
            (100, [self.superdaveleExe, request.path]),  # work the remainder
        ]
        percent = random.randint(1, 100)
        index = bisect.bisect(argvs, (percent, []))
        argv = argvs[index][1]
        #print "XXX _SuperDaveEngineDriver.scan(): argv=%s (percent=%d, index=%d)" % (argv, percent, index)

        # Run language engine and report any errors.
        p = process.ProcessOpen(argv)
        if sys.platform.startswith("win"):
            try:
                retval = p.wait(1.0) # Give it about a second to run.
            except process.ProcessError, ex:
                p.kill()
                raise LanguageEngineError("engine timed-out or crashed")
            stdout = p.stdout.read()
            stderr = p.stderr.read()
            p.close()
        else:
            raise NotImplementedError("XXX")
            #>>> try:
            #...     p.wait(os.WNOHANG)  # poll to see if is process still running
            #... except ProcessError, ex:
            #...     if ex.errno == ProcessProxy.WAIT_TIMEOUT:
            #...             print "process is still running"

        return stdout


class PythonEngineDriver(LanguageEngineDriver):
    name = "Python"
    def __init__(self):
        from codeintel import pythoncile
        self.scanner = pythoncile.scan
    def scan(self, request):
        #XXX Muck around with thread priority on Windows.  Unfortunately
        #    this doesn't seem to have any useful effect.  Leaving this
        #    in for reference if we come back to this.
        #import win32api, win32process
        #hThread = win32api.GetCurrentThread()
        #win32process.SetThreadPriority(hThread, win32process.THREAD_PRIORITY_HIGHEST)
        import parser
        import pythoncile
        try:
            return self.scanner(request.content, request.path,
                                request.md5sum, request.mtime)
        except (pythoncile.PythonCILEError, parser.ParserError), ex:
            raise LanguageEngineError(ex)


class PerlEngineDriver(LanguageEngineDriver):
    name = "Perl"

    def __init__(self):
        # Find the 'perlcile' executable to use as the Language Engine.
        dname = os.path.normpath(os.path.dirname(__file__))
        if sys.platform.startswith("win"):
            self.perlcile = os.path.join(dname, "perlcile.exe")
            self.scineplex = os.path.join(dname, "scineplex.exe")
        else:
            self.perlcile = os.path.join(dname, "perlcile")
            self.scineplex = os.path.join(dname, "scineplex")
        if not os.path.exists(self.perlcile):
            raise CodeIntelError("could not find the Perl CILE "
                                 "component '%s'\n" % self.perlcile)
        if not os.path.exists(self.scineplex):
            raise CodeIntelError("could not find the Perl CILE "
                                 "component '%s'\n" % self.scineplex)

    def scan(self, request):
        request.calculateMD5()
        argv = [self.perlcile,
                "--scineplex", self.scineplex,
                "--filename", request.path,
                "--mtime", str(request.mtime),
                "--md5", request.md5sum]

        # Run language engine and report any errors.
        p = process.ProcessOpen(argv)
        content = re.sub("(\r\n|\r)", "\n", request.content)
        p.stdin.write(content)
        p.stdin.close()
        stdout = p.stdout.read()
        stderr = p.stderr.read()
        p.close()
        return stdout


class TclEngineDriver(LanguageEngineDriver):
    name = "Tcl"

    def __init__(self):
        # Find the 'tclcile' executable to use as the Language Engine.
        dname = os.path.normpath(os.path.dirname(__file__))
        if sys.platform.startswith("win"):
            self.tclcile = os.path.join(dname, "tclcile.exe")
        else:
            self.tclcile = os.path.join(dname, "tclcile")
        if not os.path.exists(self.tclcile):
            raise CodeIntelError("could not find the Tcl CILE "
                                 "component '%s'\n" % self.tclcile)

    def scan(self, request):
        request.calculateMD5()
        argv = [self.tclcile,
                "--filename", request.path,
                "--mtime", str(request.mtime),
                "--md5", request.md5sum]

        env = dict(os.environ)
        # no access to prefs!
        #if self.prefService.prefs.hasPref("tclExtraPaths"):
        #    tclExtraPaths = self.prefService.prefs.getStringPref("tclExtraPaths")
        #    # If TCLLIBPATH is set, then it must contain a valid Tcl
        #    # list giving directories to search during auto-load
        #    # operations. Directories must be specified in Tcl format,
        #    # using "/" as the path separator, regardless of platform.
        #    # This variable is only used when initializing the
        #    # auto_path variable. 
        #    tclExtraPaths = tclExtraPaths.replace('\\', '/')
        #    TCLLIBPATH = ' '.join(tclExtraPaths.split(os.pathsep))
        #    env["TCLLIBPATH"] = TCLLIBPATH

        # Run language engine and report any errors.
        p = process.ProcessOpen(argv, env=env)
        content = re.sub("(\r\n|\r)", "\n", request.content)
        p.stdin.write(content)
        p.stdin.close()
        stdout = p.stdout.read()
        stderr = p.stderr.read()
        p.close()
        return stdout


class PHPEngineDriver(LanguageEngineDriver):
    name = "PHP"
    def __init__(self):
        from codeintel import phpcile
        self.scanner = phpcile.scan
    def scan(self, request):
        import parser
        import phpcile
        try:
            return self.scanner(request.content, request.path,
                                request.md5sum, request.mtime)
        except phpcile.PHPCILEError, ex:
            #XXX Probably will have to trap other possible errors because
            #    it doesn't look like phpcile.py really uses this one.
            raise LanguageEngineError(ex)
        except Exception, ex:
            raise LanguageEngineError("unexpected internal error in "
                                      "phpcile.py: %s" % ex)


class JavaScriptEngineDriver(LanguageEngineDriver):
    name = "JavaScript"
    def __init__(self):
        from codeintel import jscile
        self.scanner = jscile.scan
    def scan(self, request):
        import parser
        import jscile
        try:
            return self.scanner(request.content, request.path,
                                request.md5sum, request.mtime)
        except jscile.JSCILEError, ex:
            #XXX Probably will have to trap other possible errors because
            #    it doesn't look like jscile.py really uses this one.
            raise LanguageEngineError(ex)
        except Exception, ex:
            raise LanguageEngineError("unexpected internal error in "
                                      "jscile.py: %s" % ex)


class XBLEngineDriver(LanguageEngineDriver):
    name = "XBL"
    def __init__(self):
        from codeintel import xblcile
        self.scanner = xblcile.scan
    def scan(self, request):
        import xblcile
        try:
            return self.scanner(request.content, request.path,
                                request.md5sum, request.mtime)
        except xblcile.XBLCILEError, ex:
            raise LanguageEngineError(ex)


class LanguageEngineRegistry:
    """A registry of LanguageEngineDriver used to map file content type to
    appropriate Language Engines.

    Generally this registry is created automatically with the default set
    of mappings. However the user may create and customize his own such
    registry to 
    """
    DEFAULT_LED_CLASSES = {  # default language engine driver classes
        "Python": [# _SuperDaveEngineDriver, #XXX for testing
                   PythonEngineDriver],
        "PHP": [PHPEngineDriver],
        "JavaScript": [JavaScriptEngineDriver],
        "Perl": [PerlEngineDriver],
        "Tcl": [TclEngineDriver],
        #"XBL": [XBLEngineDriver],
    }

    def __init__(self, excludeDefaults=0):
        """Create a language engine registry.
        
            "excludeDefaults" (optional) is a boolean indicating if the
                default mappings should be excluded.
        """
        # mapping of <language> to list of LanguageEngineDriver instances
        self._drivers = {}
        if not excludeDefaults:
            for language, classes in self.DEFAULT_LED_CLASSES.items():
                for klass in classes:
                    self.registerDriver(language, klass())

    def registerDriver(self, language, driver):
        """Register a Language Engine Driver for a specific content type.
        
            "language" is the content type that the given language engine
                is able to process.
            "driver" is a LanguageEngineDriver instance.
        """
        self._drivers.setdefault(language, [])
        self._drivers[language].append(driver)

    def getLanguages(self):
        """Return a list of languages with registered drivers."""
        return self._drivers.keys()

    def getDrivers(self, language):
        """Return all registered Language Engines for this language.

        Returns a list of LanguageEngineDriver instances. If there are no
        engines for this langage an empty list is returned.
        """
        return self._drivers.get(language, [])


