#!/usr/bin/env python
# Copyright (c) 2004 Sophos.
# See the file LICENSE.txt for licensing information.
#
# Contributors:
#   Shane Caraveo (ShaneC@ActiveState.com)
#   Trent Mick (TrentM@ActiveState.com)

"""
    jscile - a Code Intelligence Language Engine for the JavaScript 1.5 language

    Module Usage:
        from jscile import scan
        mtime = os.stat("foo.js")[stat.ST_MTIME]
        content = open("foo.js", "r").read()
        scan(content, "foo.js", mtime=mtime)
    
    Command-line Usage:
        jscile.py [<options>...] [<JavaScript files>...]

    Options:
        -h, --help          dump this help and exit
        -V, --version       dump this script's version and exit
        -v, --verbose       verbose output
        -f, --filename <path>   specify the filename of the file content
                            passed in on stdin, this is used for the "path"
                            attribute of the emitted <file> tag.
        --md5=<string>      md5 hash for the input
        --mtime=<secs>      modification time for output info, in #secs since
                            1/1/70.

        -t, --text          plain text output mode (for debugging)
        -c, --clock         print time info for the scan (for debugging)

    One or more JavaScript files can be specified as arguments or content can
    be passed in on stdin.

    This is a Language Engine for the Code Intelligence (codeintel) system.
    Code Intelligence XML format. See:
        http://specs.activestate.com/Komodo_3.0/func/code_intelligence.html
    
    The command-line interface will return non-zero iff the scan failed.
"""

import os
import sys
import getopt
import md5
import re
import logging
import string
import time
import glob
import stat



#---- exceptions

class JSCILEError(Exception):
    pass



#---- global data

_version_ = (0, 2, 0)
log = logging.getLogger("jscile")



#---- internal routines and classes

def _getAttrStr(attrs):
    """Construct an XML-safe attribute string from the given attributes
    
        "attrs" is a dictionary of attributes
    
    The returned attribute string includes a leading space, if necessary,
    so it is safe to use the string right after a tag name.
    """
    from xml.sax.saxutils import quoteattr
    s = ''
    for attr, value in attrs.items():
        s += ' %s=%s' % (attr, quoteattr(str(value)))
    return s


def _xmlencode(s):
    """Encode the given string for inclusion in a UTF-8 XML document.
    
    Specifically, illegal or unpresentable characters are encoded as
    XML character entities.
    """
    if sys.version_info >= (2, 3):
        e = s.encode('utf-8', 'xmlcharrefreplace')
    else:
        # Python 2.2 doesn't have 'xmlcharrefreplace'. Fallback to a
        # literal '?' -- this is better than failing outright.
        e = s.encode('utf-8', 'replace')
    # As defined in the XML spec some of the character from 0x00 to 0x19
    # are not allowed in well-formed XML. We replace those with entity
    # references here.
    #   http://www.w3.org/TR/2000/REC-xml-20001006#charsets
    # (XXX It would be nice if Python has a codec for this. Perhaps we
    # should write one.)
    e = re.sub(# match 0x00-0x1f except TAB(0x09), LF(0x0A), and CR(0x0D)
               '([\x00-\x08\x0b\x0c\x0e-\x1f])',
               # replace with XML decimal char entity, e.g. '&#7;'
               lambda m: '&#%d;'%ord(m.group(1)),
               e)
    return e


class JScile:
    def __init__(self):
        # filesparsed contains all files parsed
        self.filesparsed={}
        # infile contains a list of files that the key file is included in 
        self.infile={}
        # classindex tells us what file a class definition is contained in
        self.classindex={}
        # functionindex tells us what file a function is defined in
        self.functionindex={}
        # interfaceindex tells us what file an interface definition is contained in
        self.interfaceindex={}

    def __repr__(self):
        r = ''
        for f in self.filesparsed:
            r += repr(self.filesparsed[f])
        return r + '\n'

    def toXML(self, level=0, indent="    "):
        #r = '<?xml version="1.0"?>\n' #encoding="UTF-8"?>\n'
        r = '<?xml version="1.0" encoding="UTF-8"?>\n'
        r += '<codeintel version="0.1">\n'
        for f in self.filesparsed.values():
            r += f.toXML(level+1, indent)
        r += '</codeintel>\n'
        return r

class JSArg:
    def __init__(self, data):
        self.data = data
        
    def __repr__(self):
        attrs = ''
        if self.data[0]:
            attrs += '%s ' % self.data[0]
        if self.data[1]:
            attrs += self.data[1]
        if self.data[2]:
            attrs += ' = %s' % self.data[2]
        return attrs
        
    def toXML(self, level=0, indent="    "):
        attrs = {}
        if self.data[0]:
            attrs['type'] = self.data[0]
        if self.data[1]:
            attrs['name'] = self.data[1]
        if self.data[2]:
            attrs['default'] = self.data[2]
        return '%s<argument%s/>\n' % (indent*level, _getAttrStr(attrs))

class JSArgs:
    def __init__(self, argline):
        self.args = collector.res["args_info"].findall(argline)
        self.argline = argline
    
    def __repr__(self):
        args = []
        for arg in self.args:
            args.append(repr(arg))
        return string.join(args, ', ')

    def toXML(self, level=0, indent="    "):
        out = ''
        for arg in self.args:
            out += '%s<argument name="%s"/>\n' % (indent*level, arg)
        return out

class JSVariable:
    def __init__(self, name, line, classname=''):
        self.name = name
        self.type = classname
        self.line = line

    def __repr__(self):
        return "var %s line %s type %s\n"\
               % (self.name, self.line, self.type)

    def toXML(self, level=0, indent="    "):
        type = ''
        attrs = {'name': self.name, 'line': self.line}
        if self.type:
            attrs['type'] = self.type
        return '%s<variable%s/>\n' % (indent*level, _getAttrStr(attrs))

class JSFunction:
    def __init__(self, funcname, args, lineno, depth=0, doc=None):
        self.name = funcname
        self.argline = args
        self._class = None
        self.variables = {}
        self.line = lineno
        self.lineend = 0
        self.depth = depth
        if args:
            self.args = JSArgs(args)
        else:
            self.args = None
        self.doc = doc
        
    def __repr__(self):
        return 'function %s(%s);\n' % (self.name, self.argline)

    def toXML(self, level=0, indent="    "):
        variables = args = doc = ''
        attrs = {'name': self.name, 'line': self.line}
        signature = "function "+self.name
        if self.lineend:
            attrs['lineend'] = self.lineend
        if self.doc:
            # XXX parse the docstring
            doc = '%s<doc><![CDATA[%s]]></doc>\n'\
                  % (indent*(level+1), self.doc)
        signature += "("
        if self.args:
            args = self.args.toXML(level+1, indent)
            signature += ' '.join(self.args.argline.split())
        signature += ")"
        attrs["signature"] = signature

        v = self.variables.keys()
        v.sort()
        for vk in v:
            variables += self.variables[vk].toXML(level+1, indent)

        children = doc + args + variables
        if children:
            return '%s<function%s>\n%s%s</function>\n'\
                   % (indent*level, _getAttrStr(attrs), children,
                      indent*level)
        else:
            return '%s<function%s/>\n' % (indent*level, _getAttrStr(attrs))

   
class JSClass:
    def __init__(self, name, lineno, depth, doc=None):
        self.name = name
        self.constructor = name
        self.extends = ''
        self.linestart = lineno
        self.lineend = 0
        self.depth = depth
        self.members = {} # declared class variables
        self.variables = {} # all variables used in class
        self.functions = {}
        self.doc = doc

    def __repr__(self):
        # dump our contents to human readable form
        r = "CLASS %s" % self.name
        if self.extends:
            r += " EXTENDS %s" % self.extends
        r += '\n'

        if self.members:
            r += "Members:\n"
            for m in self.members:
                r += "    var %s line %s\n"  % (m, self.members[m])

        if self.functions:            
            r += "functions:\n"
            for f in self.functions.values():
                r += "    %r" % f

        if self.variables:
            r += "variables:\n"
            for v in self.variables.values():
                r += "    %r" % v
            
        return r + '\n'

    def toXML(self, level=0, indent="    "):
        attrs = {'name': self.name,
                 'line': self.linestart,
                 'lineend': self.lineend}
        signature = "class "+self.name
        if self.extends:
            signature += " extends "+self.extends
        attrs["signature"] = signature
        out = '%s<class%s>\n' % (indent*level, _getAttrStr(attrs))
        
        if self.doc:
            # XXX parse the docstring
            out += '%s<doc><![CDATA[%s]]></doc>\n'\
                   % (indent*(level+1), self.doc)

        if self.extends:
            out += '%s<base%s/>\n'\
                   % (indent*(level+1), _getAttrStr({'name': self.extends}))

        f = self.functions.keys()
        f.sort()
        for fk in f:
            out += self.functions[fk].toXML(level+1, indent)

        m = self.members.keys()
        m.sort()
        for mk in m:
            out += self.members[mk].toXML(level+1, indent)

        v = self.variables.keys()
        v.sort()
        for vk in v:
            out += self.variables[vk].toXML(level+1, indent)
        out += '%s</class>\n' % (indent*level)
        return out


class JSFile:
    """CIX specifies that a <file> tag have zero or more <module> children.
    In JavaScript this is a one-to-one relationship, so this class represents both
    (and emits the XML tags for both).
    """
    def __init__(self, filename, content=None, md5sum=None, mtime=None):
        self.filename = filename
        if content is None:
            self.content = self.read()
        else:
            self.content = content
        if md5sum is None:
            self.md5sum = md5.new(self.content).hexdigest()
        else:
            self.md5sum = md5sum
        if mtime is None:
            self.mtime = int(time.time())
        else:
            self.mtime = mtime

        self.functions = {} # functions declared in file
        self.classes = {} # classes declared in file
        self.variables = {} # all variables used in file
        self.includes = {} # files included into this file
        self.interfaces = {} # interfaces declared in file

    def read(self):
        f = open(self.filename, "r")
        try:
            return f.read()
        finally:
            f.close()
    
    def __repr__(self):
        # dump our contents to human readable form
        r = "FILE %s\n" % self.filename

        for f, l in self.includes.items():
            r += "include %s on line %d\n" % (f, l)

        r += "functions:\n"
        for f in self.functions.values():
            r += "    %r" % f

        r += "variables:\n"
        for v in self.variables.values():
            r += "    %r" % v

        r += "classes:\n"
        for c in self.classes.values():
            r += repr(c)

        return r + '\n'
    
    def toXML(self, level=0, indent="    "):
        fileAttrs = {"language": "JavaScript",
                     "generator": "JavaScript",
                     "md5": self.md5sum,
                     "mtime": self.mtime,
                     "path": self.filename.replace('\\', '/')}
        out = '%s<file%s>\n' % (indent*level, _getAttrStr(fileAttrs))

        moduleAttrs = {"name": os.path.basename(self.filename)}
        out += '%s<module%s>\n'\
               % (indent*(level+1), _getAttrStr(moduleAttrs))
        for fn in self.includes:
            attrs = {'filename': fn, 'line': self.includes[fn]}
            out += '%s<import%s/>\n' % (indent*(level+2), _getAttrStr(attrs))

        for f in self.functions.values():
            out += f.toXML(level+2, indent)

        for v in self.variables.values():
            out += v.toXML(level+2, indent)

        for i in self.interfaces.values():
            out += i.toXML(level+2, indent)

        for c in self.classes.values():
            out += c.toXML(level+2, indent)

        out += '%s</module>\n' % (indent*(level+1))
        out += '%s</file>\n' % (indent*level)
        return out



class recollector:
    def __init__(self):
        self.res = {}
        self.regs = {}
        
    def add(self, name, reg, mods=None ):
        self.regs[name] = reg % self.regs
        #print "%s = %s" % (name, self.regs[name])
        if mods:
            self.res[name] = re.compile(self.regs[name], mods) # check that it is valid
        else:
            self.res[name] = re.compile(self.regs[name]) # check that it is valid

        
collector = recollector()
a = collector.add

a("newlines" ,          "[ \t]*(\r\n|\r|\n)")
a("comment_line" ,      "(//.*)")
a("comment_block" ,     r'(?P<doc>/\*.*?\*/)', re.S|re.M)
if sys.version_info[:2] == (2,0):
    a("string" ,        r'^(?<=(?<!\\)"|\').*?((?<!\\)\1)', re.S|re.M)
else:
    a("string" ,        r'((?<!\\)"|\').*?((?<!\\)\1)', re.S|re.M)
a("string_delim" ,      r'((?:(?<!\\)"|\'))')
a("vartypes",           r'(?:var|const)\s+',re.I)

a("variable_object",    r'var\s*(?P<classname>[\w_]*?)\s*=\s*{',re.I|re.M)
a("variable_with_type", r'(?:%(vartypes)s)(?:\s*(?P<variable>[\w_]*?)\s*=\s*new\s*(?P<classname>[\w_]*?)(?:\(.*?\)|\s*)\s*);',re.I|re.M)
a("variable" ,          r'(?:%(vartypes)s)(?:\s*(?P<variable>[\w_]*?)(?:(?:\s*=.*?)|\s*));',re.I|re.M)

a("block_begin" ,       "{")
a("block_end" ,         "}")

a("function" ,          r'\s*function\s+(?P<function>[\w_]+)\s*\((?P<args>.*?)\)',re.I|re.S|re.M)

# class type 1
# classname.prototype = {}
a("class_prototype",        r'(?P<classname>[\w_]*?)\s*\.\s*prototype\s*=\s*',re.I)

# class parent
# class.prototype = new classname();
a("class_parent",       r'(?P<classname>[\w_]*?)\s*\.\s*prototype\s*=\s*new\s+(?P<parent>[\w_]*?)(?:\(.*?\))',re.I|re.S|re.M)

# class constructor
# class.prototype.constructor = functionname;
a("class_constructor",  r'(?P<classname>[\w_]*?)\s*\.\s*prototype\s*\.\s*constructor\s*=\s*(?P<constructor>[\w_]+)\s*;',re.I|re.S|re.M)

#a("class" ,             r'(?:%(comment_block)s?\s*%(class_type1)s)',re.I|re.S|re.M)

# classmember type 1
# classname.prototype.funcname = function(args)
a("class_function",     r'(?P<classname>[\w_]*?)\s*\.\s*prototype\s*\.\s*(?P<function>[\w_]*?)\s*=\s*function\s*\((?P<args>.*?)\)\s*',re.I|re.S|re.M)

# class var getter/setter type 2
a("class_member",    r'(?P<classname>[\w_]*?)\s*\.\s*prototype\s*\.\s*__define[S|G]etter__\(\"(?P<member>[\w_]*?)\"', re.I|re.M)

# classmember type 2
#    works with class_prototype
# name: function(args)
# \s*(?P<name>[\w_]*?)\s*:\s*function\s*\((?P<args>.*?)\)
a("class_function2", r'(?P<function>[\w_]*?)\s*:\s*function\s*\((?P<args>.*?)\)',re.I)

#a("classmember" ,       r'(?:(?P<attributes>(?:%(vartypes)s)*)\$(?P<name>[\w_]*?)\s*[=|;])',re.I)

# class var getter/setter
# (?:\s+|{)(?:set|get)\s*(?P<name>[\w_]*?)\s*\(
a("class_getset",    r'(?:set|get)\s+(?P<variable>[\w_]*?)\s*\(', re.I|re.M)


# class var type 3
# (?:\s+|{)(?P<name>[\w_]+?)\s*:(?!\s*function)
a("class_var",    r'(?P<variable>[\w_]+?)\s*:(?!\s*function)', re.I|re.M)


a("label" ,             "([a-zA-Z_][a-zA-Z0-9_]*)")
a("lnum" ,              "([0-9]+)")
a("whitespace" ,        "[ \t]+")
a("any_char" ,          "(.)") # |[\n]
a("args_info",          r'(?:\s*(?P<arg>[\w_]+)(?:.*?,)?)',re.S|re.M)

SKIPTOK = 0x01 # don't consider this a token that is to be considered a part of the grammar, like '\n'
MAPTOK = 0x02  # use the token associated with the pattern when it matches
EXECFN= 0x04   # execute the function associated with the pattern when it matches
USETXT = 0x08  # if you match a single character and want its ascii value to be the token

# Lexer class borrowed from the PyLRd project,
# http://starship.python.net/crew/scott/PyLR.html
class Lexer:
    eof = -1

    def __init__(self):
        self.tokenmap = {}
        self.prgmap = {}
        self.prglist = []
        self.lasttok = -1
        self.text = ""
        self.textindex = 0
        self.tokennum2name = {}

    def nexttok(self):
        self.lasttok = self.lasttok + 1
        return self.lasttok

    def settext(self, t):
        self.text = t
        self.textindex = 0

    def addmatch(self, prg, func=None, tokname="", attributes=MAPTOK|EXECFN):
        self.prglist.append(prg)
        tok = -2
        if not func:
            attributes = attributes & ~EXECFN
        if not tokname:
            attributes = attributes & ~MAPTOK
        if attributes & MAPTOK:
            self.tokenmap[tokname] = tok = self.nexttok()
        else:
            tok = self.nexttok()
        self.prgmap[prg] = tok, attributes, func
        self.tokennum2name[tok] = tokname

    def scan(self):
        for prg in self.prglist:
            mo = prg.match(self.text, self.textindex)
            if not mo: 
                continue
            self.textindex = self.textindex + len(mo.group(0))
            tmpres = mo.group(0)
            t, attributes, fn = self.prgmap[prg]
            #log.debug("'%s' token: %r", self.tokennum2name[t], tmpres)
            if attributes & EXECFN:
                tmpres = apply(fn, (mo,))
            if attributes & USETXT:
                t = ord(mo.group(0)[0])
            return (t, tmpres)
        if self.textindex >= len(self.text):
            return (self.eof, "")
        raise "Syntax Error in lexer"


class JSLexer:
    def __init__(self, filename, content=None, md5sum=None, mtime=None):
        # hook up the lexical matches to a function that handles the token
        self.lex_matches = [
            ('newline',         collector.res["newlines"],        self.incline,           EXECFN|SKIPTOK),    
            ('comment_line',    collector.res["comment_line"],    None,                   SKIPTOK),
            #('comment_block',   collector.res["comment_block"],   self.doMultiLineBlock,  EXECFN|SKIPTOK),
            #('string',          collector.res["string"],          self.doMultiLineBlock,  EXECFN|SKIPTOK),
            #('string',          collector.res["string_delim"],    self.toggleStringState, EXECFN|SKIPTOK),
            ('block_begin',     collector.res["block_begin"],     self.incBlock,          EXECFN|SKIPTOK),    
            ('block_end',       collector.res["block_end"],       self.decBlock,          EXECFN|SKIPTOK),    
            ('variable_object', collector.res["variable_object"], self.addObject,         EXECFN|SKIPTOK),
            ('variable_with_type', collector.res["variable_with_type"], self.addVariable, EXECFN|SKIPTOK),
            ('variable',        collector.res["variable"],        self.addVariable,       EXECFN|SKIPTOK),
            ('function',        collector.res["function"],        self.addFunction,       EXECFN|SKIPTOK),
            ('classfunction',   collector.res["class_function"],  self.addClassPart,      EXECFN|SKIPTOK),
            ('classparent',     collector.res["class_parent"],    self.addClassPart,      EXECFN|SKIPTOK),
            ('classconstructor',collector.res["class_constructor"], self.addClassPart,    EXECFN|SKIPTOK),
            ('classmember',     collector.res["class_member"],    self.addClassPart,      EXECFN|SKIPTOK),
            ('classfunction2',  collector.res["class_function2"], self.addFunction,       EXECFN|SKIPTOK),
            ('class',           collector.res["class_prototype"], self.addClass,          EXECFN|SKIPTOK),
            ('classgetset',     collector.res["class_getset"],    self.addClassMember,    EXECFN|SKIPTOK),
            ('classvar',        collector.res["class_var"],       self.addClassMember,    EXECFN|SKIPTOK),
            ('label',           collector.res["label"],           None,                   0),    
            ('lnum',            collector.res["lnum"],            None,                   SKIPTOK),    
            ('whitespace',      collector.res["whitespace"],      None,                   SKIPTOK),    
            ('any_char',        collector.res["any_char"],        None,                   SKIPTOK),    
        ]

        # setup lexer and add token matching regexes
        self.l = Lexer()
        for p in self.lex_matches:
            # log.debug("adding %r",p[0])
            attributes = p[3]
            if not attributes: attributes = MAPTOK|EXECFN
            self.l.addmatch(p[1],p[2],p[0],attributes)

        self.cile = JScile()
        self.lineno = 1
        self.jscode = 1
        self._stringDelim = ""
        self.depth = 0
        self.classStack = []
        self.currentClass = None
        self.currentFunction = None
        
        self.filename = filename
        self.fileinfo = JSFile(self.filename, content, md5sum, mtime)
        
        self.scanData()
        self.cile.filesparsed[self.filename] = self.fileinfo

    def idfunc(self, m):
        if not self.jscode: return
        log.info("ID: %r",m.group(0))
        return m.group(0)

    def incline(self, m):
        self.lineno = self.lineno + 1
        # log.debug("line: %d",self.lineno)attributes
        return ""

    def toggleStringState(self, m):
        if self.jscode:
            log.info("exiting code fragment")
            self.jscode = 0
            self._stringDelim = m.group(0)
        elif self._stringDelim == m.group(0):
            log.info("entering code fragment")
            self.jscode = 1
        return ""

    # takes a block, and figures out how many lines it spans, to
    # keep lineno correct
    def doMultiLineBlock(self, m):
        if not self.jscode: return
        nl = collector.res["newlines"].findall(m.group(0))
        # log.debug("block start at lineno %d",self.lineno)
        self.lineno = self.lineno + len(nl)
        # log.debug("block had %d lines, lineno is %d", len(nl), self.lineno)

    def incBlock(self, m):
        if not self.jscode: return
        self.depth = self.depth+1
        log.debug("+:depth at %d line %d", self.depth, self.lineno)
        return ""

    def decBlock(self, m):
        if not self.jscode: return
        self.depth = self.depth-1
        log.debug("-:depth at %d line %d", self.depth, self.lineno)
        if self.currentFunction and self.currentFunction.depth == self.depth:
            log.debug("done with function %s at depth %d line %d", self.currentFunction.name, self.depth, self.lineno)
            self.currentFunction.lineend = self.lineno
            if self.currentFunction._class:
                self.currentFunction._class.lineend = self.lineno
            # XXX stacked functions used to work in javascript, need verify still is
            self.currentFunction = None
        if self.currentClass and self.currentClass.depth == self.depth:
            log.debug("done with class %s at depth %d", self.currentClass.name, self.depth)
            self.currentClass.lineend = self.lineno
            self.currentClass = self.classStack.pop()
        return ""

    def addFunction(self, m):
        if not self.jscode: return
        log.info("FUNC: %s(%s) on line %d", m.group('function'), m.group('args'), self.lineno)
        self.doMultiLineBlock(m)
        fn = JSFunction(m.group('function'), m.group('args'), self.lineno, self.depth,
                                           doc=m.groupdict().get('doc'))
        if self.currentClass:
            self.currentClass.functions[fn.name] = fn
            if fn.line < self.currentClass.linestart: self.currentClass.linestart = fn.line
            if fn.lineend > self.currentClass.lineend: self.currentClass.lineend = fn.lineend
        else:
            self.fileinfo.functions[fn.name] = fn
            self.cile.functionindex[fn.name] = self.fileinfo.filename
        self.currentFunction = fn
        return ""

    def _addClassPart(self, m):
        if not self.jscode: return
        d = m.groupdict()
        jsclass = None
        if d['classname']:
            fn = None
            if d['classname'] not in self.fileinfo.classes:
                jsclass = JSClass(d['classname'], self.lineno, self.depth,
                                             doc=d.get('doc'))
                self.fileinfo.classes[jsclass.name] = jsclass
                # log.debug("adding classindex[%s]=%s", m.group('name'), self.fileinfo.filename)
                self.cile.classindex[jsclass.name]=self.fileinfo.filename
                log.info("CLASS: %s extends %s on line %d in %s at depth %d\nDOCS: %s",
                         jsclass.name, jsclass.extends, 
                         jsclass.linestart, self.filename, self.depth,
                         jsclass.doc)

            else:
                jsclass = self.fileinfo.classes[d['classname']]
            
            if 'function' in d:
                log.info("FUNC: %s(%s) on line %d", d['function'], d['args'], self.lineno)
                fn = JSFunction(d['function'], d['args'], self.lineno, self.depth, doc=d.get('doc'))
                fn._class = jsclass
                jsclass.functions[fn.name] = fn
                self.currentFunction = fn
            if 'parent' in d:
                jsclass.extends = d['parent']
            if 'constructor' in d:
                # XXX this isn't done
                jsclass.constructor = d['constructor']
            if 'member' in d:
                v = JSVariable(d['member'], self.lineno)

            if jsclass.constructor in self.fileinfo.functions:
                # move this to the classvar now
                fn = jsclass.functions[jsclass.constructor] = self.fileinfo.functions[jsclass.constructor]
                del self.fileinfo.functions[jsclass.constructor]
                del self.cile.functionindex[jsclass.constructor]
            if fn:
                if fn.line < jsclass.linestart: jsclass.linestart = fn.line
                if fn.lineend > jsclass.lineend: jsclass.lineend = fn.lineend

            self.doMultiLineBlock(m)
        return jsclass

    # a class part using prototype.name = function
    def addClassPart(self, m):
        self._addClassPart(m)
        return ""

    def addObject(self, m):
        # this is the same as addClass, but we need to start a block also
        ret = self.addClass(m)
        self.incBlock(m)
        return ret

    # a class using classname.prototype = { ... }
    def addClass(self, m):
        if not self.jscode: return
        
        jsclass = self._addClassPart(m)
        if jsclass:
            self.classStack.append(self.currentClass)
            self.currentClass = jsclass

        return ""
    
    def addClassMember(self, m):
        if not self.jscode: return
        if self.currentClass:
            if m.group('variable') not in self.currentClass.members:
                self.currentClass.members[m.group('variable')] =\
                    JSVariable(m.group('variable'), self.lineno)
            log.info("CLASSMBR: %r", m.group('variable'))
        return ""
    
    def addVariable(self, m):
        if not self.jscode: return
        v = JSVariable(m.group('variable'), self.lineno, m.groupdict().get('classname'))
        if self.currentFunction and m.group('variable') not in self.currentFunction.variables:
            log.info("FVAR: %s on line %d", m.group('variable'), self.lineno)
            self.currentFunction.variables[m.group('variable')] = v
        elif m.group('variable') not in self.fileinfo.variables:
            log.info("VAR: %s on line %d", m.group('variable'), self.lineno)
            self.fileinfo.variables[m.group('variable')] = v
        return ""

    def scanData(self):
        # log.debug("doing: %s",data)
        self.l.settext(self.fileinfo.content)
            
        #res = []
        while 1:
            t, v = self.l.scan()
            # log.debug("lex symbol: %r %r", t, v)
            if t == self.l.eof:
                break
            #res.append((t, v))



#---- public module interface

def scan(content, filename, md5sum=None, mtime=None, useRepr=0):
    """Scan the given JavaScript content and return Code Intelligence data
    conforming the the Code Intelligence XML format.
    
        "content" is the JavaScript content to scan
        "filename" is the source of the JavaScript content (used in the
            generated output).
        "md5sum" (optional) if the MD5 hexdigest has already been calculated
            for the content, it can be passed in here. Otherwise this
            is calculated.
        "mtime" (optional) is a modified time for the file (in seconds since
            the "epoch"). If it is not specified the _current_ time is used.
            Note that the default is not to stat() the file and use that
            because the given content might not reflect the saved file state.
        "useRepr" (optional) is a boolean for debugging.

    This will raise a JSCILEError if there was an error processing.
    """
    lex = JSLexer(filename, content, md5sum, mtime)
    if useRepr:
        return repr(lex.cile)
    return lex.cile.toXML()


#---- mainline

def main(argv):
    logging.basicConfig()

    # Parse options.
    useRepr = 0
    doTiming = 0
    stdinFilename = None
    md5sum = None
    mtime = None
    try:
        opts, args = getopt.getopt(argv[1:], "Vvhf:tc",
            ["version", "verbose", "help", "filename=", "mtime=", "md5=",
             "text", "clock"])
    except getopt.GetoptError, ex:
        log.error(str(ex))
        log.error("Try `jscile --help'.")
        return 1
    for opt, optarg in opts:
        if opt in ("-h", "--help"):
            sys.stdout.write(__doc__)
            return
        elif opt in ("-V", "--version"):
            ver = '.'.join([str(part) for part in _version_])
            print "jscile %s" % ver
            return
        elif opt in ("-v", "--verbose"):
            log.setLevel(logging.DEBUG)
        elif opt in ("-f", "--filename"):
            stdinFilename = optarg
        elif opt in ("--md5",):
            md5sum = optarg
        elif opt in ("--mtime",):
            mtime = optarg
        elif opt in ("-t", "--text"):
            useRepr = 1
        elif opt in ("-c", "--clock"):
            doTiming = 1

    if len(args) == 0:
        contentOnStdin = 1
        filenames = [stdinFilename or "<stdin>"]
    else:
        contentOnStdin = 0
        filenames = []
        for arg in args:
            filenames += glob.glob(arg)

    try:
        for filename in filenames:
            if contentOnStdin:
                log.debug("reading content from stdin")
                content = sys.stdin.read()
                log.debug("finished reading content from stdin")
                if mtime is None:
                    mtime = int(time.time())
            else:
                if mtime is None:
                    mtime = int(os.stat(filename)[stat.ST_MTIME])
                fin = open(filename, 'r')
                try:
                    content = fin.read()
                finally:
                    fin.close()

            if doTiming: start = time.clock()
            data = scan(content, filename, md5sum=md5sum, mtime=mtime,
                        useRepr=useRepr)
            if doTiming: end = time.clock()
            if data:
                sys.stdout.write(data)
            if doTiming:
                print "time to parse '%s': %fs" % (filename, end-start)
    except JSCILEError, ex:
        log.error(str(ex))
        if log.isEnabledFor(logging.DEBUG):
            print
            import traceback
            traceback.print_exception(*sys.exc_info())
        return 1
    except KeyboardInterrupt:
        log.debug("user abort")
        return 1


if __name__ == "__main__":
    sys.exit( main(sys.argv) )

