#!python
# Copyright (c) 2004 Sophos Inc.
# See the file LICENSE.txt for licensing information.

"""Code Intelligence Scan Scheduling

Usage:
    XXX describe usage of the scheduler
"""

import os, sys
import threading
import time
import bisect
import Queue
import md5

import logging
import sqlite

from codeintel.common import *
from codeintel.database import Database


#---- globals

log = logging.getLogger("codeintel.scheduler")
#log.setLevel(logging.DEBUG)


#---- internal support

class _PriorityQueue(Queue.Queue):
    """A thread-safe priority queue.
    
    In order to use this the inserted items should be tuples with the
    priority first. Note that subsequent elements of the item tuples will
    be used for secondary sorting. As a result, it is often desirable to
    make the second tuple index be a timestamp so that the queue is a
    FIFO for elements with the same priority, e.g.:
        item = (PRIORITY, time.time(), element)
        
    Usage:
        q = _PriorityQueue(0)  # unbounded queue
        q.put( (2, time.time(), "second") )
        q.put( (1, time.time(), "first") )
        q.put( (3, time.time(), "third") )
        priority, timestamp, value = q.get()
    """
    def _put(self, item):
        bisect.insort(self.queue, item)


class _ScanRequestQueue(_PriorityQueue):
    """A thread-safe priority queue for ScanRequests.
    
    This queue class extends _PriorityQueue with this condition:
    - When adding a ScanRequest to the queue, if a ScanRequest for the
      same file (but not necessarily the same md5) already exists in the
      queue, then the new ScanRequest inherits the higher priority and the
      earlier timestamp of the two and _replaces_ the older ScanRequest.

    This condition is added because there is no point in scanning file
    contents from time T1 when a scan of the file contents at time T2
    (more recent) is being requested. It is important to adopt the
    higher priority (and earlier timestamp) to ensure the requestor does
    not starve.

    Note: This presumes that an "item" is this 3-tuple:
        (<priority-number>, <timestamp>, <ScanRequest instance>)
    """
    def __init__(self, maxsize=0):
        _PriorityQueue.__init__(self, maxsize)
        self.__cpath2item = {}
    def _put(self, item):
        # Remove a possible existing request for the same file (there can
        # be only one).
        priority, timestamp, request = item
        try:
            cpath = request.getCanonicalPath()
        except ValueError:
            cpath = request.path
        if cpath in self.__cpath2item:
            i = self.__cpath2item[cpath]
            self.queue.remove(i)
            p, t, r = i
            item = (min(priority, p), t, request)
        # Add the (possibly updated) item to the queue.
        self.__cpath2item[cpath] = item
        _PriorityQueue._put(self, item)
    def _get(self):
        item = _PriorityQueue._get(self)
        try:
            cpath = item[2].getCanonicalPath()
        except ValueError:
            cpath = item[2].path
        del self.__cpath2item[cpath]
        return item

class _StagingScanRequestQueue(_ScanRequestQueue):
    """A priority queue for ScanRequests with delayed staging support
    
    This queue class extends _StagingScanRequestQueue by adding the .stage()
    method. This method is like the regular .put() method except that staged
    ScanRequests are only actually placed on the queue if a certain period
    of inactivity (by default, one second) passes without subsequent scan
    requests for the same file path.
    
    This is to support reasonable performance for live updating while a
    document is being edited. Rather than executing a scan for every
    intermediate edited state, scanning is only  after a period of
    relative inactivity.
    
    One additional burden is that a "staging thread" is involved so one must
    call this queue's .finalize() method to properly shut it down.
    
    As with the _ScanRequestQueue this queue presumes that and item is this
    3-tuple:
            (<priority-number>, <timestamp>, <ScanRequest instance>)
    """
    DEFAULT_STAGING_DELAY = 1.5 # default delay from on deck -> on queue (s)

    def __init__(self, maxsize=0, stagingDelay=None):
        """Create a staging scan request queue.
        
            "maxsize" (optional) is an upperbound limit on the number of
                items in the queue (<= 0 means the queue is unbounded).
            "stagingDelay" (optional) is a number of seconds to use as a
                delay from being staged to being placed on the queue.
        """
        _ScanRequestQueue.__init__(self, maxsize)
        if stagingDelay is None:
            self._stagingDelay = self.DEFAULT_STAGING_DELAY
        else:
            self._stagingDelay = stagingDelay
        self._onDeck = {
            # <path> : (<time when due>, <queue item>)
        }
        self._nothingOnDeck = threading.Lock()
        self._nothingOnDeck.acquire()
        self._terminate = 0 # boolean telling "staging thread" to terminate
        self._stager = threading.Thread(target=self._stagingThread,
                                        name="staging thread")
        self._stager.start()

    def finalize(self):
        if self._stager:
            self._terminate = 1
            # Make sure staging thread isn't blocked so it can terminate.
            self.mutex.acquire()
            try:
                if not self._onDeck:
                    self._nothingOnDeck.release()
            finally:
                self.mutex.release()
            # Don't bother join'ing because there is no point waiting for
            # up to self._stagingDelay while the staging thread shuts down.
            #self._stager.join()

    def stage(self, item):
        self.mutex.acquire()
        try:
            priority, timestamp, request = item
            try:
                path = request.getCanonicalPath()
            except ValueError:
                path = request.path
            wasEmpty = not self._onDeck
            self._onDeck[path] = (timestamp + self._stagingDelay, item)
            if wasEmpty:
                self._nothingOnDeck.release()
        finally:
            self.mutex.release()

    def _stagingThread(self):
        """Thread that handles moving requests on-deck to the queue."""
        log.debug("staging thread: start")
        while 1:
            # If nothing is on-deck, wait until there is.
            #log.debug("staging thread: acquire self._nothingOnDeck")
            self._nothingOnDeck.acquire()
            #log.debug("staging thread: acquired self._nothingOnDeck")
            if self._terminate:
                break

            # Place any "due" items on the queue.
            self.mutex.acquire()
            somethingStillOnDeck = 1
            currTime = time.time()
            toQueue = []
            try:
                for path, (timeDue, item) in self._onDeck.items():
                    if currTime >= timeDue:
                        toQueue.append(item)
                        del self._onDeck[path]
                if not self._onDeck:
                    somethingStillOnDeck = 0
            finally:
                if somethingStillOnDeck:
                    self._nothingOnDeck.release()
                self.mutex.release()
            if toQueue: log.info("staging thread: queuing %r", toQueue)
            for item in toQueue:
                self.put(item)

            # Sleep for a bit.
            #XXX If the latency it too large we may want to sleep for some
            #    fraction of the staging delay.
            log.debug("staging thread: sleep for %.3fs", self._stagingDelay)
            time.sleep(self._stagingDelay)
        log.debug("staging thread: end")



#---- public classes

class _ScanRunner(threading.Thread):
    """A worker thread to run one scan request.
    
    The "finished" attribute is set to true when it is finished running.
    The "abort" boolean attribute can be set to true to suggest the runner
    abort mid-run.
    """
    def __init__(self, request, drivers, db, callback):
        """Create a thread to execute on ScanRequest
        
            "request" is the scan request instance.
            "drivers" is a list of relevant LanguageEngineDrivers to use for
                scanning.
            "db" is a Database instance (which provides an API to working
                with the a connection to the CIDB).
            "callback" is a callback to run on completion. The request should
                be passed to the callback.
        """
        threading.Thread.__init__(self, name="Scan Runner %s"%request.id)
        self.request = request
        self.drivers = drivers
        self.db = db
        self.callback = callback
        self.finished = 0
        self.abort = 0

    def run(self):
        TIME_SCANS = False
        try:
            #XXX What is the proper way to handle errors in the scan runner?
            #    When that is figured out then handle a possible os.error
            #    (and others, e.g. parser.ParserError) from this.
            self.request.loadContent()
            for driver in self.drivers:
                if self.abort:
                    #print "aborting request %s: '%s' (%s)"\
                    #      % (self.request.id,
                    #         os.path.basename(self.request.path),
                    #         self.request.language)
                    return

                if TIME_SCANS: startscan = time.clock()
                try:
                    cix = driver.scan(self.request)
                except LanguageEngineError, ex:
                    cix = """\
<?xml version="1.0" encoding="UTF-8"?>
<codeintel version="0.1">
    <file path="%s" generator="%s" language="%s" error="%s"/>
</codeintel>
""" % (self.request.path, driver.name, self.request.language,
       xmlencode(str(ex)))
                    #print "XXX error scanning '%s' with the %s Language Engine Driver: %s"\
                    #      % (self.request.path, driver.name, ex)
                if TIME_SCANS: endscan = time.clock()

                if self.abort:
                    #print "aborting request %s: '%s' (%s)"\
                    #      % (self.request.id,
                    #         os.path.basename(self.request.path),
                    #         self.request.language)
                    return

                #XXX If we ever get around to merging CIX documents then
                #    may want to merge here and then called self.db.update()
                #    ONCE with the merged CIX.
                if cix:
                    log.info("update database with CIX results for '%s' "
                             "(request %r)",
                             os.path.basename(self.request.path),
                             self.request.id)
                    log.debug("CIX for '%s' (request %r):\n%s",
                              self.request.path, self.request.id, cix)
                    #XXX Errors should be trapped here.
                    try:
                        if 0: # profile database update
                            import hotshot, hotshot.stats
                            prof = hotshot.Profile("codeintel_cidb_update.prof")
                            cix = prof.runcall(self.db.update, cix, self.request, driver)
                            prof.close()
                            stats = hotshot.stats.load("codeintel_cidb_update.prof")
                            stats.strip_dirs()
                            stats.sort_stats('time', 'calls')
                            stats.print_stats(20)
                        else:
                            self.db.update(cix, self.request, driver)
                    except CodeIntelError, ex:
                        pass #XXX Do something useful with the error.
                        log.error(str(ex))
                if TIME_SCANS:
                    print "scan of '%s': scan=%.3fs  update=%.3fs"\
                          % (os.path.basename(self.request.path),
                             endscan-startscan,
                             time.clock()-endscan)
        finally:
            self.finished = 1 # must set *before* callback so runner is cleaned up
            #XXX This callback should perhaps indicate if the scan and/or
            #    update was successful and if not, then why not (e.g.
            #    aborted).
            self.callback(self.request)


class Scheduler(threading.Thread):
    """The scan request scheduler thread.
    
    This thread manages a queue of ScanRequest's, scheduling the scans in
    priority order. It has two modes of usage:
        MODE_DAEMON
            The scheduler remains running until it is explicitly stopped with
            the .stop() method.
        MODE_ONE_SHOT
            All added requests are processed and then the scheduler
            terminated. Note that the .stageRequest() method is not
            allowed in this mode.

    Usage:
        from codeintel.scheduler import Scheduler
        s = Scheduler(Scheduler.MODE_DAEMON, manager)
        s.start()
        #...make s.addRequest(request) calls
        s.stop()
    """
    # Dev Notes:
    # - There is a potential race condition on request id generation
    #   if addRequest/stageRequest calls are made from multiple threads.

    POOL_SIZE = 1 # max. num of scanners that can be running simultaneously
    MODE_DAEMON, MODE_ONE_SHOT = range(2)

    class StopScheduler(Exception):
        """Used to signal that scheduler iteration should stop."""
        # Note: I *could* use StopIteration here, but I don't want to
        #       possibly misinterpret a real StopIteration.
        pass

    def __init__(self, mode, manager, requestCompletedCB=None,
                 requestStartedCB=None, completedCB=None):
        """Create a ScanRequest processing scheduler.
        
            "mode" is one of the MODE_* constants.
            "manager" is a codeintel.Manager instance
            "requestCompletedCB" (optional), if specified, is called when
                a request is completed.
            "requestStartedCB" (optional), if specified, is called when a
                request is started.
            "completedCB" (optional), if specified, is called when the
                Scheduler thread completes.
        """
        threading.Thread.__init__(self, name="Code Intel Scheduler")
        if mode not in (self.MODE_DAEMON, self.MODE_ONE_SHOT):
            raise CodeIntelError("illegal mode value: %s" % mode)
        self.mode = mode
        self.manager = manager
        self._requestCompletedCB = requestCompletedCB
        self._requestStartedCB = requestStartedCB
        self._completedCB = completedCB
        self._scannerPool = threading.Semaphore(self.POOL_SIZE)
        #XXX Do we have to worry about the _ridCounter (a Python int that
        #    will self-promote to a Python long) overflowing an exported
        #    XPCOM long type?
        self._ridCounter = 0 # counter for assigning unique request IDs
        if self.mode == self.MODE_DAEMON:
            self._requests = _StagingScanRequestQueue()
        else:
            self._requests = _ScanRequestQueue()
        self._runners = []  # list of worker _ScanRunner threads
        self._runnersLock = threading.Lock()
        self._stopping = False
        self._resumeEvent = None

    def stop(self):
        """Stop the scheduler thread (blocks until thread completes)."""
        self._stopping = True
        sentinel = ScanRequest("<stop>", None, PRIORITY_CONTROL)
        self.addRequest(sentinel)
        # ensure scheduler is not blocked waiting for "slot" for sentinel
        self._scannerPool.release()
        if self.mode == self.MODE_DAEMON:
            self._requests.finalize()
        try:
            self.join()
        except AssertionError:
            pass # thread was not started

    def pause(self):
        self._resumeEvent = threading.Event()
        self._pauseEvent = threading.Event()
        sentinel = ScanRequest("<pause>", None, PRIORITY_CONTROL)
        self.addRequest(sentinel)
        self._pauseEvent.wait() # wait until the Scheduler is actually paused
        log.debug("scheduler thread: paused")

    def resume(self):
        if self._resumeEvent:
            self._resumeEvent.set()
            self._resumeEvent = None
        log.debug("scheduler thread: resumed")

    def _abortMatchingRunner(self, path, language):
        # If a matching request is currently being processed, then tell it
        # to abort (whenever it can); there is little point in its data.
        self._runnersLock.acquire()
        try:
            for r in self._runners:
                if r.request.path == path and r.request.language == language:
                    #print "suggest runner matching new request abort: "\
                    #      "'%s' (%s)" % (os.path.basename(path), language)
                    r.abort = 1
                    break
        finally:
            self._runnersLock.release()

    def addRequest(self, request):
        """Add a scan request.
        
            "request" is a ScanRequest instance.
        """
        self._abortMatchingRunner(request.path, request.language)
        request.id = self._ridCounter; self._ridCounter += 1
        self._requests.put( (request.priority, time.time(), request) )

    def stageRequest(self, request):
        """Stage a scan request for addition to the queue.
        
            "request" is a ScanRequest instance.
        
        Staging a scan request involves putting it "on deck" for a certain
        period (called the "staging delay") before actually queuing it.
        Subsequently staged requests for the same file within that period
        replace the earlier request on deck. The point is to reduce
        unnecessary scanning for a rapidly changing file.
        """
        if self.mode == self.MODE_ONE_SHOT:
            raise CodeIntelError("cannot call .stageRequest() on a "
                                 "MODE_ONE_SHOT Scheduler")
        self._abortMatchingRunner(request.path, request.language)
        request.id = self._ridCounter; self._ridCounter += 1
        self._requests.stage( (request.priority, time.time(), request) )

    def getNumRequests(self):
        """Return the current number of requests."""
        return self._requests.qsize()

    def _requestCompleted(self, request):
        """Callback called by a _ScanRunner on completion of a ScanRequest."""
        if self._stopping:
            #HACK: Don't bother if we are shutting down. Made necessary by
            #      Komodo's given _requestCompletedCB that results in
            #      possible deadlock on proxied koICodeBrowserTreeView call.
            #      (Don't _really_ understand why.)
            return
        self._runnersLock.acquire()
        try:
            # Clean up zombie _ScanRunner threads.
            for i in range(len(self._runners)-1, -1, -1):
                if self._runners[i].finished:
                    del self._runners[i]
        finally:
            self._runnersLock.release()

        # Free this scanner slot.
        self._scannerPool.release()
        if self._requestCompletedCB:
            self._requestCompletedCB(request)

    def run(self):    # the scheduler thread run-time
        log.debug("scheduler thread: start")
        cx = self.manager.getDatabaseConnection()
        reason = "failed"
        try:
            while 1:
                try:
                    self._iteration(cx)
                except Queue.Empty:
                    reason = "completed"
                    break
                except self.StopScheduler:
                    reason = "stopped"
                    break
                except sqlite.Error, ex:
                    log.exception("unexpected SQLite error: %s: ignoring "
                                  "and continuing" % ex)
                except:
                    log.exception("unexpected internal error in scheduler "
                                  "thread: ignoring and continuing")
        finally:
            # Wait for current scan runners to complete.
            try:
                self._runnersLock.acquire()
                try:
                    activeRunners = [r for r in self._runners if not r.finished]
                finally:
                    self._runnersLock.release()
                for r in activeRunners:
                    try:
                        r.join()
                    except:
                        pass
            except:
                log.exception("unexpected error joining active scan runners")

            try:
                cx.close()
            except:
                log.exception("unexpected error closing scheduler CIDB "
                              "connection")
            
            try:
                if self._completedCB:
                    try:
                        self._completedCB(reason)
                    except:
                        pass
            except:
                log.exception("unexpected error in completion callback")
            log.debug("scheduler thread: end")

    def _iteration(self, cx):
        """one iteration of the scheduler thread loop
        
        Raises StopScheduler exception if iteration should stop.
        """
        #log.debug("scheduler: acquire scanning slot")
        self._scannerPool.acquire()  # get a slot for scanning
        #log.debug("scheduler: ACQUIRED scanning slot")
        #log.debug("scheduler: get request")
        if self.mode == self.MODE_DAEMON:
            priority, timestamp, request = self._requests.get()
        else: # mode == self.MODE_ONE_SHOT
            priority, timestamp, request = self._requests.get_nowait()
        #log.debug("scheduler: GOT request")
        if request.priority == PRIORITY_CONTROL: # sentinel
            if request.path == "<stop>":
                self._scannerPool.release()
                raise self.StopScheduler()
            elif request.path == "<pause>":
                self._scannerPool.release()
                self._pauseEvent.set() # tell .pause() that Scheduler has paused
                #log.debug("scheduler: paused")
                self._resumeEvent.wait()
                #log.debug("scheduler: resuming")
                return

        if self._requestStartedCB:
            self._requestStartedCB(request)

        # Get the appropriate Language Engine Drivers.
        drivers = self.manager.getCILEDrivers(request.language)
        if not drivers:
            self._scannerPool.release()
            log.warn("scheduler thread: drop request %r: unsupported "
                     "content type: '%s'", request.id,
                     request.language)
            return

        # If this content has already been scanned by all registered drivers
        # (judging by 'path', 'mtime' and 'md5') then there is no point in
        # this request.
        if not request.force:
            cu = cx.cursor()
            try:
                # Get file ID.
                file_id = None
                cu.execute("SELECT id FROM language WHERE name=%s",
                           request.language)
                if cu.rowcount:
                    language_id = cu.fetchone()[0]
                    cu.execute("SELECT id FROM file WHERE compare_path=%s "
                               "AND language_id=%s",
                               request.getCanonicalPath(), language_id)
                    if cu.rowcount:
                        file_id = cu.fetchone()[0]

                if file_id:
                    cu.execute("SELECT generator, mtime, md5 FROM scan "
                               "WHERE file_id=%s", file_id)
                    gen2state = {}
                    for generator, mtime, md5sum in cu.fetchall():
                        gen2state[generator] = (mtime and int(mtime), md5sum)
                    for driver in drivers:
                        state = gen2state.get(driver.name, (None, None))
                        if state[0] < request.mtime: # if mtime in CIDB is out of date
                            request.calculateMD5()
                            if state[1] != request.md5sum:
                                #print "re-scanning '%s' because out-of-date: "\
                                #      "mtime:%r<%r? and md5:%r!=%r?"\
                                #      % (os.path.basename(request.path),
                                #         state[0], request.mtime,
                                #         state[1], request.md5sum)
                                break
                    else:
                        log.info("scheduler thread: drop request %r: the "
                                 "content for '%s' has already been scanned ",
                                 request.id, request.getCanonicalPath())
                        self._scannerPool.release()
                        return

                    request.calculateMD5()
                    cu.execute("SELECT generator FROM scan WHERE file_id=%s "
                               "AND md5=%s", file_id, request.md5sum)
                    rows = cu.fetchall()
                    if rows:
                        generators = [g for (g,) in rows]
                        generators.sort()
                        driverNames = [driver.name for driver in drivers]
                        driverNames.sort()
                        if generators == driverNames:
                            log.info("scheduler thread: drop request %r: path "
                                     "(%s) and md5 are already in database",
                                     request.id, request.getCanonicalPath())
                            self._scannerPool.release()
                            return
            finally:
                cu.close()

        ## If any of the current scan runners matches this request (in
        ## path and md5) then there is no point in this request.
        ##XXX Is this common enough to be of any benefit?
        #self._runnersLock.acquire()
        #try:
        #    request.calculateMD5()
        #    for runner in self._runners:
        #        if runner.finished:
        #            continue  # zombies don't count
        #        runner.request.calculateMD5()
        #        if (request.md5sum == runner.request.md5sum
        #            and request.getCanonicalPath() ==
        #                runner.request.getCanonicalPath()):
        #            log.info("scheduler thread: drop request %r: path (%s) "
        #                     "and md5 match an active scan", request.id,
        #                     request.getCanonicalPath())
        #            self._scannerPool.release()
        #            return
        #finally:
        #    self._runnersLock.release()

        # Optimization: If the same file basename and checksum are in the
        # database, then we can shortcircuit: copy the settings from what
        # is likely a duplicate of this file in another directory.
        #XXX

        # Let's always calculate the MD5 ourselves instead of possibly
        # getting the CILE to do it. Everybody and their daughters seems to
        # calculate a different MD5 digest. I had previously thought this was
        # pretty standard.
        request.calculateMD5()

        # Start a worker thread to handle this request.
        assert(self.POOL_SIZE == 1,
               "Scheduler POOL_SIZE is not one: must not shared database "
               "connection among concurrent Scan Runners")
        db = Database(self.manager, cx)
        runner = _ScanRunner(request, drivers, db, self._requestCompleted)
        self._runnersLock.acquire()
        try:
            self._runners.append(runner)
        finally:
            self._runnersLock.release()
        log.info("scheduler thread: start request %r: path '%s'",
                 request.id, os.path.basename(request.path))
        runner.start()


class BatchScheduler(threading.Thread):
    #XXX 'splain usage
    def __init__(self, manager, progressCB=None, completedCB=None,
                 errorCB=None):
        threading.Thread.__init__(self, name="Code Intel Batch Scheduler")
        self.manager = manager
        self.errors = []  # list of any errors messages during processing
        self._progressCB = progressCB # progress callback
        self._batchCompletedCB = completedCB # completed callback
        self._errorCB = errorCB # error callback
        self._requests = Queue.Queue()
        self._scheduler = None
        # boolean indicating if batch scheduler should stop
        self._stopped = False
        self._completionReason = None

    def addRequest(self, type, path, language=None, extra=None):
        """Add a batch request
        
            "type" is one of "language" (meaning a language installation),
                "cix" (meaning import a CIX file), or "directory" (meaning
                scan a source directory).
            "path" is the full path to a language install, CIX file, or
                source directory, depending on the type.
            "language" is the language of the language installation, None
                for type=="cix", and the language of the library dir.
            "extra" is an optional (null if not used) extra value depending
                on the type, path and/or language of the request that may be
                request for processing it. For example, a PHP language batch
                update request uses the "extra" field to specify the
                "php.ini"-config-file path.
        """
        if self.isAlive():
            raise CodeIntelError("cannot add a batch update request while "
                                 "the batch scheduler is alive")
        if type in ("language", "cix", "directory"):
            self._requests.put( (type, path, language, extra) )
        else:
            raise CodeIntelError("unknown batch update request type: '%s'"
                                 % type)

    def getNumFilesToProcess(self):
        """Return the number of files remaining to process."""
        if self._scheduler:
            return self._scheduler.getNumRequests()
        else:
            return 0

    def stop(self):
        """Cancel the running batch update."""
        self._stopped = True
        if self._scheduler:
            self._scheduler.stop()
        try:
            self.join()
        except AssertionError:
            pass # thread was not started or is no longer alive

    def _subRequestStarted(self, request):
        """Used to replace managers"""
        if self._progressCB:
            self._progressCB("Scanning", request)

    def _subCompletedCB(self, reason):
        self._completionReason = reason

    #XXX This must manually be kept in sync with cidb_schema.sql.
    _indeces = [
        ("file_compare_path_index", "file", "(compare_path)"),
        ("file_language_id_index", "file", "(language_id)"),
        ("file_compare_path_and_language_id_index", "file", "(compare_path, language_id)"),
        ("module_file_id_index", "module", "(file_id)"),
        ("module_scan_id_index", "module", "(scan_id)"),
        ("symbol_file_id_index", "symbol", "(file_id)"),
        ("symbol_scan_id_index", "symbol", "(scan_id)"),
        ("symbol_module_id_index", "symbol", "(module_id)"),
        ("symbol_scope_id_index", "symbol", "(scope_id)"),
        ("base_scan_id_index", "base", "(scan_id)"),
        ("base_symbol_id_index", "base", "(symbol_id)"),
        ("type_scan_id_index", "type", "(scan_id)"),
        ("type_symbol_id_index", "type", "(symbol_id)"),
        ("type_base_id_index", "type", "(base_id)"),
        ("import_scan_id_index", "import", "(scan_id)"),
        ("import_scope_id_index", "import", "(scope_id)"),
    ]
    def _removeCIDBIndices(self, cx):
        cu = cx.cursor()
        try:
            length = len(self._indeces)
            for i in range(length):
                name, table, values = self._indeces[i]
                if self._progressCB:
                    self._progressCB("Preparing database", (name, i, length))
                try:
                    cu.execute("DROP INDEX %s;" % name)
                except sqlite.DatabaseError:
                    log.warn("database error dropping '%s' index" % name)
            cx.commit()
        finally:
            cu.close()
    def _restoreCIDBIndices(self, cx):
        cu = cx.cursor()
        try:
            length = len(self._indeces)
            for i in range(length):
                name, table, values = self._indeces[i]
                if self._progressCB:
                    self._progressCB("Restoring indices", (name, i, length))
                try:
                    cu.execute("CREATE INDEX %s ON %s %s;"
                               % (name, table, values))
                except sqlite.DatabaseError:
                    log.warn("database creating '%s' index" % name)
            cx.commit()
        finally:
            cu.close()

    def run(self):
        log.debug("batch scheduler thread: start")
        self.errors = []
        self._scheduler = Scheduler(Scheduler.MODE_ONE_SHOT,
                                    self.manager,
                                    None,
                                    self._subRequestStarted,
                                    self._subCompletedCB)
        cx = self.manager.getDatabaseConnection()
        #self._removeCIDBIndices(cx)
        try:
            while 1:
                try:
                    type, path, language, extra = self._requests.get_nowait()
                    if self._stopped:
                        log.debug("batch scheduler thread: stopped")
                        break
                    log.debug("handle %r batch update request: path=%r, "
                              "language=%r extra=%r", type, path, language,
                              extra)
                    if type == "language":
                        self._handleLanguageRequest(cx, path, language, extra)
                    elif type == "cix":
                        self._handleCIXRequest(cx, path)
                    elif type == "directory":
                        self._handleDirectoryRequest(cx, path, language)
                except Queue.Empty:
                    log.debug("batch scheduler thread: request queue empty")
                    break
                except Exception, ex:
                    self._logError("unexpected error handling batch "
                                   "update: %s" % ex)
                    break

            if self.errors:
                log.debug("batch scheduler thread: error out")
                self._completionReason = "error"
                return

            # Phase 2: start the scheduler and wait for it to complete.
            log.debug("batch scheduler thread: start sub-scheduler")
            self._scheduler.start()
            log.debug("batch scheduler thread: join sub-scheduler")
            self._scheduler.join()
            log.debug("batch scheduler thread: joined sub-scheduler")
        finally:
            #try:
            #    self._restoreCIDBIndices(cx)
            #except:
            #    pass
            cx.close()
            if self._batchCompletedCB:
                try:
                    self._batchCompletedCB(self._completionReason)
                except:
                    log.exception("error in batch scheduler completion "
                                  "callback")
        log.debug("batch scheduler thread: end")

    def _handleCIXRequest(self, cx, path):
        if self._progressCB:
            self._progressCB("Importing", path)
        #XXX Consider not bothering if MD5 of file is already in DB.
        #       md5sum = md5.new(cix).hexdigest()
        #    c.f. "Can an MD5 for a CIX file be added?" in my notes.
        try:
            fin = open(path, 'r')
            try:
                cix = fin.read()
            finally:
                fin.close()
        except EnvironmentError, ex:
            self._logError("Error importing CIX file '%s': %s" % (path, ex))
            return
        db = Database(self.manager, cx)
        try:
            db.update(cix, recover=0)
        except CodeIntelError, ex:
            self._logError("Error importing CIX file '%s': %s" % (path, ex))
            return

    def _logError(self, errmsg):
        self.errors.append(errmsg)
        if self._errorCB:
            self._errorCB(errmsg)

    def _handleLanguageRequest(self, cx, path, language, extra):
        if language == "*":
            languages = self.manager.getSupportedLanguages()
        else:
            languages = [language]

        # See if have any pre-created CIX files for this language to use
        # instead of or in addition to actually scanning the core library.
        if "Python" in languages:
            pythonCIX = os.path.join(os.path.dirname(__file__), "python.cix")
            self._handleCIXRequest(cx, pythonCIX)

        for lang in languages:
            try:
                importer = self.manager.getImportHandler(lang)
            except CodeIntelError, ex:
                if language != "*":
                    self._logError("cannot handle 'language' batch update "
                                   "request for '%s': %s" % (lang, ex))
                continue
            try:
                importer.setCorePath(path, extra)
                UPDATE_EVERY = 50
                n = 0
                for file in importer.genScannableFiles(skipRareImports=True,
                                                       importableOnly=True):
                    if self._progressCB and n % UPDATE_EVERY == 0:
                        self._progressCB("Gathering files", n)
                    r = ScanRequest(file, lang, PRIORITY_IMMEDIATE)
                    self._scheduler.addRequest(r)
                    n += 1
            except CodeIntelError, ex:
                self._logError("error handling %s language request: %s"
                               % (language, ex))

    def _handleDirectoryRequest(self, cx, path, language):
        if language == "*":
            languages = self.manager.getSupportedLanguages()
        else:
            languages = [language]

        for lang in languages:
            try:
                importer = self.manager.getImportHandler(lang)
            except CodeIntelError, ex:
                if language == "*":
                    continue
                else:
                    raise CodeIntelError("cannot handle 'directory' batch "
                                         "update request for '%s': %s"
                                         % (lang, ex))
            UPDATE_EVERY = 10
            n = 0
            for file in importer.genScannableFiles([path]):
                if self._progressCB and n % UPDATE_EVERY == 0:
                    self._progressCB("Gathering files", n)
                r = ScanRequest(file, lang, PRIORITY_IMMEDIATE)
                self._scheduler.addRequest(r)
                n += 1





