#!/usr/bin/env python
# Copyright (c) 2000-2004 ActiveState
# See the file LICENSE.txt for licensing information.
#
#*************************************************************************
# NOTE: You should not run this script directly. You should run Komodo's
#       "install.sh" which will launch this script properly.
#*************************************************************************
#
# Komodo simple install script
#

import sys, os, shutil
import socket
import tempfile


#---- exceptions

class InstallError(Exception): pass


#---- globals

defaultInstallDir = "~/Komodo-3.0"
installDirOnBuildMachine = "/export/home/jeffh/p4/Komodo-devel/install/release/user_withmoz/Komodo-Professional-3.0.0-108475-solaris8-sparc"
xpcomSubdir = "lib/python2.2/site-packages"
verbose = 0
installLog = "koinstall.log"


#---- support functions

def _log(msg):
    f = open(installLog, 'a')
    f.write(msg)
    f.close()


def _isRegistered(comp, installDir):
    """Return true iff the given XPCOM component looks like it is properly
    registered.
    
    "comp" is a dictionary with 'lib', 'contractId', and 'name' keys
        describing the component.
    "installDir" is the path to where Komodo is being installed.
    
    The test is done by trying the use the component via XPCOM.
    """
    if comp["contractId"] == "@mozilla.org/inline-plugin/application/x-scimoz-plugin":
        # Cannot use PyXPCOM to check this component on Linux because this
        # happens:
        #   - Python's import xpcom with load libxpcom.so with RTLD_LOCAL
        #   - _isregistered.py will try to load libnpscimoz.so which will try
        #     to load some symbols from libxpcom.so, but libxpcom.so has
        #     already been loaded into the process so it presumes the symbols
        #     are already defined (presuming that libxpcom.so was loaded
        #     RTLD_GLOBAL): failure ensures because that assumption is faulty.
        # Solution is to go via js code.
        cmd = '%s "%s" _isSciMozRegistered.js >>%s 2>&1'\
              % (_getMozCmdPrefix(installDir, userDataDir),
                 os.path.join(installDir, "Mozilla", "xpcshell"), installLog)
    else:
        cmd = '%s "%s" _isregistered.py "%s" >>%s 2>&1'\
              % (_getMozCmdPrefix(installDir, userDataDir), sys.executable,
                 comp["contractId"], installLog)
    _log("_isRegistered(%r): running %r\n" % (comp["contractId"], cmd))
    retval = os.system(cmd)
    _log("retval is %r\n" % retval)
    return not retval



def AskYesNo(prompt="?", default=1):
    """Ask a yes no question on the command line and return the response.
    Returns true (1) if answer is yes, false (0) if answer is no.
    """
    import string
    validChoices = {'yes':1, 'y':1, 'ye':1, 'no':0, 'n':0}
    if default == 1:
        postPrompt = ' [Y/n] '
    else:
        postPrompt = ' [y/N] '
    while 1:
        choice = string.lower(raw_input(prompt + postPrompt))
        if choice == '':
            return default
        elif choice in validChoices.keys():
            return validChoices[choice]
        else:
            print "Please repond with 'yes' or 'no' (or 'y' or 'n')."


def MyMkdir(newdir):
    """mymkdir: works the way a good mkdir should :)
        - already exists, silently complete
        - regular file in the way, raise an exception
        - parent directory(ies) does not exist, make them as well
    """
    import os
    global verbose

    if os.path.isdir(newdir):
        pass
    elif os.path.isfile(newdir):
        raise InstallError("a file with the same name as the desired " \
            "dir, '%s', already exists." % newdir)
    else:
        head, tail = os.path.split(newdir)
        if head and not os.path.isdir(head):
            MyMkdir(head)
        if tail:
            if verbose:
                print "mkdir %s" % repr(newdir)
            os.mkdir(newdir)


def MyCopy(src, dst):
    """mycopy: works the way a good copy should:
        - no source, raise an exception
        - destination directory, make a file in that dir named after src
        - source directory, recursively copy the directory to the destination
        - filename wildcarding allowed
    """
    import string, glob, os, shutil
    global verbose

    # determine if filename wildcarding is being used
    # (only raise error if non-wildcarded source file does not exist)
    if string.find(src, '*') != -1 or \
       string.find(src, '?') != -1 or \
       string.find(src, '[') != -1:
        usingWildcards = 1
        srcFiles = glob.glob(src)
    else:
        usingWildcards = 0
        srcFiles = [src]

    for srcFile in srcFiles:
        if os.path.isfile(srcFile):
            if usingWildcards:
                srcFileHead, srcFileTail = os.path.split(srcFile)
                srcHead, srcTail = os.path.split(src)
                dstHead, dstTail = os.path.split(dst)
                if dstTail == srcTail:
                    dstFile = os.path.join(dstHead, srcFileTail)
                else:
                    dstFile = os.path.join(dst, srcFileTail)
            else:
                dstFile = dst
            dstFileHead, dstFileTail = os.path.split(dstFile)
            if not os.path.isdir(dstFileHead):
                MyMkdir(dstFileHead)
            if verbose:
                print "copy %s to %s" % (repr(srcFile), repr(dstFile))
            if os.path.isfile(dstFile):
                # make sure 'dstFile' is writeable
                os.chmod(dstFile, 0755)
            shutil.copy(srcFile, dstFile)
        elif os.path.isdir(srcFile):
            srcFiles = os.listdir(srcFile)
            if not os.path.exists(dst):
                MyMkdir(dst)
            for f in srcFiles:
                s = os.path.join(srcFile, f)
                d = os.path.join(dst, f)
                try:
                    MyCopy(s, d)
                except (IOError, os.error), why:
                    raise InstallError("Can't copy %s to %s: %s"\
                          % (repr(s), repr(d), str(why)))
        elif not usingWildcards:
            raise InstallError("Source file %s does not exist" % repr(srcFile))

def _rmtreeOnError(rmFunction, filePath, excInfo):
    if excInfo[0] == OSError:
        # presuming because file is read-only
        os.chmod(filePath, 0777)
        rmFunction(filePath)

def _rmtree(dirname):
    import shutil
    shutil.rmtree(dirname, 0, _rmtreeOnError)



def _getMozCmdPrefix(installDir, userDataDir):
    """Return a command prefix to properly setup the environment to run a
    Komodo/Mozilla program.

        "installDir" is the path where Komodo is being installed.
        "userDataDir" is a directory to which to redirect Komodo's user data
            dir.

    Note, I am reproducing the environment that komodo.cpp sets up. Those
    must be kept in sync.
    """
    path = [ os.path.join(installDir, "python", "bin"),
             os.path.join(installDir, "Mozilla") ]
    if os.environ.has_key("PATH"):
        path += os.environ["PATH"].split(os.pathsep)
    ld_library_path = [ os.path.join(installDir, "Mozilla") ]
    #XXX Do we need this on LD_LIBRARY_PATH anymore?
    #       os.path.join(installDir, "Mozilla", "components")
    if os.environ.has_key("LD_LIBRARY_PATH"):
        ld_library_path += os.environ["LD_LIBRARY_PATH"].split(os.pathsep)
    pythonhome = os.path.join(installDir, "python")
    pythonpath = [ os.path.join(installDir, "Mozilla", "python") ]
    #XXX Do we need this on PYTHONPATH anymore?
    #       os.path.join(installDir, "python")
    prefix = "PATH=%s LD_LIBRARY_PATH=%s PYTHONHOME=%s PYTHONPATH=%s "\
             "KOMODO_HOSTNAME=%s KOMODO_USERDATADIR=%s %s"\
             % (os.pathsep.join(path), os.pathsep.join(ld_library_path),
                pythonhome, os.pathsep.join(pythonpath),
                socket.gethostname(), userDataDir,
                os.path.join(installDir, "Mozilla", "run-mozilla.sh"))
    return prefix


def DoInstall(userDataDir):
    print "Installing ActiveState Komodo."
    # Remove an old install log if there is one.
    if os.path.exists(installLog):
        os.unlink(installLog)
    _log("#---- Installing Komodo\n")

    # get install directory from user
    installDir = raw_input("Enter the directory for the Komodo "\
        "installation,\nor press Enter to accept the default [%s]. " %\
        defaultInstallDir)
    if installDir.strip() == "":
        installDir = defaultInstallDir
    installDir = os.path.abspath(os.path.normpath(os.path.expanduser(installDir)))
    if os.path.exists(installDir):
        print """
ERROR: '%s' already exists. You must install Komodo to a new
       directory to ensure a clean installation. Please rerun the
       install script and select a different installation location.

Aborting the install.
""" % installDir
        sys.exit(1)
    print "\nInstalling Komodo to '%s'." % installDir

    # warning about disk space size
    print "\nKomodo installation requires about 100 megabytes. Please"
    print "make sure enough free space is available before continuing."
    if not AskYesNo("Would you like to proceed?"):
        print "Aborting install."
        sys.exit(1)

    _log("#---- dumping user's environment\n")
    os.system("set >>%s" % installLog)
   
    # copy the entire "Komodo" tree to the installDir
    print "\nInstalling Komodo (this may take a couple of minutes)..."
    try:
        MyMkdir(installDir)
        MyCopy("INSTALLDIR", installDir)
    except OSError, err:
        if str(err).find("Permission denied"):
            print "\nERROR: You do not have permission to install "\
                  "Komodo in '%s'. Aborting install." % installDir
            sys.exit(1)
        else:
            raise err
    
    # put link to Komodo in /usr/local/bin?
    if installDir.startswith("/usr/local"):
        linkDir = "/usr/local/bin"
        link = os.path.join(linkDir, "komodo")
        if AskYesNo("Would you like a symbolic link to 'komodo' "\
            "from %s?" % linkDir):
            createLink = 1
            if os.path.islink(link):
                print "WARNING: A symbolic link '%s' already exists." % link
                if AskYesNo("Would you like to overwrite this link?"):
                    os.unlink(link)
                else:
                    createLink = 0
            elif os.path.isfile(link):
                print "WARNING: A file '%s' already exists." % link
                if AskYesNo("Would you like to overwrite this file?"):
                    os.unlink(link)
                else:
                    createLink = 0
            if createLink:
                print "Creating link to Komodo at '%s'." % link
                try:
                    os.symlink(os.path.join(installDir, "komodo"), link)
                except OSError, err:
                    if str(err).find("Permission denied"):
                        print "\nERROR: You do not have permission to "\
                            "create a link in '%s'. Skipping link." %\
                             linkDir
                    else:
                        raise err
            else:
                print "Not creating link."
    
    # run regxpcom
    # needed because:
    #    (1) If a priviledged user installs Komodo, regxpcom must be run
    #        with those privs to generate the chrome and XPCOM
    #        registries.
    #    (2) Komodo's license mechanism doesn't properly read the
    #        installed license if regxpcom is run in the same Mozilla
    #        session.
    _log("#---- registering XPCOM components\n")
    print "Creating XPCOM registry..."
    # first clean out registry
    filesToClean = [
        os.path.join(installDir, "Mozilla", "components", "compreg.dat"),
        os.path.join(installDir, "Mozilla", "components", "xpti.dat")]
    for file in filesToClean:
        os.system("rm -f %s" % file)
    # Would like to redirect stderr to /dev/null as well but this
    # causes, for some reason case, (2) from above to occur. Go figure.
    cmd = "(%s %s >/dev/null)" %\
          (_getMozCmdPrefix(installDir, userDataDir),
           os.path.join(installDir, "Mozilla", "regxpcom"))
    _log("cmd is %r\n" % cmd)
    os.system(cmd)

#XXX It is not clear that this mechanism of trying to hackily
#    register our CPP components is necessary anymore. As well,
#    it does not appear that the x-scimoz-plugin registration _can_
#    be checked the way it was, anymore. Skip this for now. We can
#    revive it if it proves necessary.
##    cppComponents = [
##        {"lib": os.path.join(installDir, "Mozilla", "components", "libnpscimoz.so"),
##         "contractId": "@mozilla.org/inline-plugin/application/x-scimoz-plugin",
##         "name": "SciMoz",
##        },
##        {"lib": os.path.join(installDir, "Mozilla", "components", "libkoLicense.so"),
##         "contractId": "@activestate.com/koLicense;8",
##         "name": "License Manager",
##        },
##    ]
##    failedComponents = []
##    for c in cppComponents:
##        for i in range(5): # sentinel
##            if not _isRegistered(c, installDir):
##                cmd = "touch %s" % c['lib']
##                _log("touch components: cmd=%r\n" % cmd)
##                os.system(cmd)
##                cmd = "(%s %s >>%s)" %\
##                      (_getMozCmdPrefix(installDir, userDataDir),
##                       os.path.join(installDir, "Mozilla", "regxpcom"),
##                       installLog)
##                _log("re-register components: cmd=%r\n" % cmd)
##                os.system(cmd)
##            else:
##                break
##        else:
##            failedComponents.append(c["name"])
##    #XXX Might want to read in Mozilla/components/*.dat
##    if failedComponents:
##        raise InstallError("""\
##Could not register some XPCOM components: %r.
##Please attach your install log, \"%s\", to this Komodo bug:
##   http://bugs.activestate.com/Komodo/show_bug.cgi?id=21386
##""" % (failedComponents, installLog))

    return installDir
    

def DoInstallWithRedirect():
    """Do the install, but redirect the user data dir to a temp location to
    avoid the problem described in bug 32270.
    """
    tempDir = os.path.join(tempfile.gettempdir(), "koinstall.%s" % os.getpid())
    try:
        return DoInstall(tempDir)
    finally:
        if os.path.exists(tempDir):
            _rmtree(tempDir)


#---- mainline

if __name__ == "__main__":
    try:
        installDir = DoInstallWithRedirect()
    except KeyboardInterrupt:
        print "\n\nUser abort of install.\n"
    except SystemExit:
        pass
    except InstallError, ex:
        print
        print '-'*70
        print "Install Error:"
        print ex
        print '-'*70
    except:
        import traceback
        print """
%s
The following error occurred while trying to install Komodo:
""" % ('-'*70)
        traceback.print_exc()
        print '-'*70
        
        print """
Please report installation troubles to komodo-feedback@ActiveState.com
or log a bug at http://bugs.ActiveState.com/Komodo
And please include the install log, \"%s\", with your bug report.
""" % installLog
        print
    else:
        print """
Finished installing Komodo.

Please send us your feedback:
    <mailto:komodo-feedback@ActiveState.com>

Be sure to put '%s' on your PATH.
Enjoy!
""" % installDir


